package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

// So that we can source events.
import javax.swing.event.EventListenerList;

/**
 * This class is an overall container for the DPRP Client. It can be
 * instantiated, and once done, contains a vector of all of the leases
 * currently outstanding for this client. Furthermore, it can spawn
 * new 'DPRPClientSession' threads, in order to obtain & update new
 * leases. <p>
 *
 * This is going to need some major help so that this class can fire events
 * at the GUI. Hell, maybe the thread should be the one firing the events.
 * Well, it should be here, because this is the one that is interrupting
 * the caller thread (ugly like Mike's mom). Anyways, the help for
 * <b><tt>EventListenerList</tt></b> in the JDK 1.3.1 documentation is
 * the most lucid source of information that I have at at this point. I'm
 * still confused as to what sort of interface I need to create (probably
 * <b><tt>DPRPEventListener</tt></b>), and what this interface needs to
 * define. <p>
 *
 * Everything sucks.
 *
 * @version 1.1 - November 15th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPClient implements DPRPLeaseListenerConstants,
				   DPRPClientConstants {
    /** Controls the printing of debug messages. */
    public static boolean DEBUG = false;

    /**
     * Notifies the session that we already have a valid lease, and that
     * it should simply sit on it (i.e. manage it).
     */
    public DPRPLease gotLease = null;

    /**
     * Set by the caller, whenever it wants to interrupt the thread with
     * new instructions.
     */
    public int sentRequest;
    /**
     * For the legacy client interrupt method, passes a reference to
     * a <b><tt>DPRPLeaseEvent</tt></b> object, which gives some details about
     * this event..
     */
    public DPRPLeaseEvent returnLeaseEvent;

    /** The IP address of the gateway, that we connect to. */
    private InetAddress srvIP = null;
    /**
     * The UDP port on the server to connect to. This should really specify
     * the main port on which the server is listening, and not ever
     * point to any "per session" ports.
     */
    private int srvPort = DPRPMessage.DEFAULT_PORT;
    /** The IP address on the local machine that we are using. */
    private InetAddress cliIP = null;
    /** The port on the local machine that we are using. */
    private int cliPort = 0;
    /** The port to request from the gateway. */
    private int reqPort = 0;
    /** The lease duration to request from the gateway.  */
    private int reqDuration = 0;
    /** The flags to accompany all messages to the gateway. */
    private short reqFlags = 0;
    /** The type of port to request from the gateway. */
    private byte reqProtocol = 0;

    private int extPort = 0;
    private InetAddress extIP = null;

    /**
     * Internal variable, contains a reference to all outstanding leases
     * for this client.
     */
    private Vector obtainedLeases;
    /** A reference to the thread that started this object. */
    private Thread callerThread = null;
    /**
     * A list of event listeners, who need to be notified whenever we muck
     * with a lease.
     */
    private EventListenerList listenerList = null;
    /** The Thread object for the client session. */
    private Thread sessionThreadID = null;

    /**
     * Creates a new <b><tt>DPRPClient</tt></b> object instance.
     */
    public DPRPClient() {
	// null constructor, do anything?
	obtainedLeases = new Vector();
	listenerList = new EventListenerList();
    }

    /**
     * Creates a new <b><tt>DPRPClient</tt></b> object instance.
     * @param ct A reference to a thread object, so that we can
     *  interrupt them if we so choose.
     */
    public DPRPClient (Thread ct) {
	callerThread = ct;
	obtainedLeases = new Vector();
	listenerList = new EventListenerList();
    }

    /**
     * Creates a new <b><tt>DPRPClient</tt></b> object instance. I
     * don't think that this constructor is ever called.
     */
    public DPRPClient (InetAddress sIP, InetAddress cIP, int rPort,
		       int rD, short rF, byte rProt) {
	srvIP = sIP;
	cliIP = cIP;
	reqPort = rPort;
	reqDuration = rD;
	reqFlags = rF;
	reqProtocol = rProt;

	obtainedLeases = new Vector();
	listenerList = new EventListenerList();
    }

    /**
     * Creates a new <b><tt>DPRPClient</tt></b> object from a
     * pre-existing lease. This is used when a GUI user loads a batch
     * of saved leases from disk. A <b><tt>DPRPClient</tt></b> must be
     * started for each lease (via this constructor), and the
     * appropriate threads must be started. The session threads detect
     * that they are being started from a previously negotiated (and
     * still valid) lease, and thus don't try and get a new lease at
     * the outset.
     * @param il A valid <b><tt>DPRPLease</tt></b>, to base this session
     *  around.
     */
    public DPRPClient (DPRPLease il) {
	// We're starting from a valid lease, folks.
	gotLease = il;

	// Doh. I think that I have to shake the lease-variables out here...
	srvIP = il.getServerIP();
	srvPort = il.getServerPort();
	cliIP = il.getClientIP();
	cliPort = il.getClientPort();
	reqPort = il.getExternalPort();
	reqProtocol = il.getProtocol();
	reqFlags = il.getFlags();
	reqDuration = il.getDuration();

	obtainedLeases = new Vector();
	listenerList = new EventListenerList();	
    }

    /**
     * Add a lease to the internal vector-o-leases.
     * @param lea The 'DPRPLease' to be added.
     */
    public synchronized void addLease (DPRPLease lea) {
	obtainedLeases.add (lea);
	if (DEBUG)
	    System.out.println ("PC: lease added: " + lea);

	// Interrupt whomever called me -- they have something new to
	// observe.
	DPRPLeaseEvent ev = new DPRPLeaseEvent (this, lea, NEW_LEASE_EVENT);
	ev.setErrorcode (DPRPErrors.CLI_ACQUIRE_SUCCESS);

	// If we were initialized with a lease, we'll mark this event,
	// so as to supress the "success" dialog in the GUI.
	if (gotLease != null)
	    ev.markAsOld();

	notifyController (ev);
    } /* End addLease(). */

    /**
     * Remove a lease from the internal vector-o-leases.
     * @param lea A lease to be removed.
     */
    public synchronized void rmLease (DPRPLease lea, DPRPLeaseEvent ev) {
	if (DEBUG)
	    System.out.println ("PC: removing lease: " + lea);
	if ((lea != null) && (!obtainedLeases.remove (lea)))
	    System.out.println ("** PC: Error removing lease. **");

	// Interrupt whomever called me -- they have something new to
	// observe.
	ev.setCommand (REMOVE_LEASE_EVENT);
	notifyController (ev);
    } /* End rmLease(). */

    /**
     * Update a lease in the internal vector-o-leases. This is only ever
     * called if the update has been successful. So, we always know
     * that our notification is going to be one of success. Yay.
     * @param curLea The current lease in the database.
     * @param newLea The lease that is to replace the current lease.
     */
    public synchronized void updateLease (DPRPLease curLea, DPRPLease newLea) {
	if (DEBUG) {
	    System.out.println ("PC: updating lease: ");
	    curLea.prettyPrint();
	    System.out.println ("PC: with this new one: ");
	    newLea.prettyPrint();
	    System.out.println ("PC: # of renews for this lease: " +
				(curLea.getNumRenews() + 1));
	}

	// Update the number of times renewed.
	// Why am I doing this here?
	newLea.setNumRenews (curLea.getNumRenews() + 1);

	// find the lease, and put the new one in.
	if (!obtainedLeases.remove (curLea))
	    System.out.println ("** PC: Error removing lease in update. **");
	obtainedLeases.add (newLea);

	// Interrupt whomever called me -- they have something new to
	// observe.
	DPRPLeaseEvent ev;
	ev = new DPRPLeaseEvent (this, newLea, UPDATE_LEASE_EVENT);
	ev.setErrorcode (DPRPErrors.CLI_UPDATE_SUCCESS);
	notifyController (ev);
    } /* End updateLease(). */

    /**
     * Forces this client object to instantiate a new session thread, which
     * does the actual lease gettin'/renewin'.
     */
    public void kickoffNewRequest() {
	// instantiate a new DPRPClientSession for a lease
	DPRPClientSession pcs = new DPRPClientSession (this);
	sessionThreadID = pcs.getThreadID();
	pcs.start();
    } /* End kickoffNewRequest(). */

    public void waitForSession() {
	try {
	    sessionThreadID.join();
	} catch (InterruptedException e) {
	    if (DEBUG)
		System.out.println ("DC: Got interrupted performing join(): "
				    + e);
	}
    } /* End waitForSession(). */

    /**
     * Used by the GUI in order to pass immediate orders to the session
     * (i.e. user requests immediate update or termination).
     */
    public void interruptSession() {
	if (DEBUG)
	    System.out.println ("** Interrupting session object: " +
				sessionThreadID);
	sessionThreadID.interrupt();
    }

    /**
     * Sets the thread ID of the session, so that we can properly interrupt it.
     * This is all quite hacky -- the thread must call back on us after
     * it has been told to run, because only then, does the thread ID exist
     * proper.
     * @param tid A thread ID object, which we'll interrupt later.
     */
    public void setSessionThreadID (Thread tid) {
	sessionThreadID = tid;
    }

    /**
     * Notifies whomever is controlling us that something has happend. The
     * exact details of this "happening" are included in the lease event
     * object.
     * @param ev A <b><tt>DPRPLeaseEvent</tt></b>, to be sent to whomever
     *  is controlling us (CLI or GUI).
     */
    public void notifyController (DPRPLeaseEvent ev) {
	// update stupidVariables. Should account for errors/remove events
	// differently. eh?
	extIP = ev.getLease().getExternalIP();
	extPort = ev.getLease().getExternalPort();
	reqDuration = ev.getLease().getDuration();

	if (callerThread != null) {
	    // We have an old-style CLI client.
	    returnLeaseEvent = ev;
	    callerThread.interrupt();
	} else {
	    // We have a new-style, event-driven client.
	    ev.setSource (this);
	    fireReceiveLease (ev);
	}
    } /* End notifyController(). */

    /*
     * Accessor methods.
     */
    /** Sets the port on the server to connect to, initially. */
    public void setServerPort (int sPort) {
	srvPort = sPort;
    }
    /** Sets the duration to request. */
    public void setDuration (int durata) {
	reqDuration = durata;
    }
    /** Sets the port number to request. */
    public void setRequestedPort (int rP) {
	reqPort = rP;
    }
    /** Sets the flags that accompany all requests. */
    public void setFlags (short rF) {
	reqFlags = rF;
    }
    /** Sets the protocol to request. */
    public void setProtocol (byte p) {
	reqProtocol = p;
    }
    /** Sets the protocol to request. */
    public void setProtocol (String str) {
	try {
	    reqProtocol = DPRPMessage.stringToProt (str);
	} catch (DPRPInvalidProtocolException e) {}
    }
    /** Sets our local IP address. */
    public void setClientIP (InetAddress cIP) {
	cliIP = cIP;
    }
    /** Sets the server's IP address, to connect to. */
    public void setServerIP (InetAddress sIP) {
	srvIP = sIP;
    }

    /** Gets the port to connect to on the server. */
    public int getServerPort() {
	return (srvPort);
    }
    /** Gets the servers' internal IP address. */
    public InetAddress getServerIP() {
	return (srvIP);
    }
    /** Gets our local IP address. */
    public InetAddress getClientIP() {
	return (cliIP);
    }
    /** Gets the port that we're requesting from the gateway. */
    public int getRequestedPort() {
	return (reqPort);
    }
    /** Gets the duration that we are requesting. */
    public int getDuration() {
	return (reqDuration);
    }
    /** Gets the flags that we are sending in all requests. */
    public short getFlags() {
	return (reqFlags);
    }
    /** Gets the protocol that we are requesting for our port. */
    public byte getProtocol() {
	return (reqProtocol);
    }
    /** Returns the Thread object for the session. */
    public Thread getSessionThreadID() {
	return (sessionThreadID);
    }


    /** Sets the external IP address, after the lease has been acquired. */
    public void setExternalIP (InetAddress i) {
	extIP = i;
    }
    /** Sets the external port, after the lease has been acquired. */
    public void setExternalPort (int p) {
	extPort = p;
    }
    /** Gets the current notion of the external IP address. */
    public InetAddress getExternalIP() {
	return (extIP);
    }
    /** Gets the DPRP-leased external port. */
    public int getExternalPort() {
	return (extPort);
    }

    /**
     * Returns our current client IP address, as a string. If no such
     * address has been set, then we use the standard Java routines to
     * get "our best guess" of the local address.
     * @return A string, representing the dotted-quad version of our
     *  local IP address.
     */
    public String ourAddrString() {
	String ret = null;
	try {
	    if (cliIP != null)
		ret = cliIP.getHostAddress();
	    else {
		cliIP = InetAddress.getLocalHost();
		ret = cliIP.getHostAddress();
	    }
	} catch (UnknownHostException e) {
	    System.err.println ("** oAS: " + e);
	}

	return (ret);
    } /* End ourAddrString(). */

    /**
     * Returns our current notion of the server's IP address, as a
     * string. Someday, this might try and get all fancy and return
     * hostnames if available, or IP addresses if unavailable. But for
     * now, we gots what we gots. The algorithm followed here is to
     * check and see if <b><tt>srvIP</tt></b> has been set. If so, we
     * return that. Otherwise, we see if <b><tt>cliIP</tt></b> has
     * been set. It if has, then we find it's network address (based
     * upon a 255.255.255.0 mask), and guess that the first
     * addressable machine on that network (".1") is the
     * gateway. Finally, as a last resort, we return the string
     * "<tt>0.0.0.0</tt>" if we truly have nothing else.
     * @return A string representation of the server's address.
     */
    public String ourSrvAddrString() {
	String ret = null;
	try {
	    if (srvIP != null)
		ret = srvIP.getHostAddress();
	    else if (cliIP != null) {
		String cliStr = cliIP.getHostAddress();
		int lastDot = cliStr.lastIndexOf ('.');
		String cliNet = cliStr.substring (0, lastDot);
		cliNet += ".1";
		srvIP = InetAddress.getByName (cliNet);
		ret = srvIP.getHostAddress();
	    } else
		ret = "0.0.0.0";
	} catch (UnknownHostException e) {
	    System.err.println ("** oSAS: " + e);
	}

	return (ret);
    } /* End ourSrvAddrString(). */

    /**
     * Returns the port to connect to on the server, as a string.
     */
    public String ourSrvPortString() {
	return ("" + srvPort);
    }

    /** Prints all of the initial parameters to <tt>stdout</tt>. */
    public void dumpParameters() {
	System.out.println ("************** BEGIN CLIENT DUMP **************");
	System.out.println ("Received Parameters:");
	if (srvIP != null)
	    System.out.println ("  srvIP       = " + srvIP.getHostAddress());
	else
	    System.out.println ("  srvIP       = null");
	System.out.println ("  srvPort     = " + srvPort);
	System.out.println ("  reqPort     = " + reqPort);
	System.out.println ("  reqDuration = " + reqDuration);
	System.out.println ("  reqFlags    = " + reqFlags);
	System.out.println ("  protocol    = " + reqProtocol);
	if (cliIP != null)
	    System.out.println ("  cliIP       = " + cliIP.getHostAddress());
	else
	    System.out.println ("  cliIP       = null");
	System.out.println ("*************** END CLIENT DUMP ***************");
    } /* End dumpParameters(). */

    /**
     * Adds a listener to our internal list. This is the result of me trying
     * to figure out how Swing event <b>sourcing</b> works, simply by
     * reverse-engineering the functionality out of the JDK source. Since
     * this appears to work, I'd say that I was pretty damn successful.
     * @param pl A <b><tt>DPRPLeaseListener</tt></b> to be added as the 
     *  target of fired events.
     */
    public void addDPRPLeaseListener (DPRPLeaseListener pl) {
	listenerList.add (DPRPLeaseListener.class, pl);
	if (DEBUG) {
	    System.out.println ("aPLL: Added a new lease listener. " +
				"total listener count = " +
				listenerList.getListenerCount());
	    System.out.println ("Listner class: " + pl.getClass().getName());
	}
    }

    /**
     * Removes a lease listener from our internal list.
     * @param pl A <b><tt>DPRPLeaseListener</tt></b> who no longer wishes
     *  to receive events.
     */
    public void removeDPRPLeaseListener (DPRPLeaseListener pl) {
	listenerList.remove (DPRPLeaseListener.class, pl);
	if (DEBUG) {
	    System.out.println ("aPLL: Removed a lease listener. " +
				"total listener count = " +
				listenerList.getListenerCount());
	    System.out.println ("Listner class: " + pl.getClass().getName());
	}
    }

    /**
     * Send a new event to our listeners. This just makes an
     * <b><tt>DPRPLeaseEvent</tt></b> object, and calls the real
     * <code>fireReceiveLease()</code> method.
     * @deprecated This is only for backwards compatibility.
     * @param lea A lease to send.
     * @param reason Why we're sending said lease.
     */
    protected void fireReceiveLease (DPRPLease lea, char reason) {
	DPRPLeaseEvent eventToSend = new DPRPLeaseEvent (this, lea, reason);
	fireReceiveLease (eventToSend);
    }

    /**
     * Send a new event to our listeners. This is the real magilla, and
     * as it turns out, this method is mostly comments. There is some magic
     * in the listeners, that intercepts the calls to
     * <code>receiveLease()</code>, and schedules them to occur serially
     * in the worker thread.
     * @param eventToSend A nice <b><tt>DPRPLeaseEvent</tt></b> object,
     *  which I can customize to my will, passing all sorts of data in
     *  an event.
     */
    protected void fireReceiveLease (DPRPLeaseEvent eventToSend) {
	// Get the array of listeners -- targets for our transmission.
	Object[] listeners = listenerList.getListenerList();

        // Process the listeners last to first, notifying
        // those that are interested in this event
	/*
	 * ME - I have no clue as to why they are pruning two listeners,
	 * and then skipping down the list in two's. Fucking wierd.
	 *
	 * _Oh_ _boy_! It looks like the list has two different types of
	 * objects, interleaved. The first is a listener, the second is a
	 * reference to an object that we actually send to (or maybe the
	 * thread in that object)?
	 */
        for (int i = listeners.length-2; i>=0; i-=2) {
            if (listeners[i] == DPRPLeaseListener.class) {
		((DPRPLeaseListener)listeners[i+1]).receiveLease (eventToSend);
	    } /* if */
	} /* for */
    } /* End fireReceiveLease(). */

} /* End class DPRPClient. */
