package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class represents an individual DPRP Client
 * session. Essentially, this means that this object is instantiated
 * into a new thread. Then, it starts by negotiating a lease from the
 * server. If this is successful, it sleeps until such a time until it
 * must renew the lease. If the renewal process is successful, it will
 * sleep again, repeating this cycle for as long as it can.
 *
 * @version 1.2 - October 30th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPClientSession extends Thread implements DPRPClientConstants {
    /** The IP address of the server that will give us the goods. */
    private InetAddress serverIP;
    /** The port on the server to connect to. */
    private int serverPort;
    /** A specific port that we'd like to request. */
    private int requestedPort;

    /** Our local IP address, as specified when the client was launched. */
    private InetAddress myIP;
    /** Local port, associated with the socket. */
    private int myPort;
    /** Local socket, refreshed for each new request. */
    private DPRPSocket mySock;

    /** A reference to the client object that spawned us. */
    private DPRPClient client;
    /** The lease duration to request for our lease. */
    private int reqDuration;
    /** The flags to guide the lease request. */
    private short reqFlags;
    /** The protocol to request for our lease. */
    private byte reqProtocol;

    /** A reference to the current lease that we are holding. */
    private DPRPLease curLease = null;    // current lease we're operating on.

    /**
     * Constructs a new <b><tt>DPRPClientSession</tt></b> out of a 
     * <b><tt>DPRPClient</b></tt> object instance.
     * @param pcp A <b><tt>DPRPClient</tt></b> instance, to base ourselves
     *  off of.
     */
    public DPRPClientSession (DPRPClient pcp) {
	client = pcp;

	serverIP = pcp.getServerIP();
	serverPort = pcp.getServerPort();
	requestedPort = pcp.getRequestedPort();

	myIP = pcp.getClientIP();
	myPort = 0;
	mySock = null;

	reqDuration = pcp.getDuration();
	reqFlags = pcp.getFlags();
	reqProtocol = pcp.getProtocol();

	// Initialize my local IP address and port.
	if (myIP == null)
	    try {
		myIP = InetAddress.getLocalHost();
		myPort = DPRPMessage.DEFAULT_PORT;
	    } catch (UnknownHostException e) {
		System.err.println (getName() + "PCS: Unknown local host. " +
				    "Error:\n" + e);
		System.exit(1);
	    }
	else if (DPRPClient.DEBUG)
	    System.out.println (getName() + "PCS: Connecting via IP Address: "
				+ myIP.getHostAddress());
    } /* End DPRPClientSession constructor. */

    /**
     * Used by the parent <b><tt>DPRPClient</tt></b> in order to interrupt
     * us when (and if) needed.
     */
    public Thread getThreadID() {
	return (Thread.currentThread());
    }

    /**
     * The method invoked at thread creation, to start things off. I think
     * that maybe this method should fire error events at the GUI,
     * so that nice error notification can occur.
     */
    public void run() {
	// This fixes a most interesting bug. I can't tell the client
	// my threadID, until I actually start *executing* (instantiating
	// is not enough). So, before I did this, I was interrupting the
	// AWT Event Dispatching thread, which really killed that GUI, dead.
	client.setSessionThreadID (Thread.currentThread());

	// Reason why maintainCurrentLease() halted.
	int haltReason = 0;
	// Whether we want to jump back into maintaining our lease, or not.
	boolean continueLooping = true;
	// A lease event, to be zapped to the controller.
	DPRPLeaseEvent leaEv = new DPRPLeaseEvent (this, null, '0');

	if (client.gotLease != null) {
	    // okay, we want to skip the lease-getting, and just go into
	    // the loop. This is going to entail shaking the lease details
	    // out into our local variables.
	    curLease = client.gotLease;
	} else {

	    /*
	     * Step 1: get a new lease.
	     */
	    try {
		curLease = requestLease();
	    } catch (DPRPErrorMessageException e) {
		if (DPRPClient.DEBUG)
		    System.out.println (getName() + e);

		// Should notify the controlling client of this condition.
		leaEv.markAsError();
		//leaEv.setErrorcode (DPRPErrors.CLI_ACQUIRE_ERROR);
		leaEv.setErrorcode (e.getErrorcode());
		leaEv.setCommand (DPRPLeaseListenerConstants.NEW_LEASE_EVENT);
		client.notifyController (leaEv);

		// Kill this thread.
		return;
	    }
	}

	// This can still return NULL for other reasons (like, the server
	// isn't even there, and shit).
	if (curLease == null) {
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "PCS: Got null from " +
				    "requestLease(), notifying controller.");

	    // Should notify the controlling client of this condition.
	    leaEv.markAsError();
	    leaEv.setErrorcode (DPRPErrors.CLI_ACQUIRE_ERROR);
	    leaEv.setCommand (DPRPLeaseListenerConstants.NEW_LEASE_EVENT);
	    client.notifyController (leaEv);

	    // Kill this thread.
	    return;
	}

	// Success! Add this lease.
	client.addLease (curLease);

	/*
	 * Step 2: Keep trying to maintain this lease, while also handling
	 * user-supplied interrupts. The 'continueLooping' flag allows us
	 * to put cleanup code after this loop, if necessary.
	 */
	while (continueLooping) {
	    try {
		haltReason = maintainCurrentLease();
	    } catch (DPRPErrorMessageException e) {
		if (DPRPClient.DEBUG)
		    System.out.println (getName() + e);

		// Prepare the lease event.
		leaEv.markAsError();
		leaEv.setErrorcode (e.getErrorcode());

		// Because the rmLease() method implicitly performs the
		// notification, this code snippet will notify the controller
		// in the appropriate manner.
		if (curLease != null) {
		    leaEv.setLease (curLease);
		    client.rmLease (curLease, leaEv);
		} else {
		    leaEv.setCommand (DPRPLeaseListenerConstants.UPDATE_LEASE_EVENT);
		    client.notifyController (leaEv);
		}

		// Kill this thread. I'm going to take this sort of error to
		// be fatal.
		return;
	    } /* catch */

	    // Figure out why we exited, and deal.
	    switch (haltReason) {
	    case HALT_SERVER_NACK:
		if (DPRPClient.DEBUG)
		    System.out.println (getName() + "PCS: Server NACK'd " +
					"our lease.");
		continueLooping = false;

		// Set things up so that we can notify the controlling
		// client of this situation.
		leaEv.markAsError();
		leaEv.setErrorcode (DPRPErrors.CLI_SERVER_NACK_ERROR);
		leaEv.setLease (curLease);   // This has the NACK reason in it.
		// This will implicitly perform the notification.
		client.rmLease (curLease, leaEv);
		break;
	    case HALT_T2_EXPIRED:
		System.out.println (getName() + "PCS: T2 expired before " +
				    "lease could be renewed, giving up.");
		continueLooping = false;

		leaEv.markAsError();
		leaEv.setErrorcode (DPRPErrors.CLI_T2_EXPIRED);
		leaEv.setLease (curLease);
		client.rmLease (curLease, leaEv);
		break;
	    case HALT_NO_RESPONSE:
		System.out.println (getName() + "PCS: No response received " +
				    "while attempting to renew lease.");
		continueLooping = false;

		leaEv.markAsError();
		leaEv.setErrorcode (DPRPErrors.CLI_SRV_NO_RESPOND);
		leaEv.setLease (curLease);
		client.rmLease (curLease, leaEv);
		break;
	    case HALT_INVALID_SLEEP_TIMER:
		System.out.println (getName() + "PCS: Invalid time passed " +
				    "to 'sleep()' method call.");
		continueLooping = false;

		leaEv.markAsError();
		leaEv.setErrorcode (DPRPErrors.CLI_INTERNAL_ERROR);

		// In this case, it's best to rm the lease from the server.
		try {
		    terminateLease (curLease);
		} catch (DPRPErrorMessageException e) {
		    if (DPRPClient.DEBUG) {
			System.err.println (e);
		    }
		}
		leaEv.setLease (curLease);
		client.rmLease (curLease, leaEv);
		break;
	    case HALT_INTERRUPTED:
		// Okay, the controller interrupted us, check to see what
		// we have.
		if (client.sentRequest == CLI_FORCE_LEASE_UPDATE) {
		    if (DPRPClient.DEBUG)
			System.out.println (getName() + "PCS: Client forced " +
					    "a lease update.");

		    // update the specified lease right now!
		    // I think that curLease must reference a valid DPRPLease
		    // instance...
		    DPRPLease nl = null;
		    try {
			nl = renewLease (curLease);
		    } catch (DPRPErrorMessageException e) {
			// I surely hate code duplication.
			if (DPRPClient.DEBUG)
			    System.out.println (getName() + e);

			// Prepare the lease event.
			leaEv.markAsError();
			leaEv.setErrorcode (e.getErrorcode());

			// Because the rmLease() method implicitly performs the
			// notification, this code snippet will notify the
			// controller in the appropriate manner.
			if (curLease != null) {
			    leaEv.setLease (curLease);
			    client.rmLease (curLease, leaEv);
			} else {
			    leaEv.setCommand (DPRPLeaseListenerConstants.UPDATE_LEASE_EVENT);
			    client.notifyController (leaEv);
			}

			// Kill this thread. I'm going to take this sort of
			// error to be fatal.
			return;
		    } /* catch */

		    if ((nl != null) &&
			(nl.getFlags()!=DPRPErrors.CLIENT_LEASE_NOT_RENEWED)) {
			// This implictly handles the notification.
			client.updateLease (curLease, nl);
			curLease = nl;
		    } else if (nl != null) {
			// report NACK-reason to user.
			if (DPRPClient.DEBUG)
			    System.out.println (getName() + "PCS: Forced " +
			     "lease update was NACK'd. Reason: " +
			     DPRPErrors.resolveError(curLease.getNackError()));
			leaEv.setLease (curLease);
			leaEv.markAsError();
			leaEv.setErrorcode (DPRPErrors.CLI_SERVER_NACK_ERROR);
			client.rmLease (curLease, leaEv);
			continueLooping = false;
		    } else {
			// Update completely and utterly failed, report as
			// such, but don't rm the lease, because after all
			// it is still valid. I mean, this doesn't bode well,
			// but still...
			leaEv.setCommand (DPRPLeaseListenerConstants.UPDATE_LEASE_EVENT);
			leaEv.markAsError();
			leaEv.setErrorcode (DPRPErrors.CLI_UPDATE_ERROR);
			client.notifyController (leaEv);
		    }
		} else if (client.sentRequest == CLI_FORCE_LEASE_TERMINATE) {
		    // kill the specified lease right now!
		    continueLooping = false;
		    DPRPLease rl = null;
		    try {
			rl = terminateLease (curLease);
			if (rl == null) {
			    if (DPRPClient.DEBUG)
				System.out.println (getName() + "PCS: sent " +
						    "the terminate request " +
						    "successfully, deleting " +
						    "lease.");
			    // Delete away, punchy.
			    leaEv.setLease (curLease);
			    client.rmLease (curLease, leaEv);
			} else {
			    if (DPRPClient.DEBUG)
				System.out.println (getName() + "PCS: Got " +
						    "non-null from termLea.");

			    // delete curLease anyway? Try again? What?
			    leaEv.markAsError();
			    leaEv.setErrorcode (DPRPErrors.UNKNOWN_ERROR);
			    leaEv.setLease (curLease);
			    client.rmLease (curLease, leaEv);
			}
		    } catch (DPRPErrorMessageException e) {
			// Got an error message from server, delete lease and
			// mark the event as such.
			if (DPRPClient.DEBUG) {
			    System.out.println (getName() + "PCS: Got error" +
						" packet whilst trying to " +
						"terminate with server.");
			    System.out.println (getName() + e);
			}
			leaEv.markAsError();
			leaEv.setErrorcode (e.getErrorcode());
			leaEv.setLease (curLease);
			client.rmLease (curLease, leaEv);
		    }
		} else if ( client.sentRequest == CLI_TERMINATE_THREAD) {
		    if (DPRPClient.DEBUG)
			System.out.println (getName() + "PCS: controller is" +
					    "telling us to halt thread.");
		    leaEv.setLease (curLease);
		    client.rmLease (curLease, leaEv);
		    continueLooping = false;
		} else {
		    System.out.println (getName() + "PCS: Controlling code " +
					"sent a malformed interrupt reason.");
		    continueLooping = false;
		}

		// Don't fall into the default section. Gawrsh, this is
		// some complicated shit.
		break;
	    default:
		System.out.println (getName() + "PCS: Unknown haltcode '" +
				    haltReason + "' encountered. Exiting.");
		continueLooping = false;
	    } /* switch */

	    // Re-initialize our common lease event, just-in-case.
	    leaEv = new DPRPLeaseEvent (this, null, '0');
	} /* while */
    } /* End run(), and the thread, for that matter. */

    /**
     * Looks after the upkeep of the current lease, by renewing it before it
     * expires. This is accomplished by sleeping for <tt>t1</tt> seconds, and
     * then establishing a renew conversation with the server. If this is
     * unsuccessful (but not via an explicit NACK), then we sleep for
     * half of the difference between <tt>t2</tt> and <tt>t1</tt>, and
     * try again. This process continues, until either <tt>t2</tt> is
     * expired, the lease is expired, or a new lease is successfully
     * negotiated.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return An control code, specifying the reason for loop termination.
     */
    private int maintainCurrentLease() throws DPRPErrorMessageException {
	// Local variables.
	DPRPLease newLease = null;    // new lease we want to establish.
	boolean leaseRenewed = true;  // True if the lease was renewed.
	long t1 = 0, t2 = 0;          // relative renew timers, in ms.
	long renewTimeout = 0;        // Seconds to wait in-between successive
	                              // renew attempts. Halved for every
	                              // failure.
	Date t2_nr = null;            // non-relative version of t2
	int rtrC = 0;                 // Retry counter.

	long waitSpan = 0;

	try {
	    // Keep trying to renew this lease until we're told to stop
	    // for some reason.
	    while (true) {
		if (leaseRenewed) {
		    // This lovely block of code computes the two timers.
		    // First, we compute the 'waitSpan' -- which is the
		    // chunk of time from now until the lease expires. Then,
		    // we set t1 to go off after half of the 'waitSpan' has
		    // expired. Similarily, t2 is set to go off after 7/8 of
		    // the 'waitSpan' has expired.
		    // getTime() returns milliseconds.
		    waitSpan = curLease.getStopTime().getTime() - (new Date()).getTime();
		    t1 = (long)(0.5 * waitSpan);
		    t2 = (long)(0.875 * waitSpan);
		    t2_nr = new Date();
		    t2_nr.setTime (t2_nr.getTime() + t2);
		    renewTimeout = (long)(0.5 * (t2 - t1));
		} /* if lease was renewed */

		// wait t1; sleep is in milliseconds?
		if (DPRPClient.DEBUG)
		    System.out.println (getName() + "mCL: Going to sleep " +
					"for t1 = " + (t1/1000) + " seconds.");
		sleep (t1);

		if (DPRPClient.DEBUG)
		    System.out.println (getName() + "mCL: t1 expired, " +
					"starting renew process.");

		leaseRenewed = false;
		// so long as wall-clock time is not greater than t2...
		while ((t2_nr.after (new Date())) && (!leaseRenewed)) {
		    newLease = renewLease (curLease);

		    // if fail, keep trying until past t2...
		    if (newLease == null) {
			if (DPRPClient.DEBUG)
			    System.out.println(getName()+"mCL: Got null from "
						+ "renewLease(). Sleeping for "
						+ (renewTimeout/1000) +
						" seconds.");

			// wait for a bit, and try it all again. except for
			// the t1 sleep. This should be some sort of
			// exponential back-off type deal.
			sleep (renewTimeout);
			renewTimeout = (long)(0.5 * renewTimeout);
		    } else if (newLease.getFlags() == DPRPErrors.CLIENT_LEASE_NOT_RENEWED) {
			if (DPRPClient.DEBUG)
			    System.out.println (getName() + "mCL: Server " +
						"NACK'd our renew request.");

			// stop all of this.
			curLease = newLease;
			return (HALT_SERVER_NACK);
		    } else {
			if (DPRPClient.DEBUG)
			    System.out.println (getName() + "mCL: Server " +
						"renew'd us for another " +
						newLease.getDuration() +
						" seconds.");
			
			// hooray!
			leaseRenewed = true;

			// update, baby!
			client.updateLease (curLease, newLease);
			curLease = newLease;
		    } /* else */
		} /* while curTime */

		if ((t2_nr.before (new Date())) && (!leaseRenewed)) {
		    if (DPRPClient.DEBUG)
			System.out.println (getName() + "mCL: t2 expired " +
					    "without getting a " +
					    "valid\n\t'SRV_RENEW_ACK', so we "
					    + "must terminate the current " +
					    "lease... :(");
		    // t2 has expired, and no renew has come. we out.
		    return (HALT_T2_EXPIRED);
		}
		if ((rtrC == NUM_RETRIES) && (!leaseRenewed)) {
		    if (DPRPClient.DEBUG)
			System.out.println (getName() + "mCL: tried to " +
					    "renew lease " + NUM_RETRIES +
					    " times. Exiting.");
		    // Halt this session.
		    return (HALT_NO_RESPONSE);
		} else
		    rtrC++;

	    } /* while true */
	} catch (IllegalArgumentException e) {
	    if (DPRPClient.DEBUG)
		System.err.println (getName() + "mCL: Invalid sleep time " +
				    "specified: " + e);
	    // Halt this session.
	    return (HALT_INVALID_SLEEP_TIMER);
	} catch (InterruptedException e) {
	    if (DPRPClient.DEBUG)
	    System.err.println (getName() + "mCL: got interruption from " +
				"controller:\n\t" + e);
	    return (HALT_INTERRUPTED);
	}
    } /* End maintainCurrentLease(). */

    /**
     * Uses lower-level methods to request a lease from the
     * server. This method makes use of the
     * <b><tt>sendCliRequest()</tt></b> method, and then error-checks
     * the response, and generates a lease if all is well.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return A brand-spankin'-new lease, or <tt>null</tt> if the server
     *  wasn't feelin' it.
     */
    private DPRPLease requestLease() throws DPRPErrorMessageException {
	DPRPMessage curMessage = null;
	DPRPLease curLease = null;
	int suggestedPort = 0;
	int suggestedDuration = 0;

	// Acquire our local socket, so that we can get this proverbial
	// ball rolling.
	allocNewMySock();

	// Send client request.
	curMessage = sendCliRequest();
	if (curMessage == null) {
	    // Didn't get a valid message, punt.
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "rqL-Error: looks like the " +
				    "server is unavailable.");
	    return (null);
	}

	// Got a valid message.
	suggestedPort = curMessage.getExternalPort();
	suggestedDuration = curMessage.getDuration();

	if (DPRPClient.DEBUG)
	    System.out.println (getName() + "rqL: Got a valid SRV_OFFER.");

	// Examine the port that was suggested
	if ((suggestedPort != requestedPort) && (requestedPort != 0)) {
	    // Somebody didn't get what they asked for.
	    System.out.println ("rqL-Error: Server refused your requested " +
				"port. Either try again, request " +
				"a different port, or allow the server to " +
				"allocate a free port.");
	    // if not, nack it
	    curMessage = sendPortNack (curMessage);
	    return (null);
	} else {
	    // if we like it, start the whole 'finalize' process
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "rqL: Like the offer, " +
				    "attempting to ACK");

	    curMessage = sendPortAck (curMessage);
	    if (curMessage == null) {
		if (DPRPClient.DEBUG)
		    System.out.println (getName() + "rqL-Error: Server " +
					"didn't respond with a " +
					"valid ACK, aborting lease attempt.");
		return (null);
	    }

	    // Construct a local DPRPLease, and insert it into our db.
	    curLease = new DPRPLease (curMessage, true);
	    curLease.setServerPort (serverPort);
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "rqL: New lease created: " +
				    curLease);
	} /* else */

	// boy, this is a debate-and-a-half. To close, or not to close?
	mySock.close();
	mySock = null;
	return (curLease);
    } /* End requestLease(). */

    /**
     * Attempts to renew an outstanding lease. This method will
     * transmit a <b><tt>DPRP_CLI_RENEW_REQUEST</tt></b> message, and
     * deal with the server's reply.
     * @param lea The current outstanding lease to be renewed.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return A fresh lease, or null if one could not be obtained.
     */
    private DPRPLease renewLease (DPRPLease lea) throws DPRPErrorMessageException {
	// Check to see if the socket is closed. If it is, make a new one.
	if (mySock == null) {
	    allocNewMySock();
	} /* if mySock is null */

	// Create a new message, starting from the Lease.
	DPRPMessage renewMessage = lea.messageToClient (serverIP, serverPort);
	renewMessage.flipSendRecv();
	renewMessage.setSourcePort (mySock.getLocalPort());
	renewMessage.setMessageType (DPRPMessage.CLI_RENEW_REQUEST);

	// Blast it off, and see what we get back.
	DPRPMessage recvdMessage = sendAndRecvMessage (renewMessage);
	if (recvdMessage == null) {
	    if (DPRPClient.DEBUG)
		System.out.println(getName()+"PCS-rnL: Got null from server.");
	    return (null);
	} else if (DPRPClient.DEBUG) {
	    System.out.println (getName() + "PCS-rnL: Got Client Renew " +
				"Response:");
	    recvdMessage.printDump();
	}
	DPRPLease retLease = null;
	// Now, see if we got an ACK, NACK, or 'other'.
	if (recvdMessage.getMessageType() == DPRPMessage.SRV_RENEW_ACK) {
	    // Fantastic. Make a nice lease, and return it.
	    retLease = new DPRPLease (recvdMessage, true);
	    retLease.setServerPort (serverPort);
	} else if (recvdMessage.getMessageType()==DPRPMessage.SRV_RENEW_NACK) {
	    // Boo, hiss. I think this means we should stop all of this.
	    if (DPRPClient.DEBUG) {
		System.out.println (getName() + "PCS-rNL: Server denied lease"
				    + " renew request for leaseID: " +
				    recvdMessage.getPrimaryKey() +". Reason:");
		short reason = recvdMessage.getFlags();
		System.out.println (DPRPErrors.resolveError (reason));
	    }

	    // This is dumb. I should really be throwing an exception.
	    // That would let me eliminate CLIENT_LEASE_NOT_RENEWED and
	    // having the nack errorcode in the lease.
	    retLease = new DPRPLease (recvdMessage, true);
	    retLease.setFlags (DPRPErrors.CLIENT_LEASE_NOT_RENEWED);
	    retLease.setNackError (recvdMessage.getFlags());
	} else {
	    // Other.
	    mySock.sendError (renewMessage, DPRPErrors.INVALID_PACKET_TYPE);
	    retLease = null;
	}

	// Kill the socket, we're going to sleep for awhile. (Probably)
	mySock.close();
	mySock = null;
	return (retLease);
    } /* End renewLease(). */

    /**
     * Sends a <b><tt>DPRP_CLI_REQUEST</tt></b> message to the
     * server. It gains the necessary variables to include in the
     * packet from class-wide instance variables.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return A valid offer (in the form of a
     *  <b><tt>DPRPMessage</tt></b>), or <tt>null</tt> if there was a
     *  problem.
     */
    private DPRPMessage sendCliRequest() throws DPRPErrorMessageException {
	DPRPMessage sentMessage = createNewMessage();
	sentMessage.setMessageType (DPRPMessage.CLI_REQUEST);

	DPRPMessage recvdMessage = sendAndRecvMessage (sentMessage);
	if (recvdMessage != null) {
	    if (DPRPClient.DEBUG) {
		System.out.println (getName() + "sCR: Got Server Offer " +
				    "Message; Contents:");
		recvdMessage.printDump();
	    }

	    if (recvdMessage.getMessageType() == DPRPMessage.SRV_OFFER)
		return (recvdMessage);
	    else
		mySock.sendError (sentMessage, DPRPErrors.INVALID_PACKET_TYPE);
	}

	return (null);
    } /* End sendCliRequest(). */

    /**
     * Accepts an offered port from the server. This is accomplished
     * by sending a <b><tt>CLI_OFFER_ACK</tt></b> message, and then
     * waiting for the accompanying <b><tt>SRV_COMMIT_OFFER</tt></b>
     * message.
     * @param sentMessage A skeleton of the message to send (type and
     *  addresses are not correct).
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return The server offer message, or <tt>null</tt> if such a
     *  message could not be obtained.
     */
    private DPRPMessage sendPortAck (DPRPMessage sentMessage) throws DPRPErrorMessageException {
	sentMessage.setMessageType (DPRPMessage.CLI_OFFER_ACK);

	sentMessage.setDestinationAddress (sentMessage.getSourceAddress());
	sentMessage.setDestinationPort (sentMessage.getSourcePort());
	sentMessage.setSourceAddress (myIP);
	sentMessage.setSourcePort (myPort);

	DPRPMessage recvdMessage = sendAndRecvMessage (sentMessage);
	if (recvdMessage != null) {
	    if (DPRPClient.DEBUG) {
		System.out.println (getName() + "sPA: Port Ack Message " +
				    "received. Contents:");
		recvdMessage.printDump();
	    }
	    if (recvdMessage.getMessageType() == DPRPMessage.SRV_COMMIT_OFFER)
		return (recvdMessage);
	    else
		// WTF? Return an error. 5 is invalid packet type recv'd.
		mySock.sendError (sentMessage, DPRPErrors.INVALID_PACKET_TYPE);
	} /* if ! null */

	return (null);
    } /* End sendPortAck(). */

    /**
     * Sends 'CLI_OFFER_NACK' in response to offer contained in parameter
     * 'msg'. Expects to receive a 'SRV_OFFER_NACK_ACK' in response.
     * @param msg An offer message that we don't like. Basis for the Nack.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return A 'SRV_OFFER_NACK_ACK' message.
     */
    private DPRPMessage sendPortNack (DPRPMessage msg) throws DPRPErrorMessageException {
	DPRPMessage respMessage = null;
	DPRPMessage nackMessage = (DPRPMessage)msg.clone();
	nackMessage.flipSendRecv();
	nackMessage.setMessageType (DPRPMessage.CLI_OFFER_NACK);

	respMessage = sendAndRecvMessage (nackMessage);
	if (respMessage == null) {
	    // Uh-oh, looks like the server is dead. Not a big deal, eh?
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "sPN: server is dead. Nack " +
				    "not Ack'd, exiting...");
	    return (null);
	}

	if (DPRPClient.DEBUG) {
	    System.out.println (getName() + "sPN: got message from server:");
	    respMessage.printDump();
	}

	// Okay, we got a message, and we know it's not an ERROR.
	if (respMessage.getMessageType() == DPRPMessage.SRV_OFFER_NACK_ACK)
	    // this is what we came for
	    return (respMessage);
	else {
	    // dunno what this is
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "sPN; message wasn't what we" +
				    " were looking for. Sending error");
	    mySock.sendError (respMessage, DPRPErrors.INVALID_PACKET_TYPE);
	    return (null);
	}
    } /* End sendPortNack(). */

    /**
     * Sends a <code>CLI_TERMINATE_LEASE</code> message, immediately
     * revoking the given lease.
     * @param lea A lease for which we want to immediately revoke our
     *  handle.
     * @throws DPRPErrorMessageException This is passed on from the call to
     *  <b><tt>sendAndRecvMessage()</tt></b>.
     * @return A lease based upon the return message, or <code>null</code>.
     */
    private DPRPLease terminateLease (DPRPLease lea) throws DPRPErrorMessageException {
	// Check to see if the socket is closed. If it is, make a new one.
	if (mySock == null) {
	    allocNewMySock();
	} /* if mySock is null */

	DPRPMessage inMessage = new DPRPMessage();
	DPRPMessage outMessage = createNewMessage();
	outMessage.setMessageType (DPRPMessage.CLI_TERMINATE_LEASE);
	outMessage.setExternalIP (lea.getExternalIP());
	outMessage.setExternalPort (lea.getExternalPort());
	outMessage.setDuration (lea.getDuration());

	inMessage = sendAndRecvMessage (outMessage);
	DPRPLease ret = null;
	if ((inMessage == null) || (inMessage.getMessageType() == DPRPMessage.SRV_TERMINATE_ACK)) {
	    if (DPRPClient.DEBUG)
		System.out.println (getName() + "trL: got null or valid ACK.");
	} else {
	    if (DPRPClient.DEBUG) {
		System.out.println (getName() + "trL: got a message back!");
		inMessage.printDump();
	    }
	    // Return lease based off of this message.
	    ret = new DPRPLease (inMessage, true);
	}

	// Kill the socket, we're going to sleep for awhile.
	mySock.close();
	mySock = null;
	return (ret);
    } /* End terminateLease(). */

    /***********************************************************************/
    /* Utility methods.                                                    */
    /***********************************************************************/
    /**
     * Keep sending the given message until an appropriate response is
     * received from the server. Should only return <code>null</code>
     * in the event that the server cannot be contacted, or if there
     * is a socket error. Of course, I chould just make up new
     * <b><tt>DPRPErrors</tt></b> for these situations, which is an
     * interesting thought. I could throw UNKNOWN_ERRORS for the time being.
     * <p>
     * The problem with this idea is that in the catch, I'd have to check
     * to see if this is a 'soft error' (i.e., I should retry), or if it
     * is a 'hard error' (i.e., we got an error message). So, I'm going to
     * leave the status quo for now.
     * @param outMessage The outgoing DPRPMessage to be sent.
     * @throws DPRPErrorMessageException If an actual, honest-to-goodness,
     *  error message is received from the server.
     * @return A valid DPRPMessage that was received from the server.
     */
    private DPRPMessage sendAndRecvMessage (DPRPMessage outMessage) throws DPRPErrorMessageException {
	DPRPMessage inMessage = new DPRPMessage();

	try {
	    // Send the outgoing message, foolio!
	    mySock.send (outMessage);

	    if (DPRPClient.DEBUG) {
		System.out.println (getName() + "sARM: outMessage sent. " +
				    "Contents:");
		outMessage.printDump();
	    }

	    boolean waitingToExhale = true;    // Main loop control.
	    boolean tmp;                       // Receive status.
	    int tc = 0;                        // Retry counter.
	    while ((waitingToExhale) && (tc < NUM_RETRIES)) {
		tmp = mySock.receive (inMessage);
		// This is FUCKED up! Without the curly's after this if
		// (even though they're not needed), the program won't
		// execute the true or false part of the conditional
		// if tmp is false. Gay!
		if (tmp) {
		    if (outMessage.getXid() == inMessage.getXid()) {
			if (DPRPClient.DEBUG)
			    System.out.println (getName() + "sAMR: Received " +
					      "valid message, breaking loop.");
			waitingToExhale = false;
		    } else if (inMessage.getMessageType()!=DPRPMessage.ERROR) {
			if (DPRPClient.DEBUG) {
			    System.out.println ("Valid message received, but "
						+ "xid's do not match.");
			    System.out.println ("Us: " + outMessage.getXid() +
					      "\tThem: " + inMessage.getXid());
			    inMessage.printDump();
			}

			// Notify the server of this problem, and then notify
			// the user via an exception.
			mySock.sendError (outMessage, DPRPErrors.INVALID_XID_SENT);
			throw new DPRPErrorMessageException (DPRPErrors.INVALID_XID_RECEIVED);
		    } /* else if error */
		} else {
		    // Receive returned false.
		    tc++;
		    if (DPRPClient.DEBUG)
			System.out.println ("Timed-out waiting for response " +
			    "on " + mySock.getLocalAddress().getHostAddress() +
			    ":" + mySock.getLocalPort() + 
					    "; Re-sending #" + tc + "...");
		    mySock.send (outMessage);
		}
	    } // while

	    // If we broke the loop because we sent 25 packets...
	    if ((tc == NUM_RETRIES) && (waitingToExhale)) {
		if (DPRPClient.DEBUG)
		    System.out.println ("Sent " + NUM_RETRIES +
					" packets without getting " +
					"a valid response; giving up.");
		return (null);
	    }
	} catch (SocketException e) {
	    if (DPRPClient.DEBUG)
		System.err.println ("** sARM-Error: problem sending " +
				    "message:\n\t" + e);
	    return (null);
	} catch (IOException e) {
	    if (DPRPClient.DEBUG)
		System.err.println ("** sARM-Error: problem sending " +
				    "message:\n\t" + e);
	}

	/*
	 * Okay, so all error-handling is here, obscured from the caller.
	 */
	if (inMessage.getMessageType() == DPRPMessage.ERROR) {
	    if (DPRPClient.DEBUG) {
		System.out.println ("Got an error packet from the server.");
		System.out.println (DPRPErrors.resolveError (inMessage.getFlags()));
	    }

	    // Report error to user via an exception. How novel!
	    throw new DPRPErrorMessageException (inMessage);
	    }

	return (inMessage);
    } /* End sendAndRecvMessage(). */

    /**
     * Simply creates a new <b><tt>DPRPMessage</tt></b>, and fills it
     * in with the parameters that we know.
     * @return A fully-formed <b><tt>DPRPMessage</tt></b>.
     */
    private DPRPMessage createNewMessage() {
	DPRPMessage ret = new DPRPMessage(serverIP, serverPort);

	ret.setSourceAddress (myIP);
	ret.setSourcePort (myPort);
	ret.setExternalPort (requestedPort);
	ret.setDuration (reqDuration);
	ret.setFlags (reqFlags);
	ret.setProtocol (reqProtocol);

	return (ret);
    } /* End createNewMessage(). */

    /**
     * Allocates a new socket. This code was broken out, because a new
     * socket is now allocated for every renew request. The problem
     * with this code, is that it not only returns a new socket, but
     * it also damages the local <b><tt>myPort</tt></b>. That's frugli.
     * The way to do this would be to return void, and just go ahead and
     * damage both local variables.
     */
    private void allocNewMySock() {
	try {
	    // Establish a new socket, listening on any free UDP port.
	    mySock = new DPRPSocket (0, myIP);
	    myPort = mySock.getLocalPort();
	} catch (SocketException e) {
	    System.err.println ("Socket bind error. Local IP Address or " +
				"port invalid.");
	    System.err.println ("Exception contents: " + e);
	    System.exit(1);
	} catch (SecurityException e) {}

	// Set the name of this thread. It has to be here, because we're always
	// getting a new socket, and the name is dependant on that, buddy!
	String srvName = null;
	if (serverIP == null)
	    srvName = "null";
	else
	    srvName = serverIP.getHostAddress();
	String curName = "thr_" + myPort + "->" + srvName + ":" +
	    serverPort + ": ";
	setName (curName);

	// Print a nice debugging message before we exit.
	if (DPRPClient.DEBUG) {
	    System.out.println (getName() + "gNPS: Got new socket:" +
				"\n\tlocal IP = " +
				mySock.getLocalAddress().getHostAddress() +
				"\tlocal Port = " +
				mySock.getLocalPort());
	} /* debug */
    } /* End getNewDPRPSocket(). */

} /* End class DPRPClientSession. */
