package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This abstract class represents all of the possible error messages
 * that can be transmitted during the course of DPRP
 * operations. Essentailly, this boils down to a bunch of constant
 * variables, and a method that converts a number to a string. Ain't
 * progress wonderful?
 *
 * @version 1.3 - October 26th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
abstract public class DPRPErrors {

    /**
     * When an error happens, and you don't know who to call.
     */
    public static final short UNKNOWN_ERROR = 1;
    /**
     * Server doesn't have any free ports to allocate; Client should
     * retry later.
     */
    public static final short NO_FREE_PORTS = 2;
    /**
     * The client isn't allowed to connect, by rule of the host
     * security configuration file. Contact the administrator.
     */
    public static final short CLIENT_DENIED = 3;
    /**
     * Requested port cannot be allocated, due to limited range
     * allowed by server rule.
     */
    public static final short CLIENT_RANGE_DENIED = 4;
    /**
     * The <b><tt>messageType</tt></b> specified in the packet didn't
     * jive with what was expected.
     */
    public static final short INVALID_PACKET_TYPE = 5;
    /**
     * The <b><tt>XID</tt></b> field in the sent-out message wasn't valid 
     * for the session.
     */
    public static final short INVALID_XID_SENT = 6;
    /**
     * The <b><tt>XID</tt></b> field in the received message wasn't for the
     * session, so the controller must be notified.
     */
    public static final short INVALID_XID_RECEIVED = 7;
    /**
     * The message didn't decode properly -- i.e. the message was
     * totally corrupted.
     */
    public static final short INVALID_PACKET_FMT = 8;
    /**
     * Protocol specified wasn't TCP or UDP.
     */
    public static final short INVALID_PROTOCOL = 9;
    /**
     * Values specified in the packet were found to be invalid
     * (negative ports, and such).
     */
    public static final short INVALID_PACKET_CONTENTS = 10;
    /**
     * Notifies sender that they are having a problem assembling
     * packets for transmission.
     */
    public static final short YOU_SENT_INVALID_PKT_FMT = 11;
    /**
     * This error is rather silly -- it's used internally in the
     * client, in order to pass a NACK server response from one method
     * to another (because we're passing lease objects around, not
     * messages).
     */
    public static final short CLIENT_LEASE_NOT_RENEWED = 12;

    // These are for sending back in 'SRV_RENEW_NACK' packets.
    /**
     * Lease couldn't be renewed, because there was no record of it on
     * the server's end.
     */
    public static final short CLIENT_LEASE_NOT_FOUND = 13;
    /**
     * Client barred from renewing, by configuration rule.
     */
    public static final short CLIENT_NOT_ALLOWED_RENEW = 14;
    /**
     * Client barred from renewing, because the number of lease
     * renewals was exceeded (set by configuration rule).
     */
    public static final short NUM_RENEWS_EXCEEDED = 15;
    /**
     * Duration specified by client was invalid (negative, or goes
     * against configuration rule).
     */
    public static final short INVALID_DURATION_SPECIFIED = 16;

    /**
     * The server had some (unknown to the client) problem in reserving an
     * external port. This error might be retry-able.
     */
    public static final short PORT_ALLOCATION_ERROR = 17;

    /** Client session succesful event responses. */
    public static final short CLI_SUCCESS               = 300;
    /** A lease update was performed without error. */
    public static final short CLI_UPDATE_SUCCESS        = 301;
    /** Acquiring the first lease occured without error. */
    public static final short CLI_ACQUIRE_SUCCESS       = 302;
    /** Client session errorful event responses. */
    public static final short CLI_ERROR                 = 400;
    /** The server NACK'd an update request. */
    public static final short CLI_SERVER_NACK_ERROR     = 404;
    /**
     * A generic error occured whilst trying to acquire the first lease. This
     * means that the server probably didn't respond.
     */
    public static final short CLI_ACQUIRE_ERROR         = 405;

    /** A server error occured whilst trying to perform an update. */
    public static final short CLI_UPDATE_ERROR          = 406;
    /** The t2 timer expired without getting a valid lease renewal. */
    public static final short CLI_T2_EXPIRED            = 407;
    /** The server didn't respond to repeated lease renew requests. */
    public static final short CLI_SRV_NO_RESPOND        = 408;
    /** Client tried to sleep for an invalid time (negative). */
    public static final short CLI_INTERNAL_ERROR        = 409;

    /**
     * This method resolves a specific error into a human-readable
     * string. The returned string will have "Errorcode(#): " prepended
     * to it.
     * @param errorcode A error number, to be resolved.
     * @return A complete human-readable string representation of the error.
     */
    public static String resolveError (short errorcode) {
	return ("Errorcode(" + errorcode + "): " +
		getErrorStr (errorcode));
    }

    /**
     * Gets the raw string representation of the error, without any junk
     * postfixed to it.
     * @param errorcode A error number, to be resolved.
     * @return The raw human-readable string representation of the error.
     */
    public static String getErrorStr (short errorcode) {
	switch (errorcode) {
	case UNKNOWN_ERROR:
	    return ("Unknown Error.");
	case NO_FREE_PORTS:
	    return ("No free ports available.");
	case CLIENT_DENIED:
	    return ("Client IP address not allowed to request lease.");
	case CLIENT_RANGE_DENIED:
	    return ("Client IP address not allowed to request " +
		    "port in range.");
	case INVALID_PACKET_TYPE:
	    return ("Invalid packet type received.");
	case INVALID_XID_SENT:
	    return ("Invalid XID field sent.");
	case INVALID_XID_RECEIVED:
	    return ("Invalid XID field received.");
	case INVALID_PACKET_FMT:
	    return ("Invalid packet format received.");
	case INVALID_PROTOCOL:
	    return ("Invalid IP protocol specified.");
	case INVALID_PACKET_CONTENTS:
	    return ("Invalid values specified in packet.");
	case YOU_SENT_INVALID_PKT_FMT:
	    return ("You sent a packet that did not deserialize properly.");
	case CLIENT_LEASE_NOT_RENEWED:
	    return ("Got NACK packet back from server.");
	case CLIENT_LEASE_NOT_FOUND:
	    return ("Client lease wasn't found in Server's db.");
	case CLIENT_NOT_ALLOWED_RENEW:
	    return ("Client not allowed to renew lease by Server rule.");
	case NUM_RENEWS_EXCEEDED:
	    return ("Client attempted to renew lease too many times.");
	case INVALID_DURATION_SPECIFIED:
	    return ("A duration that was too long or too short was " +
		    "specified in renew request.");
	case PORT_ALLOCATION_ERROR:
	    return ("Due to some problem, the server was unable to " +
		    "allocate an external port. Please contact the site " +
		    "administrator.");
	case CLI_ACQUIRE_ERROR:
	    return ("A server error occured while attempting\n" +
		    "to acquire the initial lease. Verify that the server\n" +
		    "is running, and that the parameters are correct.");
	case CLI_SERVER_NACK_ERROR:
	    return ("The server NACK'd this update request.");
	case CLI_UPDATE_ERROR:
	    return ("A server error occured during the update operation.");
	case CLI_T2_EXPIRED:
	    return  ("The internal timer expired before a lease renewal\n" +
		     "was successfully completed; discarding lease.");
	case CLI_SRV_NO_RESPOND:
	    return ("The server did not respond to repeated renew\n" +
		    "requests; discarding lease.");
	case CLI_INTERNAL_ERROR:
	    return ("An internal error occured, most likely because the\n" +
		    "specified lease duration overflowed the integer limit.");
	default:
	    return ("Errorcode not defined.");
	} /* switch */
    } /* End resolveError(). */
} /* End class DPRPErrors. */
