package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.net.*;

/**
 * A generic DPRP exception. Generated when I'm too lazy to go and
 * create a specific exception.
 *
 * @version 1.0 - September 19th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPException extends Exception {
    /**
     * Totally default constructor.
     */
    public DPRPException() {
	super();
    }

    /**
     * Make a new exception, with a given human-readable error string.
     */
    public DPRPException (String s) {
	super (s);
    }
} /* End DPRPException. */

/**
 * Thrown when an invalid protocol is specified. This means, that it
 * wasn't either the string "TCP", "UDP", or their "/etc/services"
 * numeric representations.
 *
 * @version 1.0 - September 19th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPInvalidProtocolException extends DPRPException {
    /**
     * Default constructor.
     */
    public DPRPInvalidProtocolException() {
	super();
    }

    /**
     * Make a new exception, with a given human-readable error string.
     */
    public DPRPInvalidProtocolException (String s) {
	super (s);
    }
} /* End DPRPInvalidProtocolException. */

/**
 * Generated when a parse error occurs in the host security
 * configuration file.
 *
 * @version 1.1 - September 19th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPConfigException extends DPRPException {
    /**
     * The original error-string that was passed in.
     */
    private String origStr = null;

    /**
     * Default constructor.
     */
    public DPRPConfigException() {
	super();
    }

    /**
     * Make a new exception, with a given human-readable error string.
     */
    public DPRPConfigException (String s) {
	super (s);
	origStr = s;
    }
    /**
     * Get just the error string, not all of the junk that preceeds it.
     */
    public String getErrorMsg() {
	return (origStr);
    }
} /* End DPRPConfigException. */

/**
 * Generated when the client receives an error message from the server, or
 * if an invalid <code>XID</code> is detected.
 *
 * @version 1.0 - October 8th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPErrorMessageException extends DPRPException {
    /** The errorcode sent from the server. */
    private short errorcode;
    /** A generic error string, for debug-printing. */
    private static final String genErr = "** Error message received; Reason: ";

    /**
     * Constructor from a <b><tt>DPRPMessage</tt></b>. For the time being,
     * all we extract is the flags, which is working quite nicely.
     */
    public DPRPErrorMessageException (DPRPMessage im) {
	super (genErr + DPRPErrors.resolveError(im.getFlags()));
	errorcode = im.getFlags();
    }

    /**
     * Constructor from an errorcode, used when directly making up this
     * exception for internal use.
     */
    public DPRPErrorMessageException (short err) {
	super (genErr + DPRPErrors.resolveError(err));
	errorcode = err;
    }

    /** Returns the current errorcode. */
    public short getErrorcode() {
	return (errorcode);
    }
} /* End class DPRPErrorMessageException. */

// this allows me to get rid of the gay-ass nackReason field from the lease.
// class DPRPServerNACKException extends DPRPException {
//     short nackReason;
// }

/**
 * This exception is used to notify the server that a problem occured in
 * reserving the external port on the gateway. In the current incarnation,
 * this means that something went wrong exec'ing the <code>iptables</code>
 * command for netfilter.
 *
 * @version 1.0 - October 30th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPAllocatePortException extends DPRPException {
    /** The DPRPError -- i.e., the reason behind this exception. */
    private short errorcode;
    /**
     * Any optional text, included so that I didn't have to go making a
     * whole bunch of new DPRPErrors.
     */
    private String optionalText;
    /** Flag that denotes if extra text is available. */
    private boolean hasText;

    /** 
     * Constructor for an exception with just an errorcode.
     * @param ec The errorcode to attach to this exception.
     */
    public DPRPAllocatePortException (short ec) {
	super (DPRPErrors.resolveError (ec));
	hasText = false;
	optionalText = null;
	errorcode = ec;
    }

    /**
     * Constructor for an exception with an errorcode and optional text.
     * @param ec The errorcode to attach to this exception.
     * @param extraText The extra descriptive bits to attach.
     */
    public DPRPAllocatePortException (short ec, String extraText) {
	super (DPRPErrors.resolveError (ec));
	hasText = true;
	optionalText = extraText;
	errorcode = ec;
    }

    /** Returns true if extra text has been attached. */
    public boolean hasExtraText() {
	return (hasText);
    }
    /** Gets the extra text. */
    public String getExtraText() {
	return (optionalText);
    }
    /** Gets the errorcode. */
    public short getErrorcode() {
	return (errorcode);
    }
} /* End class DPRPAllocatePortException. */

/**
 * This exception is only thrown by the <b><tt>handleNewRequest()</tt></b>
 * method, and it serves to put all of the port-cleanup code in one place.
 *
 * @version 1.0 - October 30th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPOfferException extends DPRPException {
    /** The client's IP address. */
    private InetAddress cliIP;
    /** The client's port, so that an error-message can be generated. */
    private int cliPort;
    /** The error message associated with this exception. */
    private String errorText;
    /** The reserved port. */
    private int offeredPort;
    /** The protocol type of the reserved port. */
    private byte offeredProtocol;

    /**
     * Constructs a new offer exception.
     * @param ci Client's IP address.
     * @param cp Client's session port.
     * @param op The port that we have offered (and in all likelyhood, are
     *  now recending).
     * @param opr The protocol on the offered port.
     * @param et The error-text that accompanies this exception.
     */
    public DPRPOfferException (InetAddress ci, int cp, int op, byte opr,
			       String et) {
	super ("Client " + ci.getHostAddress() + ":" + cp +
	       " errored thusly: " + et + "\n\tUnallocating port: " + op);
	cliIP = ci;
	cliPort = cp;
	errorText = et;
	offeredPort = op;
	offeredProtocol = opr;
    }

    /** Returns the port that was offered, so that it can be cleaned up. */
    public int getOfferedPort() {
	return (offeredPort);
    }

    /** Gets the client's IP address. */
    public InetAddress getClientIP() {
	return (cliIP);
    }

    /** Returns the protocol of the reservation. */
    public byte getProtocol() {
	return (offeredProtocol);
    }
} /* End class DPRPOfferException. */
