package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;
import java.text.*;      // For the DateFormat class.

/**
 * This class represents all of the details that comprise a
 * "lease". In the DPRP sense, this means that we have several private
 * variables, comprising the external IP address and port on the
 * gateway, as well as duration, protocol, flags, and the starting time
 * for this lease. This class also contains methods to write out a
 * lease into a machine-parsable string, and of course, to suck in a
 * lease from such a string.
 * 
 * @version 1.3 - October 17th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPLease {
    /** Constant representing the default duration, in seconds. */
    public static final int DEFAULT_DURATION = 1000;
    /** Constant representing the default flags for a new lease. */
    public static final short DEFAULT_FLAGS = 1;

    /** The length of time, in seconds, during which this lease is valid. */
    private int duration;
    /** The wall-clock time at which this lease was created. */
    private Calendar startTime;
    /** The external IP address (on the gateway) for this lease. */
    private InetAddress externalIP;
    /** The external port (on the gateway) for this lease. */
    private int externalPort;
    /** The IP address of the client that registered this lease. */
    private InetAddress clientIP;
    /** The port of the registering client. */
    private int clientPort;
    /**
     * I've been forced to make the lease track the IP address of the
     * server that generated it. This makes the lease valid after it
     * has been saved to disk, and reloaded by the client.
     */
    private InetAddress serverIP;
    /** The server's main listening port; where new connections go. */
    private int serverPort;

    /** The protocol type that the gateway will forward. */
    private byte protocol;
    /** The flags that govern this lease. */
    private short flags;
    /** Extra field to pass server lease NACK errorcodes. */
    private short nackError;
    /** Number of times this lease has been renewed. */
    private int numRenews;

    /**
     * In order to ease JAR file creation, this variable is now public, and
     * set to be true whenever the server or client sets it's debug variables
     * to true.
     */
    public static boolean DEBUG = false;

    /**
     * DPRPLease constructor from DPRPMessage.
     * @param msg A DPRPMessage, containing the parameters to make a
     *  lease out of.
     * @param dst Wether the proposed lease's 'ClientIP' refers to the
     *  destination or the source of the message.
     * @return A fully-formed DPRPLease, based upon the message.
     */
    public DPRPLease (DPRPMessage msg, boolean dst) {
	externalIP = msg.getExternalIP();
	externalPort = msg.getExternalPort();

	// ??
	if (dst) {
	    clientIP = msg.getDestinationAddress();
	    clientPort = msg.getDestinationPort();
	    serverIP = msg.getSourceAddress();
	    serverPort = msg.getSourcePort();
	} else {
	    clientIP = msg.getSourceAddress();
	    clientPort = msg.getSourcePort();
	    serverIP = msg.getDestinationAddress();
	    serverPort = msg.getDestinationPort();
	}

	duration = msg.getDuration();
	protocol = msg.getProtocol();
	flags = msg.getFlags();
	startTime = Calendar.getInstance();
	numRenews = 0;
    } /* End DPRPLease from DPRPMessage constructor. */

    /**
     * Simple constructor for DPRPLease.
     */
    public DPRPLease() {
	startTime = Calendar.getInstance();
	numRenews = 0;
    }

    /**
     * Returns the external IP address of this lease.
     */
    public InetAddress getExternalIP() {
	return (externalIP);
    }
    /**
     * Sets the external IP address of this lease.
     */
    public void setExternalIP (InetAddress a) {
	externalIP = a;
    }
    /**
     * Gets the external port of this lease.
     */
    public int getExternalPort() {
	return (externalPort);
    }
    /**
     * Sets the external port for this lease.
     * @param extP An integer value, representing the port.
     */
    public void setExternalPort (int extP) {
	externalPort = extP;
    }
    /**
     * Sets the external port for this lease.
     * @param extS A string value, to be converted to an integral port
     *  value.
     * @throws NumberFormatException If the string doesn't contain a
     *  valid integer.
     */
    public void setExternalPort (String extS) throws NumberFormatException {
	Integer epI = new Integer (extS);
	externalPort = epI.intValue();
    }

    /**
     * Returns the client IP address that 'owns' this lease.
     */
    public InetAddress getClientIP() {
	return (clientIP);
    }
    /**
     * Sets the client IP address for this lease.
     */
    public void setClientIP (InetAddress a) {
	clientIP = a;
    }

    /**
     * Returns the port of the client that established this lease.
     */
    public int getClientPort() {
	return (clientPort);
    }
    /**
     * Sets the client port for this lease.
     * @param extP An integral port value.
     */
    public void setClientPort (int extPortInt) {
	clientPort = extPortInt;
    }
    /**
     * Sets the client port for this lease.
     * @param extPortStr A string, hopefully containing an integer port value.
     * @throws NumberFormatException If the string doesn't contain a
     *  valid integer value.
     */
    public void setClientPort (String extPortStr) throws NumberFormatException {
	Integer cpI = new Integer (extPortStr);
	clientPort = cpI.intValue();
    }

    /** Get the server's IP address. */
    public InetAddress getServerIP() {
	// Since this was a last-minute addition, I thought that a little
	// sanity check was in order.
	if (serverIP == null)
	    throw new RuntimeException ("Trying to return null server IP!");
	return (serverIP);
    }
    /** Sets the server's IP. */
    public void setServerIP (InetAddress sIP) {
	serverIP = sIP;
    }
    /** Gets the server port. */
    public int getServerPort() {
	// Sanity check.
	if (serverPort == 0)
	    throw new RuntimeException ("Trying to get null server port!");
	return (serverPort);
    }
    /** Sets the port on the server, from an int. */
    public void setServerPort (int xp) {
	serverPort = xp;
    }
    /** Sets the port on the server, from a string. */
    public void setServerPort (String ps) throws NumberFormatException {
	Integer cpI = new Integer (ps);
	serverPort = cpI.intValue();
    }

    /**
     * Returns the duration of this lease, in seconds.
     */
    public int getDuration() {
	return (duration);
    }
    /**
     * Sets the duration of this lease.
     * @param dur An integer number of seconds, to comprise the new
     *  duration.
     */
    public void setDuration (int dur) {
	duration = dur;
    }
    /**
     * Sets the duration of this lease.
     * @param sDur An string, representing an integer number of seconds.
     * @throws NumberFormatException If the string doesn't contain a
     *  valid integer value.
     */
    public void setDuration (String sDur) throws NumberFormatException {
	Integer durI = new Integer (sDur);
	duration = durI.intValue();
    }

    /**
     * Return the current protocol for this lease.
     */
    public byte getProtocol() {
	return (protocol);
    }
    /**
     * Return the current protocol as a human-readable string.
     */
    public String getProtocolStr() {
	return (DPRPMessage.protToString (protocol));
    }
    /**
     * Sets the raw protocol.
     */
    public void setProtocol (byte prot) {
	protocol = prot;
    }
    /**
     * Sets the protocol, given a human-readable string.
     * @throws DPRPInvalidProtocolException If the given string can't
     *  be matched to a known numeric representation.
     */
    public void setProtocol (String sP) throws DPRPInvalidProtocolException {
	protocol = DPRPMessage.stringToProt (sP);
    }

    /**
     * Returns the current value of the flags.
     */
    public short getFlags() {
	return (flags);
    }
    /**
     * Sets the flags for this lease.
     */
    public void setFlags (short f) {
	flags = f;
    }
    /**
     * Set the flags for this lease, given a string representation of a short.
     * @throws NumberFormatException If the string given can't be
     *  convereted to a numeric value.
     */
    public void setFlags (String fStr) throws NumberFormatException {
	Short fs = new Short (fStr);
	flags = fs.shortValue();
    }

    /**
     * Returns the nack reason.
     * @deprecated This will be removed, when I refactor the NACK stuff out
     *  of this class.
     */
    public short getNackError() {
	return (nackError);
    }
    /** Sets the nack reason. */
    public void setNackError (short r) {
	nackError = r;
    }
    /**
     * Updates the <b><tt>startTime</tt></b> for this lease with the current
     * wall-clock time.
     */
    public void touchStartTime() {
	startTime = Calendar.getInstance();
    }
    /**
     * Sets the <b><tt>startTime</tt></b> to the Date object given.
     */
    public void setStartTime (Date t) {
	startTime.setTime (t);
    }

    /**
     * Returns the "primary key" for this lease.
     * This is a database term, which means the tuple that uniquely
     * identifies an object. In this case, it is the conjunction of
     * the external IP address, the external port, and the protocol,
     * separated by colons.
     * @return A string representing the primary key.
     */
    public String getPrimaryKey() {
	if (externalIP == null)
	    return (null);
	return (externalIP.getHostAddress() + ":" + externalPort +
		":" + DPRPMessage.protToString(protocol));
    }

    /**
     * Gets the number of times this lease has been renewed.
     */
    public int getNumRenews() {
	return (numRenews);
    }
    /**
     * Increments the number of times this lease has been renewed.
     */
    public void incNumRenews() {
	numRenews++;
    }
    /**
     * Sets the number of times this lease has been renewed.
     * @param i The number of renews are set equal to this value.
     */
    public void setNumRenews (int i) {
	numRenews = i;
    }
    /**
     * Sets the number of times this lease has been renewed.
     * @param s A string, representing an integer number.
     * @throws NumberFormatException If the string doesn't accurately
     *  represent an integer.
     */
    public void setNumRenews (String s) throws NumberFormatException {
	Integer intObj = new Integer (s);
	numRenews = intObj.intValue();
    }

    /**
     * This method checks to see if a lease has expired.
     * This check is performed by adding the duration to the
     * <tt>startTime</tt>, and checking this value against the current
     * "wall clock" time.
     * @return True if the lease has expired, false otherwise.
     */
    public boolean isExpired() {
	Date now = new Date();
	if (now.after (getStopTime()))
	    return (true);
	else
	    return (false);
    } /* End isExpired(). */

    /**
     * Determines when this lease expires.
     * @return A <b><tt>Date</tt></b> instance, containing the exact time that
     *  this lease expires.
     */
    public Date getStopTime() {
	Date stopTime = startTime.getTime();
	stopTime.setTime (stopTime.getTime() + duration * 1000);
	return (stopTime);
    } /* End getStopTime(). */

    /**
     * Spits out a fully-formed <b><tt>DPRPMessage</tt></b>, addressed
     * to the client.
     * @param sourceIP The source IP address (server).
     * @param sourcePort The source IP port (server).
     * @return A fully-formed <b><tt>DPRPMessage</tt></b> based on
     *  this lease.
     */
    public DPRPMessage messageToClient (InetAddress sourceIP, int sourcePort) {
	DPRPMessage ret = new DPRPMessage (clientIP, clientPort);
	ret.setSourceAddress (sourceIP);
	ret.setSourcePort (sourcePort);
	ret.setExternalIP (externalIP);
	ret.setExternalPort (externalPort);
	ret.setProtocol (protocol);
	ret.setDuration (duration);
	ret.setFlags (flags);

	return (ret);
    } /* End messageToClient(). */

    /**
     * Compares the parameters in the given message to those in this lease.
     * @param msg A <b><tt>DPRPMessage</tt></b>, of which values certain
     *  fields will be compared to those in this lease.
     * @return True if all values are identical, false otherwise.
     */
    public boolean compareToMsg (DPRPMessage msg) {
	if (msg.getDuration() != duration)
	    return (false);
	if (msg.getExternalPort() != externalPort)
	    return (false);
	if (!externalIP.equals (msg.getExternalIP()))
	    return (false);
	if (msg.getProtocol() != protocol)
	    return (false);

	// Else, return true.
	return (true);
    } /* End compareToMsg(). */

    /**
     * Overloading the exact <tt>equals()</tt> method that appears in
     * <b><tt>Object</tt></b>.
     * @param obj An object, to compare our values against, if it is in
     *  fact a <b><tt>DPRPLease</tt></b> instance.
     * @return Returns the result of the localized <tt>equals()</tt> method.
     */
    public boolean equals (Object obj) {
	if (obj.getClass() == this.getClass())
	    return (this.equals ((DPRPLease)obj));
	else
	    return (false);
    } /* End equals(). */

    /**
     * Localized method for determining the equality of two leases. I am
     * being fairly loose about this -- if the two "primary keys" are the
     * same, then the leases are said to be equivalent. They refer to the
     * same external IP, port, and protocol, at least. They can have
     * different start times, duration, flags, etc., however.
     * @param cmp A <b><tt>DPRPLease</tt></b>, to compare our values against.
     * @return True if the leases are equal, false otherwise.
     */
    public boolean equals (DPRPLease cmp) {
	if (!externalIP.equals (cmp.getExternalIP()))
	    return (false);
	if (externalPort != cmp.getExternalPort())
	    return (false);
	if (protocol != cmp.getProtocol())
	    return (false);
	if  (!serverIP.equals (cmp.getServerIP()))
	    return (false);

	// If we get here, then they are equal.
	return (true);
    } /* End equals(). */

    /**
     * Returns an integer represenation of this lease, for use in hashing.
     * Since I overloaded <code>equals()</code>, I also have to overload
     * this method. Essentially, I just take the bits of the primary key,
     * and smash them all together into a 32-bit quantity.
     */
    public int hashCode() {
	// Ug. Hash something together, eh?
	int hc = externalPort;
	byte[] extIPbytes = externalIP.getAddress();
	hc = (hc << 8) + extIPbytes[3];
	if (protocol == DPRPMessage.TCP_PROTOCOL)
	    hc *= -1;

	return (hc);
    } /* End hashCode(). */

    /**
     * Writes out a lease in a computer-parseable format. This format is
     * suitable for an on-disk database of leases. Here is an example of the
     * format:
     * <pre>
     *   extIP:extPort:protocol|cliIP:cliPort|srvIP:srvPort|startTime|duration|flags|numRenews
     * </pre>
     * Where <tt>extIP:extPort:protocol</tt> is the primary key.
     *
     * @return A string, uniquely representing the lease.
     */
    public String toString() {
	String ret = externalIP.getHostAddress() + ":" + externalPort + ":";
	ret += DPRPMessage.protToString(protocol) + "|";
	ret += clientIP.getHostAddress() + ":" + clientPort + "|";
	ret += serverIP.getHostAddress() + ":" + serverPort + "|";

	// Print the non-bastardized form of the time.
	DateFormat df = DateFormat.getDateTimeInstance (DateFormat.FULL,
							DateFormat.FULL);
	ret += df.format(startTime.getTime()) + "|";

	// Finish it off with the duration & flags & numRenews.
	ret += duration + "|" + flags + "|" + numRenews;
	
	return (ret);
    } /* End toString(). */

    /**
     * Converts a string representation of a lease into a
     * <b><tt>DPRPLease</tt></b> object.
     * @param str A String representing a lease, to be converted.
     * @return A <b><tt>DPRPLease</tt></b> object, with the fields
     *  matching those of the string.
     */
    public static DPRPLease fromString (String str) {
	DPRPLease retLea = new DPRPLease();

	// Parse the initial string up into various vectors.
	// @0 - primary key
	// @1 - client ip & port
	// @2 - server ip & port
	// @3 - start time
	// @4 - duration
	// @5 - flags
	// @6 - numRenews
	Vector mainVec = parseLeaseString (str, '|');
 	if (mainVec.size() != 7) {
	    if (DEBUG) {
		System.out.println ("** Lease.fS-ERROR: Didn't get correct " +
				    "number of elements from parseLeaseString()! Got:");
		System.out.println ("   - size mainVec = " + mainVec.size());
	    }
	    return (null);
	}

	Vector priKey = parseLeaseString ((String)mainVec.elementAt (0), ':');
	Vector cliNfo = parseLeaseString ((String)mainVec.elementAt (1), ':');
	Vector srvNfo = parseLeaseString ((String)mainVec.elementAt (2), ':');

	// Error-check what we've supposedly parsed.
	if ((priKey.size() != 3) || (cliNfo.size() != 2) || (srvNfo.size() != 2)) {
	    if (DEBUG) {
		System.out.println ("** Lease.fS-ERROR: Didn't get correct " +
				    "number of elements from parseLeaseString()! Got:");
		System.out.println ("   - size priKey = " + priKey.size());
		System.out.println ("   - size cliNfo = " + cliNfo.size());
		System.out.println ("   - size srvNfo = " + srvNfo.size());
	    }
	    return (null);
	}

	try {
	    // Get the information about the port reserved.
	    retLea.setExternalIP (InetAddress.getByName ((String)priKey.elementAt (0)));
	    retLea.setExternalPort ((String)priKey.elementAt (1));
	    retLea.setProtocol ((String)priKey.elementAt (2));

	    // Get the client information.
	    retLea.setClientIP (InetAddress.getByName ((String)cliNfo.elementAt (0)));
	    retLea.setClientPort ((String)cliNfo.elementAt (1));

	    // Get the server information.
	    retLea.setServerIP (InetAddress.getByName ((String)srvNfo.elementAt (0)));
	    retLea.setServerPort ((String)srvNfo.elementAt (1));

	    // Get the startTime.
	    DateFormat df = DateFormat.getDateTimeInstance (DateFormat.FULL, DateFormat.FULL);
 	    Date tmpDate = df.parse ((String)mainVec.elementAt (3));
	    retLea.setStartTime (tmpDate);

	    // Get the duration.
	    retLea.setDuration ((String)mainVec.elementAt (4));

	    // Get the flags.
	    retLea.setFlags ((String)mainVec.elementAt (5));

	    // Get the numRenews.
	    retLea.setNumRenews ((String)mainVec.elementAt (6));
	} catch (UnknownHostException e) {
	    if (DEBUG)
		System.out.println (e);
	    return (null);
	} catch (ParseException e) {
	    // from date/time parsing.
	    if (DEBUG)
		System.out.println (e);
	    return (null);
	} catch (DPRPInvalidProtocolException e) {
	    if (DEBUG)
		System.out.println (e);
	    return (null);
	} catch (NumberFormatException e) {
	    if (DEBUG)
		System.out.println (e);
	    return (null);
	}

	return (retLea);
    } /* End fromString(). */

    /**
     * An internal method that breaks a string into components around
     * the separator. I wrote this before I knew about
     * <b><tt>java.util.StringTokenizer</tt></b>.
     * @param str A String object to parse.
     * @param separator A character that separates out the fields.
     * @return A vector of each substring in between the separators.
     */
    private static Vector parseLeaseString (String str, char separator) {
	    int curIdx = 0;
	    int prvIdx = 0;
       	    Vector retVec = new Vector();

	    while (curIdx != -1) {
		curIdx = str.indexOf (separator, prvIdx);
		if (curIdx != -1) {
		    retVec.add (str.substring (prvIdx, curIdx));
		    if (DEBUG)
			System.out.println ("Parsed (" + prvIdx + "->" +
					    curIdx + "): '" +
					    retVec.lastElement() + "'.");
		    prvIdx = curIdx + 1;
		}
	    } /* while */

	    // pickup the last field at the end.
	    if (prvIdx < str.length()) {
		retVec.add (str.substring (prvIdx));
		if (DEBUG)
		    System.out.println ("Parsed (" + prvIdx + "->" +
					curIdx + "): '" +
					retVec.lastElement() + "'.");
	    }

	    return (retVec);
    } /* End parseString(). */

    /**
     * Prints a lease to stdout in a human-readable format.
     */
    public void prettyPrint() {
	System.out.println ("=============== Begin Lease Object ===============");
	System.out.println ("  - startTime    = " + startTime.getTime());
	if (externalIP != null)
	    System.out.println ("  - externalIP   = " +
				externalIP.getHostAddress());
	else
	    System.out.println ("  - externalIP   = null");
	System.out.println ("  - externalPort = " + externalPort);
	if (clientIP != null)
	    System.out.println ("  - clientIP     = " +
				clientIP.getHostAddress());
	else
	    System.out.println ("  - clientIP     = null");
	System.out.println ("  - clientPort   = " + clientPort);
	if (serverIP != null)
	    System.out.println ("  - serverIP     = " +
				serverIP.getHostAddress());
	else
	    System.out.println ("  - serverIP     = null");
	System.out.println ("  - serverPort   = " + serverPort);
	System.out.println ("  - protocol     = " +
			    DPRPMessage.protToString (protocol));
	System.out.println ("  - duration     = " + duration);
	System.out.println ("  - flags        = " + flags);
	System.out.println ("  - nackError    = " +
			    DPRPErrors.resolveError (nackError));
	System.out.println ("  - numRenews    = " + numRenews);
	System.out.println ("================ End Lease Object ================");
    } /* End prettyPrint(). */

    /**
     * Reads ASCII lease representations from disk, and returns them in
     * a collection.
     * @param fname The name of the file to be read.
     * @return A collection of leases, as found on in the file.
     */
    public static HashSet readLeaseFile (File fname) throws IOException {
	DPRPLease tmpLease = null;
	HashSet retSet = new HashSet();

	FileReader inR = new FileReader (fname);
	LineNumberReader inLNR = new LineNumberReader (inR);
	String curLine = inLNR.readLine();
 
	while (curLine != null) {
	    if (DEBUG)
		System.out.println ("rLF: lease read: " + curLine);
	    
	    tmpLease = DPRPLease.fromString (curLine);
	    if (tmpLease == null) {
		// some sort of error transpired.
		// It could just be a comment, or something. We'll just try
		// and get the most valid leases that we can.
		curLine = inLNR.readLine();
		continue;
	    }

	    // check to see that it's not expired.
	    if (tmpLease.isExpired()) {
		if (DEBUG)
		    System.out.println ("rLF: read-in expired lease, not " +
					"adding to db.");
	    } else {
		// else, we're good. Add it.
		retSet.add (tmpLease);
	    }

	    // Advance to the next line.
	    curLine = inLNR.readLine();
	}

	inLNR.close();
	inR.close();
	return (retSet);
    } /* End readLeaseFile(). */

    /**
     * Writes a given collection of leases to disk, and returns the leases
     * that were actually written. Because there is no explicit <i>"lease
     * garbage collector"</i>, leases are expired at file-write time. Thus,
     * no guarantees are made that all of the leases given in the
     * collection are actually written to disk.
     * @param fname A <b><tt>File</tt></b> instance, containing the name
     *  and location of the file to be written.
     * @param leasesToWrite A collection of leases to be written.
     * @return A collection of the leases that actually were written.
     */
    public static HashSet writeLeaseFile (File fname,
					  HashSet leasesToWrite) {
	try {
	    // Check to make sure there are leases to write. If not,
	    // delete the file. This has a problem of not getting
	    // things perfect when expired leases are afoot.
	    if (leasesToWrite.size() == 0) {
		if (DEBUG)
		    System.out.println ("wLF: Didn't find any leases " +
					"to write. Deleting file.");

		//File fi = new File (fname);
		fname.delete();
		return (null);
	    }

	    // Setup the file.
	    writeLeaseFileComment (fname);

	    // Append to the comment written in the above method.
	    FileWriter outW = new FileWriter (fname.toString(), true);
	    BufferedWriter outWR = new BufferedWriter (outW);

	    // Now, loop through all the leases, and spit them out.
	    Iterator it = leasesToWrite.iterator();
	    while (it.hasNext()) {
		DPRPLease lea = (DPRPLease)it.next();

		// If this lease is expired, don't write it, and
		// remove it from the map.
		if (lea.isExpired()) {
		    String key = lea.getPrimaryKey();
		    if (DEBUG)
			System.out.println ("wLF: Found expired lease " +
					    "in map '" + key +
					    "' while writing file. Removing.");
		    // I have to use the iterator-safe way to remove. Otherwise,
		    // bad things will happen to good people.
		    it.remove();
		} else {
		    outWR.write (lea.toString());
		    outWR.newLine();
		}
	    } /* while */
	    outWR.flush();
	    outWR.close();
	} catch (IOException e) {
	    if (DEBUG)
		System.out.println ("** IO error in writeLeaseFile(): " + e);
	    return (null);
	}    

	return (leasesToWrite);
    } /* End writeLeaseFile(). */

    /**
     * Writes out the standard comment block at the top of the on-disk
     * lease database.
     * @throws IOException If there was a problem writing to the file.
     */
    private static void writeLeaseFileComment (File fname) throws IOException {
	FileWriter outW = new FileWriter (fname.toString(), false);
	PrintWriter pw = new PrintWriter (outW);
	pw.println ("#");
	pw.println ("# This file contains a textual representation " +
		    "of all of the leases in");
	pw.println ("# the server's database, in the following format:");
	pw.println ("#");
	pw.println ("#    extIP:extPort:protocol|cliIP:cliPort|srvIP:srvPort|startTime|duration|flags|numRenews");
	pw.println ("#");
	pw.println ("# This file is constantly being over-written by " +
		    "the server, so it's not");
	pw.println ("# recommended that any changes be made here. See " +
		    "the 'README' file ");
	pw.println ("# for details.");
	pw.println ("#");
	pw.flush();
	pw.close();
	outW.close();
    } /* End writeLeaseFileComment(). */
} /* End class DPRPLease. */
