package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.net.*;
import java.util.*;
// Gay. For AbstractTableModel
import javax.swing.table.*;

/**
 * This class represents my custom model for the <i>lease management table</i>.
 * Because I want the data in the table to be dynamic, I had to create a
 * model that understood DPRP. The table itself is actually just a <i>view</i>
 * on this model. Thus, it is important to understand how the model gets its
 * data. I decided to simply make the model a <b><tt>DPRPLeaseListener</tt></b>.
 * This means, that whenever a <b><tt>DPRPClient</tt></b> instance fires a
 * lease event, it will flow both into the main GUI code (for displaying any
 * appropriate dialogs), but also into this code, which will update the
 * table.<p>
 *
 * <b>TODO: column text alignment & fonts.</b>
 *
 * @version 0.9a - October 30th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
class DPRPLeaseTableModel extends AbstractTableModel implements DPRPLeaseListener {
    /** External IP address column index. */
    private static final int COL_EXTERNAL_IP = 0;
    /** Reservation External Port column index. */
    private static final int COL_PORT = 1;
    /** Reservation protocol column index. */
    private static final int COL_PROTOCOL = 2;
    /** The halt time for the reservation, column index. */
    private static final int COL_STOP_TIME = 3;
    /** The column index for the number of renewals. */
    private static final int COL_RENEWALS = 4;
    /** The column index for the flags. */
    private static final int COL_FLAGS = 5;

    /** Human-readable text descriptions for each column. */
    private final String[] columnNames = {"External IP",
					  "Port",
					  "Protocol",
					  "Stop Time",
					  "Renewals",
					  "Flags"};

    /** A list of all of the leases in the table. */
    private Vector leaseData;

    /**
     * The next index in the list to use (I'm sort of using this Vector
     * like an array, so as to make the JTable happy).
     */
    private int nextIndex;

    /** Default constructor. Simply creates a new vector, and sets
     * <code>nextIndex</code> to zero.
     */
    public DPRPLeaseTableModel() {
	leaseData = new Vector();
	nextIndex = 0;
    }

    /** Returns the number of rows in the table. */
    public int getRowCount() {
	return (leaseData.size());
    }
    /** Returns the number of columns in the table. */
    public int getColumnCount() {
	return (columnNames.length);
    }
    /**
     * Gets the name of a specific column. If the specified column is out
     * of bounds, a runtime exception is thrown.
     * @param col The column number to lookup.
     * @return A string representation of the column number (typically 
     *  the header).
     */
    public String getColumnName (int col) {
	return columnNames[col];
    }
    /** Gets the class of the column headers. */
    public Class getColumnClass (int c) {
	return (getValueAt (0, c).getClass());
    }
    /** Gets the specific object behind a specific cell. */
    public Object getValueAt (int row, int column) {
	DPRPLease l = (DPRPLease)leaseData.get (row);
	return (getTableData (l, column));
    }
    /** Fetches the <b><tt>DPRPLease</tt></b> object that backs a given row. */
    public DPRPLease getLeaseAt (int row) {
	return ((DPRPLease)leaseData.get (row));
    }

    /**
     * Returns the data, in the appropriate format, for the given column
     * in the given lease. Overloaded from the default implementation
     * provided by the superclass. Here are the formats, as they stand
     * currently:<p>
     * <blockquote><pre>
     * COL_EXTERNAL_IP -- String
     * COL_PORT        -- Integer
     * COL_PROTOCOL    -- String
     * COL_STOP_TIME   -- String, format varies depending upon date distance.
     * COL_RENEWALS    -- Integer
     * COL_FLAGS       -- HexString
     * </pre></blockquote>
     */
    private Object getTableData (DPRPLease lea, int column) {
	switch (column) {
	case COL_EXTERNAL_IP:
	    InetAddress ia = lea.getExternalIP();
	    if (ia != null) {
		return (ia.getHostAddress());
	    } /* if */
	    break;
	case COL_PORT:
	    return (new Integer (lea.getExternalPort()));
	case COL_PROTOCOL:
	    return (lea.getProtocolStr());
	case COL_STOP_TIME:
	    // This is just going to return a Date object, I may
	    // want to do some additional formatting on this
	    // shit, for brevity-of-display purposes.
	    String ret = "";
	    Calendar now = Calendar.getInstance();
	    Calendar stp = Calendar.getInstance();
	    stp.setTime (lea.getStopTime());
	    if (stp.after (now) && (stp.get (Calendar.DAY_OF_YEAR) != now.get (Calendar.DAY_OF_YEAR))) {
		// next day format. "MO/DA/YR @ ". Calendar is gay, in that
		// months start at 0, not 1.
		if ((stp.get (Calendar.MONTH) + 1) < 10)
		    ret += "0";
		ret += (stp.get (Calendar.MONTH) + 1) + "/";
		if (stp.get (Calendar.DAY_OF_MONTH) < 10)
		    ret += "0";
		ret += stp.get (Calendar.DAY_OF_MONTH) + "/";
		String year = "" + stp.get (Calendar.YEAR);
		ret += year.substring (2) + " @ ";
	    } 
	    
	    // else current day format. "HH:MM [am/pm]". Gay again, because
	    // 12:00 am is represented as 0:00 am.
	    if (stp.get (Calendar.HOUR) == 0)
		ret += 12 + ":";
	    else
		ret += stp.get (Calendar.HOUR) + ":";
	    if (stp.get (Calendar.MINUTE) < 10)
		ret += "0";
	    ret += stp.get (Calendar.MINUTE);
	    if (stp.get (Calendar.AM_PM) == Calendar.PM)
		ret += " pm";
	    else
		ret += " am";

	    return (ret);
	case COL_RENEWALS:
	    return (new Integer (lea.getNumRenews()));
	case COL_FLAGS:
	    return ("0x" + Integer.toHexString ((int)lea.getFlags()));
	default:
	    System.out.println ("** gTD-Error: Unk. column # received.");
	} /* switch */

	// error.
	return (null);
    } /* End getTableData(). */

    /**
     * This method represnts a rather interesting design decision. By
     * making this table model into a <code>DPRPLeaseListener</code>,
     * it can intercept lease events directly. This means that the table
     * will be updated without any direct intervention from the GUI code.
     * The alternative would have been to simply have the
     * <code>receiveLease()</code> handler in the GUI simply call methods
     * on this model. But, that would have required making the model
     * reference class-global, and also, it wouldn't have been anywhere
     * <i>near</i> as slick as this!
     * @param ple A <b><tt>DPRPLeaseEvent</tt></b>, representing some sort
     *  of event that must be noted in the table.
     */
    public void receiveLease (DPRPLeaseEvent ple) {
	// The row that this lease occupies in the table.
	int rowIdx = 0;

	// Filter-out error events.
	if (ple.isError() && (ple.getCommand() != REMOVE_LEASE_EVENT)) {
	    if (DPRPClient.DEBUG)
		System.out.println ("Table-rL: Got an error event:" +
				    "\n\tCommand = " + ple.getCommand() +
				    "\n\tErrorcode = " + ple.getErrorcode());
	    return;
	}

	// A newly refactored switch statement, that should make this
	// command-action handle properly.
	switch (ple.getCommand()) {
	case REMOVE_LEASE_EVENT:
	    rowIdx = leaseData.indexOf (ple.getLease());
	    if (rowIdx != -1) {
		// Remove the lease at this index, update table.
		leaseData.remove (rowIdx);
		fireTableRowsDeleted (rowIdx, rowIdx);

		// Compact the vector, and update the index.
		Vector tmp = new Vector (leaseData);
		leaseData = tmp;
		nextIndex--;
	    } else {
		if (DPRPClient.DEBUG)
		    System.out.println ("** Table-rL: Lease to be removed " +
					"not found in internal vector.");
	    }
	    break;
	case NEW_LEASE_EVENT:
	    if (DPRPClient.DEBUG) {
		System.out.println ("Table-rL: Adding following lease " +
				    "to table at row: " + nextIndex);
		ple.getLease().prettyPrint();
	    }
	    leaseData.add (nextIndex, ple.getLease());
	    fireTableRowsInserted (nextIndex, nextIndex);
	    nextIndex++;
	    break;
	case UPDATE_LEASE_EVENT: 
	    rowIdx = leaseData.indexOf (ple.getLease());
	    if (rowIdx != -1) {
		if (DPRPClient.DEBUG)
		    System.out.println ("Table-rL: updating lease.");
		// Update this lease in the vector, JTable.
		leaseData.setElementAt (ple.getLease(), rowIdx);
		fireTableRowsUpdated (rowIdx, rowIdx);
	    } else {
		if (DPRPClient.DEBUG)
		    System.out.println ("** Table-rL: Lease to be updated " +
					"not found in internal vector.");
	    }
	    break;
	default:
	    if (DPRPClient.DEBUG)
		System.out.println ("** Table-rL: Invalid lease event command"
				    + " received! Command: " +
				    ple.getCommand());
	} /* switch */
    } /* End receiveLease(). */

} /* End class DPRPLeaseTableModel. */

