package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.net.*;
import java.io.*;
import java.util.*;

/**
 * This class represents an individual DPRP Message. Interestingly,
 * this class is actually the cornerstone of this entire piece of
 * software. Not only does it lay down the format of the messages passed
 * in the protocol, but it also includes key code for serializing message
 * objects into byte streams (and also deserializing them back out
 * again). It also overrides the 'equals()' and 'hashcode()' methods from
 * class 'Object' (the only class to do so). Furthermore, this class
 * contains many support methods that make dealing with messages easier
 * for both the client and the server. Finally, this file contains a
 * series of custom exceptions for the DPRPManager.
 *
 * @version 1.1 - October 31st, 2001 
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPMessage extends Object implements Cloneable {

    /************************************************************************/
    /* The main message variables.                                          */
    /************************************************************************/
    /** The message type, defines the intended nature of the message. */
    private byte messageType;
    /** The protocol, either TCP (6) or UDP (17). */
    private byte protocol;
    /** The flags, which can represent control options. */
    private short flags;
    /** The session identifier, ala DHCP. */
    private int xid;
    /** The proposed lease duration, in seconds. */
    private int duration;

    /** The source IP address of the sender. */
    private InetAddress srcAddr;
    /** The source port on the sender. */
    private int srcPort;

    /** The destination IP address of the message. */
    private InetAddress dstAddr;
    /** The destination port for the message. */
    private int dstPort;

    /** The external IP address of the gateway. */
    private InetAddress extAddr;
    /** The external Port reserved by the Gateway. */
    private int extPort;

    /************************************************************************/
    /* Local miscellaneous class variables.                                 */
    /************************************************************************/
    /** The <b><tt>Random</tt></b> object singleton instance. */
    private static Random rando;
    /**
     * An IP address of all zeros, in byte array format. Used as in
     * initializer.
     */
    private static byte[] nullIPAddr = {0, 0, 0, 0};

    /************************************************************************/
    /* DPRP Message Types.                                                  */
    /************************************************************************/
    /**
     * Code for <tt>DPRP_ERROR</tt> Message. This is sent whenever an
     * error occurs, either on the client or server side. A specific
     * error message (as defined in the <b><tt>DPRPErrors</tt></b>
     * class) is then placed in the <tt>flags</tt> variable.
     */
    public static final byte ERROR = 1;

    /**
     * Code for <tt>DPRP_CLI_REQUEST</tt> Message. This is sent solely
     * by the client, in order to request a new lease.
     */
    public static final byte CLI_REQUEST = 2;

    /**
     * Code for <tt>DPRP_SRV_OFFER</tt> Message. This is sent by the
     * server, in order to respond to the client's request with a
     * lease offer.
     */
    public static final byte SRV_OFFER = 3;

    /**
     * Code for <tt>DPRP_CLI_OFFER_NACK</tt> Message. This is sent by
     * the client, in order to refuse the server's offer.
     */
    public static final byte CLI_OFFER_NACK = 4;

    /**
     * Code for <tt>DPRP_SRV_OFFER_NACK_ACK</tt> Message. This is sent
     * by the server, in order to acknowledge that it processed the
     * client's offer refusal. This is necessary, because the in the
     * process of generating an offer, the server reserves certain
     * state, which the client must make sure is released.
     */
    public static final byte SRV_OFFER_NACK_ACK = 5;

    /**
     * Code for <tt>DPRP_OFFER_ACK</tt> Message. This is sent by the
     * client, in order to accept the server's generous offer.
     */
    public static final byte CLI_OFFER_ACK = 6;

    /**
     * Code for <tt>DPRP_SRV_COMMIT_OFFER</tt> Message. This is sent
     * by the server, when it has committed the state. Thus, when the
     * client receives this message, it knows that it is free to start
     * using the external port.
     */
    public static final byte SRV_COMMIT_OFFER = 7;

    /**
     * Code for <tt>DPRP_CLI_RENEW_REQUEST</tt> Message. This is sent
     * by the client in order to attempt to renew an outstanding
     * lease.
     */
    public static final byte CLI_RENEW_REQUEST = 8;

    /**
     * Code for <tt>DPRP_SRV_RENEW_ACK</tt> Message. This is sent by
     * the server, telling the client that it has in fact renewed the
     * lease. The client then updates its lease based upon the
     * parameters contained in this message, and then goes to sleep.
     */
    public static final byte SRV_RENEW_ACK = 9;

    /**
     * Code for <tt>DPRP_SRV_RENEW_NACK</tt> Message. This is sent by
     * the server, when it cannot renew the client's lease. In the
     * flags, the server will place an errorcode (again, from
     * <b><tt>DPRPErrors</tt></b>), which will give the client a
     * reason for the renew refusal.
     */
    public static final byte SRV_RENEW_NACK = 10;

    /**
     * Sent by the client, when the the user wishes to immediately terminate
     * a lease.
     */
    public static final byte CLI_TERMINATE_LEASE = 11;

    /**
     * Sent by the server, responding to a termination request.
     */
    public static final byte SRV_TERMINATE_ACK = 12;

    /************************************************************************/
    /* Set-style accessor methods.                                          */
    /************************************************************************/
    /**
     * The default port that the server listens on.
     */
    public static final int DEFAULT_PORT = 1669;
    /**
     * The numeric representation of the TCP protocol, as found in
     * "/etc/services".
     */
    public static final byte TCP_PROTOCOL = 6;
    /**
     * The numeric representation of the UDP protocol, as found in
     * "/etc/services".
     */
    public static final byte UDP_PROTOCOL = 17;

    /** Return the supported protocols, in text format. */
    public static final String[] supportedProtocols = {"TCP", "UDP"};

    /**
     * Constructor for DPRPMessage.
     * @param inServerName the destination IP address
     * @param inServerPort the destination port
     */
    public DPRPMessage (InetAddress inServerIP, int inServerPort) {
	defaultValues();
	dstAddr = inServerIP;
	dstPort = inServerPort;
    }

    /**
     * Constructor for DPRPMessage.
     */
    public DPRPMessage () {
	defaultValues();
    }

    /**
     * Fills in various variables with default values.
     */
    private void defaultValues() {
	messageType = 0;
	protocol = 0;
	flags = 0;
	xid = getNextXid();
	duration = 0;
	srcAddr = null;
	srcPort = 0;
	dstAddr = null;
	dstPort = 0;
	extAddr = null;
	extPort = 0;
    } /* End defaultValues(). */

    /**
     * Converts a <b><tt>DPRPMessage</tt></b> into a byte stream. This
     * process readies a message for transmission over the network. It
     * is essentially a form of serialization, but it should be
     * confused with the JDK's serialization framework (I'm not using
     * any of that blarney).
     * @return A byte array, containing all of the essentials of this message.
     */
    public byte[] serialize() {
	ByteArrayOutputStream outBStream = new ByteArrayOutputStream ();
	DataOutputStream outStream = new DataOutputStream (outBStream);
	
	try {
	    outStream.writeByte(messageType);
	    outStream.writeByte(protocol);
	    outStream.writeShort(flags);
	    outStream.writeInt(xid);
	    // Finally, poop out the lease duration.
	    outStream.writeInt(duration);

	    // Write out the client IP address.
	    if (srcAddr != null)
		outStream.write(srcAddr.getAddress(), 0, 4);
	    else
		outStream.write(nullIPAddr, 0, 4);
	    outStream.writeInt(srcPort);

	    // Write out the server IP address.
	    if (dstAddr != null)
		outStream.write(dstAddr.getAddress(), 0, 4);
	    else
		outStream.write(nullIPAddr, 0, 4);
	    outStream.writeInt(dstPort);

	    // Write out the external IP address & port of the gateway.
	    if (extAddr != null)
		outStream.write(extAddr.getAddress(), 0, 4);
	    else
		outStream.write(nullIPAddr, 0, 4);
	    outStream.writeInt(extPort);

	} catch (IOException e) {
	    System.err.println(e);
	}  // end catch
	
	// extract the byte array from the Stream
	byte data[] = outBStream.toByteArray ();
	
	return data;
    } /* End serialize(). */

    /**
     * This method converts a byte array representation of a
     * <b><tt>DPRPMessage</tt></b> into an actual
     * <b><tt>DPRPMessage</tt></b> object. This is called in the
     * <b><tt>DPRPSocket</tt></b> class, in order to build a
     * <b><tt>DPRPMessage</tt></b> object as soon as the bytes come
     * off of the wire.
     *
     * <p>
     *
     * This method works by plucking out a series of bytes, and
     * attempting to convert them into a more complex data
     * type. Assuming all of these conversions, these complex objects
     * are stuffed into a <b><tt>DPRPMessage</tt></b>, which is then
     * returned. In case the decoding does not proceed properly, this
     * method returns an generic message of type
     * <code>DPRP_ERROR</code>, with the
     * <code>INVALID_PACKET_FMT</code> errorcode set. Receivers will
     * treat this as any other error, and notify the sender
     * appropriately.
     * @param inBytes An array of bytes, containing the key bits of a
     *  <b><tt>DPRPMessage</tt></b> object.
     * @return A valid <b><tt>DPRPMessage</tt></b> object, based on the
     *  byte array.
     */
    public DPRPMessage deserialize (byte[] inBytes) {
	byte[] tmpExtAddr = new byte[4];
	byte[] tmpSrcAddr = new byte[4];
	byte[] tmpDstAddr = new byte[4];

        ByteArrayInputStream inBStream = new ByteArrayInputStream
            (inBytes, 0, inBytes.length);
        DataInputStream inStream = new DataInputStream (inBStream);

	// Read everything out of the buffer first, and make complex types
	// later.
	try {
	    // Start by reading in the byte-length fields.
	    messageType = inStream.readByte();
	    protocol = inStream.readByte();
	    flags = inStream.readShort();
	    xid = inStream.readInt();
	    duration = inStream.readInt();

	    // Decode the client IP address & port.
	    inStream.readFully (tmpSrcAddr, 0, 4);
	    srcPort = inStream.readInt();

	    // Decode the server IP address & port.
	    inStream.readFully (tmpDstAddr, 0, 4);
	    dstPort = inStream.readInt();

	    // Get the extAddr
	    inStream.readFully (tmpExtAddr, 0, 4);

	    // Decode the port number
	    extPort = inStream.readInt();
	} catch (IOException e) {
	    // throw up an invalid_packet_fmt error here?
	    System.err.println (e);
	    messageType = ERROR;
	    flags = DPRPErrors.INVALID_PACKET_FMT;
	    return this;
	}

	/*
	 * Convert the raw IP addresses into InetAddress objects.
	 */
	try {
	    // Convert the 'extAddr' into an actual InetAddress.
	    if (Arrays.equals (tmpExtAddr, nullIPAddr)) {
		// Reset to default.
		extAddr = null;
		// This is a bad, bad, bad thing to do, if the client is
		// trying to request a port on ANY external IP.
//  		extPort = 0;
	    } else {
		int[] unsIP = fixupSigned (tmpExtAddr);
		extAddr = InetAddress.getByName (IParyToString (unsIP));
	    } /* else */

	    // Convert the 'srcAddr' into an actual InetAddress.
	    if (Arrays.equals (tmpSrcAddr, nullIPAddr)) {
		// Reset to default.
		srcAddr = null;
		srcPort = 0;
	    } else {
		int[] unsIP = fixupSigned (tmpSrcAddr);
		srcAddr = InetAddress.getByName (IParyToString (unsIP));
	    } /* else */

	    // Convert the 'srvCtlAddr' into an actual InetAddress.
	    if (Arrays.equals (tmpDstAddr, nullIPAddr)) {
		// Reset to default.
		dstAddr = null;
		dstPort = 0;
	    } else {
		int[] unsIP = fixupSigned (tmpDstAddr);
		dstAddr = InetAddress.getByName (IParyToString (unsIP));
	    } /* else */
	} catch (UnknownHostException e) {
	    // throw up an invalid_packet_fmt error here?
	    // and the rest. this should all be in one try/catch.
	    messageType = ERROR;
	    flags = DPRPErrors.INVALID_PACKET_FMT;
	    return this;
	}

	return this;
    } /* End deserialize(). */

    /*
     * Eventually, this will seed a random number generator with our
     * local IP address + port, and then return the next random number
     * in the sequence (we'll do lazy initialization for the random
     * number generator).
     *
     * Of course, now that I've figured out how to manipulate the IP
     * address in order to do this, I'm thinking that it's a dumb idea.
     * If we seed off of the IP, then client restarts will reset the
     * seed to the same deterministic sequence, and we could have 
     * "xid overlap" (gasp!). So, maybe I'll just use the default
     * seed (system time)? Here is that code:

	byte[] flah = me.getAddress();
	System.out.println ("we got: " + flah.length + " bytes out.");
	int i = 0, j = 24, k = 0;
	while (i < flah.length)
	    {
		System.out.println ("byte " + i + " is: " + flah[i]);
		k = k + (flah[i]<<j);
		j -= 8;
		i++;
	    }
	System.out.println ("IP addr as an int: " + k);

     */
    /**
     * Returns the next <tt>xid</tt>. This is done by creating a
     * <b><tt>Random</tt></b> singleton, and getting the next integer
     * in the sequence.
     * @return A random number, in sequence.
     */
    private int getNextXid() {
	if (rando == null)
	    rando = new Random();
	return (rando.nextInt());
    } /* End getNextXid(). */

    /************************************************************************/
    /* Set-style accessor methods.                                          */
    /************************************************************************/
    /** Set the message type (constants in this class). */
    public void setMessageType (byte mt) {
	messageType = mt;
    }

    /** Set the IP protocol. */
    public void setProtocol (byte inp) {
	protocol = inp;
    }

    /** Set the destination IP address. */
    public void setDestinationAddress (InetAddress dip) {
	dstAddr = dip;
    }

    /** Set the destination port. */
    public void setDestinationPort (int dp) {
	dstPort = dp;
    }

    /** Set the source IP address. */
    public void setSourceAddress (InetAddress cip) {
	srcAddr = cip;
    }

    /** Set the source port. */
    public void setSourcePort (int cp) {
	srcPort = cp;
    }

    /** Set the external IP address. */
    public void setExternalIP (InetAddress eip) {
	extAddr = eip;
    }

    /** Set the external port for this message. */
    public void setExternalPort (int ep) {
	extPort = ep;
    }

    /** Set the flags for this message. */
    public void setFlags (short ifl) {
	flags = ifl;
    }

    /** Set the duration this message. */
    public void setDuration (int id) {
	duration = id;
    }
    /**
     * Sets the <tt>xid</tt>.
     * @param xxx Moonshine just like mom used to make.
     */
    public void setXid (int xxx) {
	xid = xxx;
    }

    /************************************************************************/
    /* Get-style accessor methods.                                          */
    /************************************************************************/
    /** Get the message type. */
    public byte getMessageType() {
	return (messageType);
    }

    /** Get the protocol, as a byte. */
    public byte getProtocol() {
	return (protocol);
    }

    /** Get the protocol, as a string. */
    public String getProtocolStr() {
	return (protToString (protocol));
    }

    /** Get the source IP address. */
    public InetAddress getSourceAddress() {
	return (srcAddr);
    }

    /** Get the source port. */
    public int getSourcePort() {
	return (srcPort);
    }

    /** Get the destination address. */
    public InetAddress getDestinationAddress() {
	return (dstAddr);
    }

    /** Get the destination port. */
    public int getDestinationPort() {
	return (dstPort);
    }

    /** Get the external IP address. */
    public InetAddress getExternalIP() {
	return (extAddr);
    }

    /** Get the external port. */
    public int getExternalPort() {
	return (extPort);
    }

    /** Get the flags. */
    public short getFlags() {
	return (flags);
    }

    /** Get the duration, in seconds. */
    public int getDuration() {
	return (duration);
    }

    /** Get the session ID. */
    public int getXid() {
	return (xid);
    }

    /************************************************************************/
    /* Utility methods.                                                     */
    /************************************************************************/
    /**
     * Gets the unsigned equivalent of each byte in an array. This
     * accomplished by looping through the array, and adding 256 to
     * each byte that is negative. The resulting values are copied
     * into an equivalent <tt>int</tt> array, and returned.
     * @param byteAry An array of bytes to be de-signed.
     * @return An integer array, of the same length as the input.
     */
    public static int[] fixupSigned (byte[] byteAry)
    {
	int[] ret = new int[byteAry.length];
	int c = 0;

	for (c = 0; c < byteAry.length; c++)
	{
	    if (byteAry[c] < 0)
		ret[c] = 256 + byteAry[c];
	    else
		ret[c] = (int)byteAry[c];
	}

	return ret;
    } /* End fixupSigned(). */

    /**
     * Converts a <tt>byte[4]</tt> array into dotted-quad IP address notation.
     * @param ba A byte array.
     * @return A string, in "x.y.z.q" form.
     */
    public static String IParyToString (byte[] ba)
    {
	if (ba.length != 4)
	    return (null);
	else
	    return ("" + ba[0] + "." + ba[1] + "." + ba[2] + "." + ba[3]);
    } /* End IParyToString(). */

    /**
     * Converts an <tt>int[4]</tt> array into dotted-quad IP address notation.
     * @return A string, in the form "x.y.z.q".
     * @param ba An array of ints (to capture the unsigned byte properties)
     */
    public static String IParyToString (int[] ba)
    {
	if (ba.length != 4)
	    return (null);
	else
	    return ("" + ba[0] + "." + ba[1] + "." + ba[2] + "." + ba[3]);
    } /* End IParyToString(). */

    /**
     * Converts a IP protocol identifier into a human-readable string.
     * @param p A byte, representing the protocol (as found in
     *  '/etc/protocols').
     * @return A 3-letter string, representing the protocol identifier.
     */
    public static String protToString (byte p) {
	switch (p) {
	case TCP_PROTOCOL:
	    return ("tcp");
	case UDP_PROTOCOL:
	    return ("udp");
	default:
	    return ("unk");
	} /* switch */
    } /* End protToString(). */

    /**
     * Converts a protocol string to its numeric representation.
     * @param str A string to be converted.
     * @throws DPRPInvalidProtocolException If we don't know what the
     *  hell this string is supposed to represent.
     * @return The numeric representation, in byte form.
     */
    public static byte stringToProt (String str) throws DPRPInvalidProtocolException {
	if (str.equalsIgnoreCase ("tcp")) {
	    return (TCP_PROTOCOL);
	} else if (str.equalsIgnoreCase ("udp")) {
	    return (UDP_PROTOCOL);
	} else
	    throw new DPRPInvalidProtocolException ("Error: invalid protocol '"
						    + str + "' specified");
    } /* End stringToProt(). */

    /**
     * Returns the primary key for this message:
     * "<tt>externalIPAddr:externalPort:protocol</tt>".
     */
    public String getPrimaryKey() {
	if (extAddr == null)
	    return (null);
	return (extAddr.getHostAddress() + ":" + extPort +
		":" + protToString(protocol));
    } /* End getPrimaryKey(). */

    /**
     * Returns the source key, which is "<tt>sourceIPAddr:sourcePort</tt>".
     */
    public String getSourceKey() {
	if (srcAddr == null)
	    return (null);
	return (srcAddr.getHostAddress() + ":" + srcPort);
    } /* End getSourceKey(). */

    /**
     * Flips the source and destination addresses in place.
     */
    public void flipSendRecv() {
	InetAddress tmpAddr = srcAddr;
	int tmpPort = srcPort;
	srcAddr = dstAddr;
	srcPort = dstPort;
	dstAddr = tmpAddr;
	dstPort = tmpPort;
    } /* End flipSendRecv(). */

    /**
     * Overloaded from class <b><tt>Object</tt></b>, creates a copy of
     * this message. This is done by calling <tt>clone()</tt> from the
     * superclass, and then filling in each field on this new object.
     * @return An identical copy of this object. It would be much
     * cooler if I could overload the return value, and have it return
     * an actual <b><tt>DPRPMessage</tt></b>. There is entirely too
     * much casting in this language.
     */
    protected Object clone() {
	// Apparently, this is hella bad.
	//DPRPMessage ret = new DPRPMessage (dstAddr, dstPort);
	try {
	    DPRPMessage ret = (DPRPMessage)super.clone();
	    ret.setDestinationAddress (dstAddr);
	    ret.setDestinationPort (dstPort);
	    ret.setSourceAddress (srcAddr);
	    ret.setSourcePort (srcPort);
	    ret.setExternalIP (extAddr);
	    ret.setExternalPort (extPort);
	    ret.setProtocol (protocol);
	    ret.setDuration (duration);
	    ret.setFlags (flags);

	    return (ret);
	} catch (CloneNotSupportedException e) {
	    // this should never happen
	    throw new InternalError();
	}

    } /* End clone(). */

    /**
     * Overloaded from Object, once again.
     * The only requirement for this method is that the <tt>int</tt>
     * should be consistent for objects that are unchanging. Right now
     * I just return the <tt>xid</tt>, which is a poor substitute for
     * "the real thing".
     * @return The <tt>xid</tt> of this message.
     */
    public int hashCode() {
	return (xid);
    }

    /**
     * Compares this message to another object for equality. This
     * method solved a very, very, <i>very</i> pernicious bug, in that
     * when Java invokes the equals method in something like a
     * <b><tt>Vector</tt></b>, it first casts the parameter object to
     * class <b><tt>Object</tt></b>. Thus, we need to trap this
     * invocation, and check to see if the <b><tt>Object</tt></b> is
     * really a <b><tt>DPRPMessage</tt></b>. If it is, then we invoke
     * the <i>real</i> equals method. Otherwise, the objects are said
     * to be not equal.
     * @param o A generic object.
     * @return False if the object isn't a
     *  <b><tt>DPRPMessage</tt></b>, the result of the real
     *  <tt>equals()</tt> method otherwise.
     */
    public boolean equals (Object o) {
	if (o.getClass() == this.getClass())
	    return (equals ((DPRPMessage)o));
	else
	    return (false);
    }

    /**
     * Compares two <b><tt>DPRPMessage</tt></b> objects for
     * equality. This entails going through each class variable, and
     * checking for equality. If every field is identical, then the
     * messages are said to be identical.
     * @param cmp The <b><tt>DPRPMessage</tt></b> to compare against.
     * @return True if the messages are field-equal, false otherwise.
     */
    public boolean equals (DPRPMessage cmp) {
	if (xid != cmp.getXid())
	    return (false);

	if ((srcAddr != null) && (cmp.getSourceAddress() != null))
	    if (!srcAddr.equals (cmp.getSourceAddress()))
		return (false);
	if (srcPort != cmp.getSourcePort())
	    return (false);

	if ((dstAddr != null) && (cmp.getDestinationAddress() != null))
	    if (!dstAddr.equals (cmp.getDestinationAddress()))
		return (false);
	if (dstPort != cmp.getDestinationPort())
	    return (false);

	// God. Have to handle the edge case where the external
	// address is null (that throws equals() into a tizzy).
	if ((extAddr != null) && (cmp.getExternalIP() != null))
	    if (!extAddr.equals (cmp.getExternalIP()))
		return (false);
	if (extPort != cmp.getExternalPort())
	    return (false);
	if (protocol != cmp.getProtocol())
	    return (false);
	if (flags != cmp.getFlags())
	    return (false);
	if (duration != cmp.getDuration())
	    return (false);

	// if we get here, then true it is.
	return (true);
    } /* End equals(). */

    /**
     * Converts a messageType to a human-readable string equivalent.
     */
    public String getMessageStr() {
	String[] mt = {"", "ERROR\t", "CLI_REQUEST", "SRV_OFFER\t",
		       "CLI_OFFER_NACK", "SRV_OFFER_NACK_ACK",
		       "CLI_OFFER_ACK", "SRV_COMMIT_OFFER",
		       "CLI_RENEW_REQUEST", "SRV_RENEW_ACK", "SRV_RENEW_NACK"};
	return (mt[messageType]);
    } /* End getMessageStr(). */

    /**
     * Prints a human-readable version of the message to stdout.
     */
    public void printDump()
    {
	String[] mt = {"", "ERROR\t", "CLI_REQUEST", "SRV_OFFER\t",
		       "CLI_OFFER_NACK", "SRV_OFFER_NACK_ACK",
		       "CLI_OFFER_ACK", "SRV_COMMIT_OFFER",
		       "CLI_RENEW_REQUEST", "SRV_RENEW_ACK", "SRV_RENEW_NACK",
		       "CLI_TERMINATE_LEASE", "SRV_TERMINATE_ACK"};

	System.out.println ("==================== PACKET DUMP =============" +
			    "=======");
	System.out.println ("messageType = " + mt[messageType] +
			    "\t|| xid = " + xid);
	System.out.println ("protocol = " + protToString(protocol) +
			    "\t\t\t|| flags = " + flags);
	if (extAddr != null)
	    System.out.println ("extAddr = " +
		       IParyToString (fixupSigned(extAddr.getAddress())) +
				"\t\t|| extPort = " + extPort);
	else
	    System.out.println ("extAddr = null\t\t\t|| extPort = " + extPort);

	if (srcAddr != null)
	    System.out.println ("srcAddr = " +
		       IParyToString (fixupSigned(srcAddr.getAddress())) + 
				"\t\t|| srcPort = " + srcPort);
	else
	    System.out.println ("srcAddr = null\t\t\t|| srcPort = " + srcPort);

	if (dstAddr != null)
	    System.out.println ("dstAddr = " +
		       IParyToString (fixupSigned(dstAddr.getAddress())) +
				"\t\t|| dstPort = " + dstPort);
	else
	    System.out.println ("dstAddr = null\t\t\t|| srvCtlPort = " +
				dstPort);

	System.out.println ("duration = " + duration);

	if (messageType == ERROR) {
	    System.out.println ("****************************************" +
				"*************");
	    System.out.println (DPRPErrors.resolveError (flags));
	    System.out.println ("****************************************" +
				"*************");
	}

	System.out.println ("================ END OF PACKET DUMP =========" +
			    "========");
    } /* End printDump(). */
} /* End class DPRPMessage. */


