package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class comprises a set of restrictions against a potential DPRP
 * client. In the interests of security, there are several operations
 * that the site administrator may wish to constrain for certain
 * clients. These capabilities include:
 * <ul>
 *    <li>The ability to use the DPRP service.</li>
 *    <li>The ability to renew leases.</li>
 *    <li>The number of times a lease is allowed to be renewed.</li>
 *    <li>The range of ports that a client is allowed to request.</li>
 *    <li>The allowed lease durations that the client is allowed to
 *        request.</li>
 * </ul>
 * These restrictions are fed into the server via an ASCII text
 * configuration file, the format of which is described in the
 * <b><tt>ServerPortMgr</tt></b> class.
 *
 * @version 1.0 - September 18th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPRestriction {
    /**
     * The IP address of the client to be restricted.
     */
    InetAddress clientIP;
    /**
     * If the client is even allowed to speak DPRP.
     */
    boolean denied;
    /**
     * If the client is allowed to renew at all.
     */
    boolean renewOK;
    /**
     * Number of times client is allowed to renew lease (to prevent
     * running unauthorized servers).
     */
    int allowedNumRenewals;
    /**
     * The shortest lease time the client is allowed (to prevent DoS).
     */
    int minDuration;
    /**
     * The longest lease time the client is allowed (to prevent
     * running unauthorized servers).
     */
    int maxDuration;
    /**
     * The smallest port the client is allowed to request.
     */
    int minPort;
    /**
     * The highest port the client is allowed to request.
     */
    int maxPort;

    /**
     * Default minimum duration (500).
     */
    public static final int defaultMinDuration = 500;
    /**
     * Default maximum duration (MAX_INT).
     */
    public static final int defaultMaxDuration = Integer.MAX_VALUE;
    /**
     * Default maximum allowed port (65536).
     */
    public static final int defaultMaxPort = 65536;

    /**
     * The main constructor for a new restriction. It gives everything
     * default values.
     * @param cip The client IP address.
     */
    public DPRPRestriction (InetAddress cip) {
	clientIP = cip;
	denied = false;
	renewOK = true;
	allowedNumRenewals = 0;   // go forever, baby.
	minDuration = defaultMinDuration;
	maxDuration = defaultMaxDuration;
	minPort = 1024;
	maxPort = defaultMaxPort;
    } /* End constructor. */

    /**
     * Checks the given duration, to see if it is allowed.
     * @param inDur A duration, checked to see that it is greater than
     *  (or equal to) <b><tt>minDuration</tt></b>, but less than
     *  (or equal to) <b><tt>maxDuration</tt></b>.
     * @return True if the parameter satisfies the conditional, false
     *  otherwise.
     */
    public boolean checkDuration (int inDur) {
	if ((inDur >= minDuration) && (inDur <= maxDuration))
	    return (true);
	else
	    return (false);
    } /* End checkDuration(). */

    /**
     * Checks the given port, to see if it is allowed.
     * @param inPort A port number, checked to see that it is greater than
     *  (or equal to) <b><tt>minPort</tt></b>, but less than
     *  (or equal to) <b><tt>maxPort</tt></b>.
     * @return True if the parameter satisfies the conditional, false
     *  otherwise.
     */
    public boolean checkPort (int inPort) {
	if ((inPort >= minPort) && (inPort <= maxPort) && !denied)
	    return (true);
	else
	    return (false);
    } /* End checkPort(). */

    /**
     * Checks if the number of times a lease has been renewed is allowed.
     * @param nr A number of renews, that should be less than the
     *  number of NumRenewals allowed for this client.
     * @return True if the paramater satisfies the conditional, false
     *  otherwise.
     */
    public boolean checkNumRenews (int nr) {
	if (allowedNumRenewals == 0)
	    return (true);
	else if (nr < allowedNumRenewals)
	    return (true);
	else
	    return (false);
    } /* End checkNumRenews(). */

    /************************************************************************/
    /* Set-style accessors.                                                 */
    /************************************************************************/
    /**
     * Sets the maximum allowed duration.
     */
    public void setMaxDuration (int mad) {
	maxDuration = mad;
    }
    /**
     * Sets the minimum allowed duration.
     */
    public void setMinDuration (int mid) {
	minDuration = mid;
    }
    /**
     * Sets the maximum allowed port.
     */
    public void setMaxPort (int map) {
	maxPort = map;
    }
    /**
     * Sets the minimum allowed port.
     */
    public void setMinPort (int mip) {
	minPort = mip;
    }
    /**
     * Denies the client.
     */
    public void setIsDenied() {
	denied = true;
    }
    /**
     * Stops the client from being able to renewing any leases.
     */
    public void setRenewNotOK() {
	renewOK = false;
    }
    /**
     * Sets the maximum number of lease renewals.
     */
    public void setNumRenewals (int mr) {
	allowedNumRenewals = mr;
    }
    /**
     * Sets the client IP address for this restriction.
     */
    public void setClientIP (InetAddress cr) {
	clientIP = cr;
    }

    /************************************************************************/
    /* Get-style accessors.                                                 */
    /************************************************************************/
    /**
     * Returns <b><tt>maxDuration</tt></b>.
     */
    public int getMaxDuration() {
	return (maxDuration);
    }
    /**
     * Returns <b><tt>minDuration</tt></b>.
     */
    public int getMinDuration() {
	return (minDuration);
    }
    /**
     * Returns <b><tt>maxPort</tt></b>.
     */
    public int getMaxPort() {
	return (maxPort);
    }
    /**
     * Returns <b><tt>minPort</tt></b>.
     */
    public int getMinPort() {
	return (minPort);
    }
    /**
     * Returns true if client is denied from accessing server.
     */
    public boolean isDenied() {
	return (denied);
    }
    /**
     * Returns true if client allowed to renew leases.
     */
    public boolean renewAllowed() {
	return (renewOK);
    }
    /**
     * Returns <b><tt>allowedNumRenewals</tt></b>.
     */
    public int getNumRenewals() {
	return (allowedNumRenewals);
    }
    /**
     * Returns the client IP address.
     */
    public InetAddress getClientIP() {
	return (clientIP);
    }

    /**
     * A utility method, for displaying the restriction contents in a
     * human-readable way.
     */
    public void prettyPrint() {
	System.out.println ("========== Begin Restriction Object ==========");
	if (clientIP != null)
	    System.out.println ("  - clientIP           = " +
				clientIP.getHostAddress());
	else
	    System.out.println ("  - clientIP           = null");
	System.out.println ("  - denied             = " + denied);
	System.out.println ("  - renewOK            = " + renewOK);
	System.out.println ("  - allowedNumRenewals = " + allowedNumRenewals);
	System.out.println ("  - minDuration        = " + minDuration);
	System.out.println ("  - maxDuration        = " + maxDuration);
	System.out.println ("  - minPort            = " + minPort);
	System.out.println ("  - maxPort            = " + maxPort);
	System.out.println ("=========== End Restriction Object ===========");

    } /* End prettyPrint(). */
} /* End class DPRPRestriction. */
