package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class, when instantiated, will spawn in its own thread, and
 * handle a single client request (be it to request a new lease, or
 * renew an old one). It contains a reference to the global <i>Server
 * Port Manager</i>, which was instantiated in the
 * <b><tt>SimpleDPRPServer</tt></b> class. Using this reference, this
 * server session is able to check the priviliges of the client,
 * regarding its request. Furthermore, this session allocates a local
 * UDP socket, for use in it's replies to the client. This separates
 * this transaction from other transactions that are going on
 * concurrently in the server.
 *
 * @version 0.9c - October 27th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class DPRPServerSession extends Thread {
    /**
     * This message reference is set by the server singleton, whenever
     * it receives an errant message from the client, that is really
     * destined for this session. This occurs in error situations --
     * either the client is re-transmitting it's first request to main
     * server port, or it is sending some sort of error to the main
     * server port. Either way, the server will set this variable, and
     * then call <i>Thread.interrupt()</i>. Thus, at certain points in
     * the code (usually in the exception handler for
     * <i>Thread.sleep()</i>), this code will check this variable, and
     * if it is non-null, it will consume the message, and take the
     * appropriate action.
     */
    public DPRPMessage passedMessage = null;
    /**
     * The message that started this session, to be removed from the
     * server singleton's map.
     */
    private DPRPMessage startMessage;
    /**
     * The current <b><tt>XID</tt></b> of this session, for sanity checking.
     */
    private int curXid;
    /**
     * The external IP address of the gateway.
     */
    private InetAddress externalIP;
    /**
     * The internal IP address of the gateway.
     */
    private InetAddress internalIP;
    /**
     * Local socket, allocated to keep this session separate from the rest.
     */
    private DPRPSocket localSock = null;
    /**
     * Reference to the <i>Server Port Manager</i> singleton.
     */
    private ServerPortMgr SPM;

    /**
     * Constant directing the number of times to retry sending a message.
     */
    private static final int NUM_RETRIES = 15;
    /**
     * The amount of time to wait for a message from the client, where
     * we know that they received our final message of the transaction.
     */
    private static final long RETRY_TIMEOUT = 60 * 1000; // 60 seconds.

    /**
     * New session constructor.
     * @param inMessage The message that caused the creation of a new
     *  session.
     * @param threadName The name for this thread, as generated by the
     *  server singleton.
     * @param extIP Our external IP address, as specified in a
     *  parameter to the server.
     */
    public DPRPServerSession (DPRPMessage inMessage, String threadName,
			      InetAddress extIP) {
	super (threadName);
	startMessage = inMessage;
	externalIP = extIP;
	internalIP = inMessage.getDestinationAddress();
	SPM = SimpleDPRPServer.PortManager;
    } /* End DPRPServerSession constructor. */

    /**
     * Run method for DPRPServerSession. This method parses the packet
     * that started this session, and either initiates a new request,
     * or a renew request, depending.
     */
    public void run() {
	final String sessionKey = startMessage.getSourceKey();

	// I want this on a free port, but I really want it to be bound
	// to a specific IP address. I guess you can't have your cake
	// and eat it too, eh?
	// But, it looks like you can if you specify a port of zero (this
	// is hella undocumented).
	try {
	    localSock = new DPRPSocket (0, internalIP);
	} catch (SocketException e) {
	    System.err.println (getName() + "unk: Can't create new socket on "
				+ internalIP.getHostAddress() + ".");
	    System.err.println ("\tReason: " + e);
	    System.exit(1);
	} catch (SecurityException e2) {}

	// Update the thread name w/ our new local port.
	setName (getName() + localSock.getLocalPort() + ": ");

	if (SimpleDPRPServer.DEBUG) {
	    System.out.println (getName() + "R: Started.");
	    System.out.println (getName() + "R: Got the following packet:");
	    startMessage.printDump();
	}

	// This xid had better be the same throughout the session.
	curXid = startMessage.getXid();

	// This packet can either be a CLI_REQUEST or a CLI_RENEW
	if (startMessage.getMessageType() == DPRPMessage.CLI_REQUEST) {
	    try {
		handleNewRequest();
	    } catch (DPRPOfferException e) {
		if (SimpleDPRPServer.DEBUG)
		    System.out.println (getName() + e);

		// unallocate de offered port, eh? Send error?
		if (!SPM.unallocExternalPort (e.getClientIP(),
					      e.getOfferedPort(),
					      e.getProtocol()) &&
		    SimpleDPRPServer.DEBUG)
		    System.out.println (getName() + "R: Huh. " +
					"unallocExternalPort() failed.");

		// send error? we can...
	    } /* catch */
	} else if (startMessage.getMessageType()==DPRPMessage.CLI_RENEW_REQUEST)
	    handleRenewRequest();
	else {
	    DPRPMessage tmp_msg = (DPRPMessage)startMessage.clone();
	    tmp_msg.flipSendRecv();
	    localSock.sendError (tmp_msg, DPRPErrors.INVALID_PACKET_TYPE);
	}

	// all done.
	SimpleDPRPServer.removeSession (sessionKey);
	localSock.close();
    } /* End run(). */

    /* !!!!NOTE!!!!
     * if the protocol wasn't specified, we can either bonk the client w/ an
     * error, or take that to mean "both". is there any legitimate reason
     * why you'd want both? icq, maybe?
     *
     * NOTE2: Maybe this needs to throw an exception, so as to handle
     * the cleanup of allocated iptables ports properly? Otherwise,
     * I'm in for a bitch of code-replication in a whole bunch of
     * if branches.
     */
    /**
     * Handles a client request for a new lease. This entails checking
     * the client's specified parameters (if any) with the <i>Server
     * Port Manager</i>. Furthermore, if the client specified an
     * external port, we must attempt to acquire this port. If it
     * cannot be acquired, than any free port will be offered
     * instead. The offer, once constructed, is sent, and is only
     * commited after the client has properly ACK'd it.
     * @throws DPRPOfferException If a problem occurs after a port has
     *  been allocated. The handler will attempt to cleanup the invalid
     *  state in a sane manner.
     */
    private void handleNewRequest() throws DPRPOfferException {

	// Local variables (now).
	DPRPMessage curInMessage = (DPRPMessage)startMessage.clone();
	DPRPMessage curOutMessage = null;

	// Construct a temporary lease for this packet -- maybe it'll stick...
	DPRPLease tmpLease = new DPRPLease (curInMessage, false);
	tmpLease.setExternalIP (externalIP);

	// See if the client requested a port, and if so, attempt to
	// satisfy said request.
	if ((tmpLease.getExternalPort() != 0) && (SPM.checkPort (tmpLease))) {
	    // We are able to give them the port that they requested,
	    // so do it! We assume that the portsResvd[] array has
	    // been properly touched by the checkPort() method.
	    try {
		SPM.allocLinuxPort (tmpLease.getClientIP(),
					    tmpLease.getExternalPort(),
					    tmpLease.getProtocol());
		tmpLease.setExternalPort (tmpLease.getExternalPort());
		if (SimpleDPRPServer.DEBUG)
		    System.out.println (getName() + "Requested port " +
					tmpLease.getExternalPort() + ":" +
					tmpLease.getProtocolStr() +
					" was allocated to client.");
	    } catch (DPRPAllocatePortException e) {
		// What state needs to be cleaned up? The portsResvd[]
		// array, for sure.
		return;
	    } /* catch */
	} else {
	    // We are either unable to satisfy the requested port, or
	    // no port was requested. So, just give 'em something that
	    // is free.
	    if (SimpleDPRPServer.DEBUG)
		System.out.println (getName() + "Client didn't request a " +
				    "specific port.");

	    try {
		int p;
		p = SPM.allocFreeExternalPort (tmpLease.getClientIP(),
					       tmpLease.getProtocol());
		tmpLease.setExternalPort (p);
		if (SimpleDPRPServer.DEBUG)
		    System.out.println (getName() + "Free port " +
					tmpLease.getExternalPort() + ":" +
					tmpLease.getProtocolStr() +
					" was allocated to client.");
	    } catch (DPRPAllocatePortException e) {
		if (SimpleDPRPServer.DEBUG) {
		    System.out.println (getName() + "Got port " +
					"reservation error:");
		    System.out.println ("\t" + e);
		    if (e.hasExtraText())
			System.out.println ("\tAdditionally, this text " +
					    "was attached:\n\t" +
					    e.getExtraText());
		} /* debug */

		// Now what'n the hell do we do? What do we send instead
		// of an offer? What state needs to be cleaned up? I
		// guess we need to send an error message. Wild.
		curInMessage.flipSendRecv();
		if (e.getErrorcode() == DPRPErrors.NO_FREE_PORTS)
		    localSock.sendError (curInMessage,
					 DPRPErrors.NO_FREE_PORTS);
		else
		    localSock.sendError (curInMessage,
					 DPRPErrors.PORT_ALLOCATION_ERROR);
		return;
	    } /* catch */
	} /* else port allocate */

	// What if the default duration isn't valid for the client (by rule)?
	if (tmpLease.getDuration() == 0) {
	    tmpLease.setDuration (DPRPLease.DEFAULT_DURATION);
	}
	if (!SPM.checkDuration (tmpLease)) {
	    // Requested duration not allowed, override with default.
	    if (SimpleDPRPServer.DEBUG)
		System.out.println (getName() + "requested duration of " +
				    tmpLease.getDuration() + " seconds " +
				    "denied. Trying " +
				    DPRPLease.DEFAULT_DURATION + ".");
	    tmpLease.setDuration (DPRPLease.DEFAULT_DURATION);

	    if (!SPM.checkDuration (tmpLease)) {
		// Default duration not allowed(!), override with allowed max.
		int newDuration = SPM.getMaxAllowedDuration (tmpLease);
		if (SimpleDPRPServer.DEBUG)
		    System.out.println (getName() + "requested duration of " +
					tmpLease.getDuration() + " seconds " +
					"denied. Finalizing on max allowed: " +
					newDuration + ".");
		tmpLease.setDuration (newDuration);
	    } /* if default duration doesn't work. */
	} /* if req duration doesn't work. */

	/*
	 * For right now, I really don't need the flags...
	 */
	if (tmpLease.getFlags() == 0) {
	    tmpLease.setFlags (DPRPLease.DEFAULT_FLAGS);
	}

	/*
	 * Construct the outgoing message.
	 */
	curOutMessage = tmpLease.messageToClient (internalIP,
						  localSock.getLocalPort());

	// This is *very* important!
	curOutMessage.setXid (curXid);
	curOutMessage.setMessageType (DPRPMessage.SRV_OFFER);

	// Send prepared message.
	if (SimpleDPRPServer.DEBUG) {
	    System.out.println (getName() + "Sending offer message: ");
	    curOutMessage.printDump();
	}

	// God, I hate this method *so* much!
	curInMessage = sendAndMaybeRecv (curOutMessage, (byte)0, true);

	/*
	 * If we exhausted the number of retries, then we need to bomb out
	 * of this thread entirely.
	 */
	if (curInMessage == null) {
	    // Let's give up on this client, eh?
	    if (SimpleDPRPServer.DEBUG) {
		System.err.println (getName() + "the client has died.");
		System.err.println ("\tClientInfo: " +
		       curOutMessage.getDestinationAddress().getHostAddress() +
		       ":" + curOutMessage.getDestinationPort());
	    }

	    // *** Cleanup any port that I allocated in iptables?
	    throw new DPRPOfferException(curOutMessage.getDestinationAddress(),
					 curOutMessage.getDestinationPort(),
					 curOutMessage.getExternalPort(),
					 curOutMessage.getProtocol(),
					 "No respose to offer.");
	}

 	if (curInMessage.getMessageType() == DPRPMessage.CLI_OFFER_NACK) {
	    // They nack'd us! Those bastards! All I have to do is release
	    // any port that I've set aside, ACK, and exit the thread.
	    curOutMessage = (DPRPMessage)curInMessage.clone();
	    curOutMessage.flipSendRecv();
	    curOutMessage.setMessageType (DPRPMessage.SRV_OFFER_NACK_ACK);
	    curInMessage = sendAndMaybeRecv (curOutMessage,
					     DPRPMessage.CLI_OFFER_NACK,
					     false);
	    // *** Have to clean-up allocated port.
	    throw new DPRPOfferException(curOutMessage.getDestinationAddress(),
					 curOutMessage.getDestinationPort(),
					 curOutMessage.getExternalPort(),
					 curOutMessage.getProtocol(),
					 "Offer was NACK'd.");
	} else if (curInMessage.getMessageType()!=DPRPMessage.CLI_OFFER_ACK) {
	    // Funk 'dis.
	    localSock.sendError (curOutMessage,
				 DPRPErrors.INVALID_PACKET_TYPE);

	    if (SimpleDPRPServer.DEBUG) {
		System.out.println (getName() + "Didn't get a " +
				    "CLI_OFFER_ACK. Instead, we got: ");
		curInMessage.printDump();
	    }

	    // *** Have to clean-up allocated port.
	    throw new DPRPOfferException (curInMessage.getSourceAddress(),
					  curInMessage.getSourcePort(),
					  curOutMessage.getExternalPort(),
					  curOutMessage.getProtocol(),
					  "Invalid packet type received: "+
					  curInMessage.getMessageType() +
					  "; expected offer ACK or NACK.");
	}

	if (!tmpLease.compareToMsg (curInMessage)) {
	    // They didn't agree to what we sent, bastards!
	    // Maybe we should send a DPRP_SRV_COMMIT_NACK, or something...
	    // It should be an error, because this represents a non-conformant
	    // client.
	    localSock.sendError (curOutMessage,
				 DPRPErrors.INVALID_PACKET_CONTENTS);
	    // *** Have to clean-up allocated port.
	    throw new DPRPOfferException (curInMessage.getSourceAddress(),
					  curInMessage.getSourcePort(),
					  curOutMessage.getExternalPort(),
					  curOutMessage.getProtocol(),
					  "ACK didn't agree with what was " +
					  "offered.");
	}

	/*
	 * Well, if we get here, everything is all good. Commit the damn
	 * lease, and then send the appropriate message.
	 */
	tmpLease.touchStartTime();
	SPM.addReservation (tmpLease, true);
	// wack it out to stable storage?
	// wack it to iptables, or do we have to do that up above???

	/*
	 * Send SRV_COMMIT_OFFER message.
	 */
	curOutMessage = tmpLease.messageToClient (internalIP,
						  localSock.getLocalPort());

	// This is *very* important!
	curOutMessage.setXid (curXid);
	curOutMessage.setMessageType (DPRPMessage.SRV_COMMIT_OFFER);

	// Send prepared message.
	if (SimpleDPRPServer.DEBUG) {
	    System.out.println (getName() + "Sending commit message: ");
	    curOutMessage.printDump();
	}

	// *** What if this method fails to send the ACK?
	curInMessage = sendAndMaybeRecv (curOutMessage,
					 DPRPMessage.CLI_OFFER_ACK, false);
	if (curInMessage != null) {
	    // this should never happen
	    System.out.println ("** ERROR **");
	}
    } /* End handleNewRequest(). */

    /**
     * Takes care of a client renewal request. This entails checking
     * the parameters specified against the <i>Server Port
     * Manager</i>, and either issuing a
     * <b><tt>DPRP_SRV_RENEW_ACK</tt></b> or
     * <b><tt>DPRP_SRV_RENEW_NACK</tt></b>, as appropriate.
     */
    private void handleRenewRequest() {
	// Local variables (now).
	DPRPMessage inMessage = (DPRPMessage)startMessage.clone();
	DPRPMessage outMessage = null;

	// Construct a temporary lease for this packet -- maybe it'll stick...
	DPRPLease tmpLease = SPM.getReservation (inMessage);

	if (tmpLease == null) {
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("hRR: Got null from getReservation(), " +
				    "which means that lease wasn't found. " +
				    "Sending NACK.");
	    // error, no lease found for this message.
	    // send 'em a nack?
	    tmpLease = new DPRPLease (inMessage, false);
	    sendRenewNack (tmpLease, DPRPErrors.CLIENT_LEASE_NOT_FOUND);
	    return;
	}

	// Slam in the new client port. This is pretty lame.
	tmpLease.setClientPort (inMessage.getSourcePort());

	// Slam in the new duration, in case it's different.
	tmpLease.setDuration (inMessage.getDuration());

	// Check to see if the num_retries is kosher.
	if (!SPM.checkRetries (tmpLease)) {
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("hRR: Client has renewed " +
				    tmpLease.getNumRenews() +
				    " times, which exceeds their set limit; " +
				    "Sending NACK.");
	    sendRenewNack (tmpLease, DPRPErrors.NUM_RENEWS_EXCEEDED);
	    return;
	}

	// Check to make sure the requested duration is valid.
	if (!SPM.checkDuration (tmpLease)) {
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("hRR: Client-specified duration " +
				    "isn't right; Sending NACK.");
	    sendRenewNack (tmpLease,DPRPErrors.INVALID_DURATION_SPECIFIED);
	    return;
	}

	// Okay, at this point, we're good to go. Let's generate some
	// acks, shall we?
	tmpLease.incNumRenews();
	tmpLease.touchStartTime();
	SPM.updateReservation (tmpLease);

	outMessage = tmpLease.messageToClient (internalIP,
					       localSock.getLocalPort());
	outMessage.setDestinationAddress (inMessage.getSourceAddress());
	outMessage.setDestinationPort (inMessage.getSourcePort());
	outMessage.setXid (inMessage.getXid());
	outMessage.setMessageType (DPRPMessage.SRV_RENEW_ACK);

	if (!sendRenewResp (outMessage) && SimpleDPRPServer.DEBUG)
	    System.out.println ("*** hRR-ERROR: didn't properly send " +
				"ACK packet. Dying.");

    } /* End handleRenewRequest(). */

    /**
     * Constructs and transmits a valid <b><tt>DPRP_SRV_RENEW_NACK</tt></b>
     * packet. Also, this method cleans up any server state regarding
     * this deleted port, and then sends the NACK to the client.
     *
     * @param lea The DPRPLease to be terminated.
     * @param reason The reason (according to the
     *  <b><tt>DPRPErrors</tt></b> database) that the lease is being
     *  terminated.
     * @see DPRPErrors
     */
    private void sendRenewNack (DPRPLease lea, short reason) {
	// Local variables.
	DPRPMessage outMessage = null;

	// Expire the lease
	SPM.delReservation (lea, true);

	// Prepare outgoing message.
	outMessage = lea.messageToClient (internalIP,
					  localSock.getLocalPort());
	outMessage.setMessageType (DPRPMessage.SRV_RENEW_NACK);
	outMessage.setFlags (reason);
	outMessage.setXid (startMessage.getXid());

	// Send the outgoing message.
	if (!sendRenewResp (outMessage) && SimpleDPRPServer.DEBUG)
	    System.out.println ("*** sRN-ERROR: didn't properly send " +
				"NACK packet. Dying.");

    } /* End sendRenewNack(). */

    /**
     * This method sends a "renew" response to the client. This
     * response can be either an ACK or a NACK, as decided by the
     * callee. All this method really does is to fire out the message,
     * and then block the thread. If we get interrupted, that means
     * that the main server has gotten another message from this
     * client session. In that case, our response must have been lost,
     * so we wake up, consume the client's messsage, and if all is
     * well, retransmit.
     *
     * @param msg The DPRPMessage to send to the client.
     * @return True if the send came off without a hitch, false if we
     *  never were able to ram it through.
     */
    public boolean sendRenewResp (DPRPMessage msg) {
	boolean clientStupid = true;
	int curNR = 0;

	if (SimpleDPRPServer.DEBUG) {
	    System.out.println ("SRR: sending '" + msg.getMessageStr() +
		       "', for a maximum of " + NUM_RETRIES + " times.");
	    msg.printDump();
	}

	while ((clientStupid) && (curNR < NUM_RETRIES)) {
	    try {
		localSock.send (msg);
		clientStupid = false;
		curNR++;
		sleep (RETRY_TIMEOUT);
	    } catch (IOException e) {
		System.err.println (e);
	    } catch (InterruptedException e) {
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("SRR: got an interrupted exception.");
		if (passedMessage != null) {
		    if (SimpleDPRPServer.DEBUG)
			System.out.println ("SRR: found a passed-in message " +
					    "from main(). Examining...");

		    if (passedMessage.getMessageType() == DPRPMessage.ERROR) {
			if (SimpleDPRPServer.DEBUG)
			    System.out.println (DPRPErrors.resolveError (passedMessage.getFlags()));
			passedMessage = null;
			return (false);
		    } else if (passedMessage.getMessageType() == DPRPMessage.CLI_RENEW_REQUEST) {
			if (SimpleDPRPServer.DEBUG)
			    System.out.println ("SRR: re-sending #" +
						curNR + ".");
			clientStupid = true;
			passedMessage = null;
		    } else {
			// Got an unknown message.
			if (SimpleDPRPServer.DEBUG)
			    System.out.println ("SRR: got an unknown " +
						"message, sending error.");
			localSock.sendError (msg, DPRPErrors.INVALID_PACKET_TYPE);
			passedMessage = null;
			return (false);
		    }
		} /* if passedMessage != null */
	    } /* catch */
	} /* while */

	// If we terminated because we exceeded NUM_RETRIES, then return
	// we should note that we just plum gave up.
	if ((curNR == NUM_RETRIES) && (clientStupid))
	    return (false);

	return (true);
    } /* End sendRenewResp(). */

    /**
     * The new way to call the <i>send and maybe receiver</i>, which
     * will eventually take over.
     * @param outMessage A message to send.
     * @param msgShouldntRecv The message type that we don't want to
     *  get back. If this is zero, then getting back no message is what
     *  we want.
     * @return A message that was received, if we want one.
     */
    private DPRPMessage sendAndMaybeRecv (DPRPMessage outMessage,
					  byte msgShouldntRecv) {
	if (msgShouldntRecv == 0)
	    return (sendAndMaybeRecv (outMessage, (byte)0, true));
	else
	    return (sendAndMaybeRecv (outMessage, msgShouldntRecv, false));
    }

    /**
     * Takes in a fully-formed DPRPMessage, and spits out a response message
     * form the client. The 'maybe' flag controls whether a response from
     * the client is necessary or not.
     *
     * To be pedantic, I can fold the last two parameters into one. Scrap
     * the boolean, and if msgShouldntReceive is 0, then expectingMsg is
     * (symbolically) true. Otherwise, it's false, and we're in
     * we-get-a-message-is-bad mode. Confused yet? This method is _entirely_
     * too complicated.
     *
     * @param outMessage The message to send out to the client.
     * @param msgShouldntRecv The type of message that is expected in
     *  response. If 0, don't expect anything.
     * @param expectingMsg Controls wether a message should be received or not.
     * @return A DPRPMessage if we got a valid one, null otherwise.
     */
    private DPRPMessage sendAndMaybeRecv (DPRPMessage outMessage,
					  byte msgShouldntRecv,
					  boolean expectingMsg) {
	DPRPMessage inMessage = new DPRPMessage();

	if (SimpleDPRPServer.DEBUG) {
	    System.out.println (getName() + "SAMR starting, shouldRcv = " +
				msgShouldntRecv + "; expct = " + expectingMsg);
// 	    outMessage.printDump();
	}

	try {
	    localSock.send (outMessage);
	} catch (IOException e) {
	    System.err.println (e);
	}

	/*
	 * Now, we need to hang out to make sure that the client got
	 * "our package". If they did, then all is well. If not, then
	 * we have to re-send until they get it. Sometimes, we'll
	 * know that the client didn't get it because they didn't send
	 * us an expected response (such as an ACK). Other times, we'll
	 * know they didn't get it because they're still sending us the
	 * _last_ packet in the sequence. Hence all of the complexity.
	 */
	for (int wt = 0; wt < NUM_RETRIES; wt++) {
	    if (localSock.receive (inMessage)) {
		if (SimpleDPRPServer.DEBUG) {
		    System.out.println (getName() + "received message " +
					"in SAMR:");
		    inMessage.printDump();
		}

		// Uh-oh, we got something. Better check it out.
		if (!expectingMsg && (inMessage.getMessageType() == msgShouldntRecv)) {
		    try {
			// for half-duplex connections (client can
			// send packets to us, we can't send back)
			// this strategy will loop forever.
			if (SimpleDPRPServer.DEBUG)
			    System.out.println (getName() + "got " +
						"shouldReceive, re-sending.");
			wt = 0;  // reset the counter.
			localSock.send (outMessage);
		    } catch (IOException e) {
			System.err.println (e);
		    }
		} else if (inMessage.getMessageType() == DPRPMessage.ERROR) {
		    // "In my house, we call them 'uh-ohs'". -RW
		    if (SimpleDPRPServer.DEBUG)
			System.out.println (getName() + "Got error from " +
					    "client: " +
					    DPRPErrors.resolveError (inMessage.getFlags()));
		    return (null);
		} else if (expectingMsg) {
		    // Fantastic! We were waiting for a message, and
		    // we got it! All is happy! Right? Well, first, make sure
		    // we got a valid XID in our new spiffy packet.
		    if (inMessage.getXid() != curXid) {
			if (SimpleDPRPServer.DEBUG) {
			    System.out.println (getName() + "message with " +
						"invalid xid recieved:");
			    inMessage.printDump();
			}

			// send error
			localSock.sendError (outMessage,
					     DPRPErrors.INVALID_XID_SENT);
			// exit
			return (null);
		    }
		    // Well, now it's all good.
		    return (inMessage);
		}
	    } else if (expectingMsg) {
		// Looks like receive timed out, we should re-send.
		try {
		    if (SimpleDPRPServer.DEBUG)
			System.out.println (getName() + "Receive timed-" +
					    "out, resending #" + wt + "...");
		    localSock.send (outMessage);
		} catch (IOException e) {
		    System.err.println (e);
		}
	    } /* else if */
	} /* for */

	// If we get here, it's bad.
	return (null);
    } /* End sendAndMaybeRecv(). */
} /* End class DPRPServerSession. */
