//package edu.bucknell.net.JDHCP;
package edu.uiuc.cs.net.DPRPManager;

import java.net.*;

/**
 * This class represents a Socket for sending DPRP Messages. It was adapted
 * from the JDHCP package, available at <a
 * href="http://www.dhcp.org/javadhcp/main.html">http://www.dhcp.org/javadhcp/main.html</a>.
 * This package implements a simple DHCP client in Java, and was the
 * inspiration for a lot of the DPRP code base. However, this is the
 * only file that was actually ripped from the JDHCP distribution, and
 * adapted to the DPRP cause. Aside from some major tweaks to
 * <code>receive()</code>, I pretty much just did a search-and-replace
 * job.
 *
 * @version 1.0 - September 28th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 * @version 1.1.1  9/06/1999
 * @author Jason Goldschmidt (The DHCP incarnation)
 * @see java.net.DatagramSocket
 */
public class DPRPSocket extends DatagramSocket  {
    /** The size of a packet -- defaults to 1500 bytes (Ethernet MTU). */
    static protected int PACKET_SIZE = 1500;
    /** Default socket timeout (defaults to 4 seconds). */
    private int defaultSOTIME_OUT = 4000;
    /**
     * Default receive socket timeout (defaults to 4 seconds). Originally,
     * I was trying to differentiate between send and receive timeouts, but
     * I don't know if the actual code reflects this.
     */
    private int defaultRcvTIME_OUT = 4000;
    /** Default send timeout (defaults to 2 seconds). */
    private int defaultSndTIME_OUT = 2000;

    /**
     * Constructor for creating a DPRPSocket on the "next available"
     * local port on the machine.
     * @throws SocketException If the socket cannot be created/bound.
     */
    public DPRPSocket() throws SocketException {
	super();
	setSoTimeout(defaultRcvTIME_OUT); // set default time out
    }

    /** 
     * Constructor for creating DPRPSocket on a specific port on the local
     * machine. 
     * @param inPort the port for the application to bind.
     * @throws SocketException If the socket cannot be created/bound.
     */
    public DPRPSocket (int inPort) throws SocketException {
	
	super(inPort);
	setSoTimeout(defaultRcvTIME_OUT); // set default time out
    }
    
    /**
     * Constructor, that binds the new socket to a specific IP address &
     * port combination.
     * @param inPort The port to which we should attempt our bind.
     * @param laddr The local IP address to which we should attempt to bind.
     * @throws SocketException If the socket cannot be created/bound.
     * @return A valid <code>DPRPSocket</code>, or <code>null</code> if one
     *  couldn't be created & bound.
     */
    public DPRPSocket (int inPort, InetAddress laddr) throws SocketException {
	super (inPort, laddr);
	setSoTimeout(defaultRcvTIME_OUT); // set default time out
    }

    /**
     * Sets the Maximum Transfer Unit for the UDP DHCP Packets to be set.
     * Default is 1500, MTU for Ethernet
     * @param inSize integer representing desired MTU
     */
    
    public void setMTU(int inSize) {
	PACKET_SIZE = inSize;
    }

    /**
     * Returns the set MTU for this socket
     * @return the Maximum Transfer Unit set for this socket
     */
    
    public int getMTU() {
	return PACKET_SIZE;
    }
    
    /**
     * Sends a DHCPMessage object to a predifined host. In the JDHCP
     * incarnation, this is protected via the <code>synchronized</code> flag,
     * which enforces a monitor lock, and makes all calls to this
     * method ordered. I have lifted this flag, because the way I use
     * this method, only one thread will ever be hitting upon a
     * specific socket at a time.
     * @param inMessage well-formed DHCPMessage to be sent to a server
     */
    public void send (DPRPMessage inMessage)
	 throws java.io.IOException {
	byte data[] = new byte[PACKET_SIZE];
	data = inMessage.serialize();

	DatagramPacket outgoing = 
	    new DatagramPacket(data,
			       data.length,
			       inMessage.getDestinationAddress(),
			       inMessage.getDestinationPort());
	//gSocket.
	send(outgoing); // send outgoing message
	
    } /* End send(). */

    /** 
     * Receives a datagram packet containing a DHCP Message into
     * a DHCPMessage object. Again, I lifted the <code>synchronized</code>
     * restriction here.
     * @return true if message is received, false if timeout occurs.  
     * @param outMessage DHCPMessage object to receive new message into
     */
    public boolean receive (DPRPMessage outMessage) {
	DatagramPacket incoming = null;
	if (outMessage == null) {
	    if (DPRPLease.DEBUG)
		System.out.println ("Passed in null Message to receive().");
	    return (false);
	}

	try {
	    incoming = 
		new DatagramPacket(new byte[PACKET_SIZE], 
				   PACKET_SIZE);
	    //gSocket.
	    receive(incoming); // block on receive for SO_TIMEOUT
	    
	} catch (java.io.InterruptedIOException e) {
	    // receive timed out
	    if (DPRPLease.DEBUG)
		System.err.println ("InterruptedIOException in receive: " + e);

	    // What does this mean? We got a partial packet? None at all?
	    return (false);
	} catch (java.net.SocketException e) {
	    // remote end not listening
	    if (DPRPLease.DEBUG)
		System.err.println ("SocketException thrown in receive: " + e);

	    try {
		java.lang.Thread.sleep (defaultSOTIME_OUT);
	    } catch (java.lang.InterruptedException ie) {
		if (DPRPLease.DEBUG)
		    System.out.println ("got interruptedException: " + ie);
		return (false);
	    }
	    return (false);
	} catch (java.io.IOException e) {
	    if (DPRPLease.DEBUG)
		System.err.println ("IOException thrown in receive: " + e);
	    return (false);
        }  // end all catches
	
	if (incoming != null) {
	    byte[] t = incoming.getData();
	    outMessage.deserialize(t);
	} else
	    return (false);

	/*
	 * I'm not so sure about this, but what-the-hey!
	 * On the off-chance that we can't deserialize the packet,
	 * the deserialize() method will give us a dummy error.
	 * We'll just handle it here, and not even notify the caller.
	 * Awful TCP-like, aren't I... :)
	 */
	if ((outMessage.getMessageType() == DPRPMessage.ERROR) &&
	    (outMessage.getFlags() == DPRPErrors.INVALID_PACKET_FMT)) {
	    DPRPMessage tmp = (DPRPMessage)outMessage.clone();
	    tmp.flipSendRecv();
	    sendError (tmp, DPRPErrors.YOU_SENT_INVALID_PKT_FMT);
	    return (false);
	}
	
	return (true);
    } /* End receive(). */

    /**
     * A method that I crafted, that sends an error message. It's basically
     * a useful short-hand for pumping out error messages in times of
     * trouble.
     * @param errMsg A message to send, with the proper source and destination
     *  addresses set.
     * @param errno An error number, to stuff into the message before sending.
     */
    public void sendError (DPRPMessage errMsg, int errno) {
	errMsg.setMessageType (DPRPMessage.ERROR);
	errMsg.setFlags ((short)errno);

	byte data[] = new byte[PACKET_SIZE];
	data = errMsg.serialize();

	DatagramPacket outgoing = 
	    new DatagramPacket(data,
			       data.length,
			       errMsg.getDestinationAddress(),
			       errMsg.getDestinationPort());

	// I'm probably going to want these exceptions to just go to /dev/null.
	try {
	    send(outgoing);
	} catch (java.net.SocketException e) {
	    System.err.println ("sendError(): " + e);
	} catch (java.io.IOException e) {
	    System.err.println ("sendError(): " + e);
	}
    } /* End sendError(). */
} /* End class DPRPSocket. */
 
