package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * The Server Port Manager - handles the management of leases and
 * restrictions in the Server. This class has a lot of code because
 * of it's dual nature -- it must track leases, as well as the
 * server's security (client restrictions). When the server starts,
 * it will instantiate a 'ServerPortManager', and then query it
 * as requests are processed.
 *
 * <p>
 *
 * <b>ABOUT THE LINUX PORT ALLOCATION CODE:</b><br>
 * If an appropriate Linux-based machine with netfilter support is detected,
 * the server will enable "linux mode" in the server port manager. In
 * this mode, the <b><tt>allocFreeExternalPort()</tt></b> method will
 * place a reservation rule for the client in the "destination NAT"
 * netfilter chain. This will enable actual packets, sourced externally,
 * to pass through the gateway (through the port specified) to the internal
 * client host. This reservation is setup via the <code>iptables</code>
 * command, which is prepared and exec'd by this code. The commands
 * executed conform to the following template:
 *
 * <pre>
 *   /usr/sbin/iptables -v -t nat -A PREROUTING -i eth0 -j DNAT \
 *      -p &lt;protocol&gt; --destination &lt;external IP&gt; \
 *      --dport &lt;allocPort&gt; --to &lt;clientIP&gt;:&lt;allocPort&gt;
 * </pre>
 *
 * The following table details each parameter, and it's intended function:
 *
 * <p>
 *
 * <table border=1>
 *  <tr>
 *    <th>Parameter:</th>
 *    <th>Description:</th>
 *  </tr>
 *  <tr>
 *    <td><code>-v</code></td>
 *    <td>Enables verbosity, which is only useful for debugging.</td>
 *  </tr>
 *  <tr>
 *    <td><code>-t nat</code></td>
 *    <td>Selects the "nat" table.</td>
 *  </tr>
 *  <tr>
 *    <td><code>-A PREROUTING</code></td>
 *    <td>Selects the "PREROUTING" chain of the nat table. This means that
 *        the packets will be mangled <i>before</i> being passed to the IP
 *        routing code.</td>
 *  </tr>
 *  <tr>
 *    <td><code>-i eth0</code></td>
 *    <td>Specifies that the inbound interface for packets matching this
 *        rule should be <i><code>eth0</code></i>. <i>NOTE: This should
 *        <b>not</b> be hard-coded, and will be parameterized in a future
 *        release.</i></td>
 *  </tr>
 *  <tr>
 *    <td><code>-j DNAT</code></td>
 *    <td>Causes the packet match to jump to the "destination nat" builtin
 *        target.</td>
 *  </tr>
 *  <tr>
 *    <td><code>-p &lt;protocol&gt;</code></td>
 *    <td>Sets the protocol (tcp or udp) type that packets must be in order
 *        to match.</td>
 *  </tr>
 *  <tr>
 *    <td><code>--destination &lt;external IP&gt;</code></td>
 *    <td>Ensures that the packets are bound for the external address of
 *        the gateway before they are mangled.</td>
 *  </tr>
 *  <tr>
 *    <td><code>--dport &lt;allocPort&gt;</code></td>
 *    <td>This is the actual port that we have reserved -- if the packet is
 *        bound for the external IP address, and matches this port, then it
 *        will be NAT'd and routed into our client.</td>
 *  </tr>
 *  <tr>
 *    <td><code>--to &lt;clientIP&gt;:&lt;allocPort&gt;</code></td>
 *    <td>The ultimate destination for these packets -- an IP address and
 *        port on the client that made this reservation, and wants to eat
 *        these packets for lunch.</td>
 *  </tr>
 * </table>
 *
 * <p>
 *
 * The final caveat to mention about this port allocation code is that I
 * needed to devise some way in order to avoid conflicts between ports
 * reserved by DPRP clients, and ports reserved in the gateway during the
 * normal course of source NAT routing. Since the 'iptables' command doesn't
 * emit an error in the case of such a conflict, I decided to make this an
 * administrative issue. The range of ports allowable for use with SNAT needs
 * to be restricted, and the left-overs given to the DPRP code.
 *
 * <p>
 *
 * <b>ABOUT THE CLIENT RESTRICTION CODE:</b><br>
 * The most code-intensive part of the client restrictions aspect is actually
 * reading the restrictions in from disk. This involves a lot of text parsing,
 * and the like, and is quite a pain. Nevertheless, here is the format
 * for the restriction entries that the server can read in from disk:
 *
 * <p>
 *
 * <blockquote><pre>
 *   host <hostname/ip>
 *        [renewAllowed = true | false]    default = true
 *        [numRetries = <#>]               default = 0 (unlimited)
 *        [portRange = <#> - <#>]          default = 0 - 65536
 *        [minDuration = <#>]              default = 500
 *        [maxDuration = <#>]              default = 2147483647
 *        [deny = true | false]            default = false
 *
 * At least one option must be specified for each host block. Comments can be 
 * inserted anywhere, so long as they begin wtih a '#' character.
 * </pre></blockquote>
 *
 * @version 1.3 - October 30th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class ServerPortMgr {
    /**
     * Map that stores all of the ports that have been reserved. It
     * maps primary keys to <b><tt>DPRPLease</tt></b> objects.  Also,
     * it is important to remember that the <code>HashMap</code> is
     * not thread-safe, and thus all accesses to it must be protected
     * by a semaphore.
     */
    private HashMap portResv;

    /**
     * Map that stores all of the client restrictions. This maps
     * client <b><tt>InetAddress</tt></b> objects to
     * <b><tt>DPRPRestriction</tt></b> objects. Also, it is important
     * to remember that the <code>HashMap</code> is not thread-safe,
     * and thus all accesses to it must be protected by a semaphore.
     */
    private HashMap clientRestrictions;

    /** The name of the on-disk lease database. */
    private String leaseFileName;

    /** The default minimum external port in the range. */
    private static final int DEFAULT_MIN_PORT = 65500;
    /** The default maximum external port in the range. */
    private static final int DEFAULT_MAX_PORT = 65535;
    /** The minimum allocatable external port, as configured. */
    private int minExtPort;
    /** The maximum external port of those which we are allowed to allocate. */
    private int maxExtPort;
    /** Set to true if server is running on a compliant Linux/netfilter box. */
    private boolean hostIsLinux;
    /** The current port to be allocated in the external range. */
    private int curPort;
    /** A bit field, marking the ports in range that are reserved or not. */
    private boolean[] resvdPorts;
    /** The external IP address upon which we are allocating ports (lame). */
    private InetAddress externalIP;

    /**
     * Creates a new ServerPortManager instance.
     */
    public ServerPortMgr() {
	// initialize data structures.
	portResv = new HashMap();
	clientRestrictions = new HashMap();
	leaseFileName = null;

	// initialize port dealie values.
	hostIsLinux = false;
	minExtPort = DEFAULT_MIN_PORT;
	maxExtPort = DEFAULT_MAX_PORT;
	curPort = 0;

	// Allocate a default notion of the reserved ports array. Will every
	// slot in this array default to false?
	resvdPorts = new boolean[maxExtPort - minExtPort];
	if (SimpleDPRPServer.DEBUG)
	    System.out.println ("SPM::SPM: Allocated default # " +
				(maxExtPort - minExtPort) +
				" of slots for the reserved ports array.");
    } /* End ServerPortMgr constructor. */

    /**
     * Checks to see if the external port in a given lease is okay. By
     * okay, I mean it checks the restrictions set in place for a
     * particular client, and determines if the port is in the allowed
     * range. This also checks to make sure that the specified
     * protocol is kosher as well.
     * @param lea A <b><tt>DPRPLease</tt></b>, whose external port is
     *  to be checked.
     * @return True if the port is allowed, false otherwise.
     */
    public synchronized boolean checkPort (DPRPLease lea) {
	if (lea.getProtocolStr().equals ("unk")) {
	    // We're going to require that if a port is specified, 
	    // it had better have a protocol too (none of this
	    // "both" shit).
	    return (false);
	}

	/*
	 * Check to see if the specified port is allowed by
	 * configuration rule.
	 */
 	DPRPRestriction pr;
	pr = (DPRPRestriction)clientRestrictions.get (lea.getClientIP());
	if (pr != null)
	    if ((lea.getExternalPort() < pr.getMinPort()) ||
		(lea.getExternalPort() > pr.getMaxPort()))
		return (false);

	// Now, see if this requested port is in the range of ports that
	// we can actually *allocate* from!
	if ((lea.getExternalPort() < minExtPort) ||
	    (lea.getExternalPort() > maxExtPort))
	    return (false);

	// If we get here, than it is in range (miraculously). Let's see if
	// it's free, shall we? Shouldn't this be test and set,
	// concurrency being what it is?
	if (resvdPorts[lea.getExternalPort() - minExtPort])
	    return (false);

	// Okay, it's free and we want it -- grab it!
	// instead, maybe i should just call allocExternalPort()??? In fact,
	// i could do that instead of checking the array, and just eat
	// any exceptions and return false.
	resvdPorts[lea.getExternalPort() - minExtPort] = true;

	// If we get here, then, err, um, true!
	return (true);
    } /* End checkPort(). */

    /**
     * Checks to see that the specified duration is in-range, with
     * respect to client restrictions. This is done by extracting the
     * duration in the given lease, and comparing it to the bounds
     * specified in the restrictions for the client.
     * @param lea A lease, specifying a client IP address and a duration.
     * @return True if the duration satisfies the bounds, false otherwise.
     */
    public boolean checkDuration (DPRPLease lea) {
	DPRPRestriction pr;
	pr = (DPRPRestriction)clientRestrictions.get (lea.getClientIP());
	if (pr == null)
	    return (true);
	// check for any duration restrictions
	else if ((lea.getDuration() <= pr.getMaxDuration()) &&
		 (lea.getDuration() >= pr.getMinDuration()))
	    return (true);
	
	return (false);
    } /* End checkDuration(). */

    /**
     * Checks to see if a specific client is allowed to negotiate for
     * a lease renewal.
     * @param cli A client IP address to be looked up.
     * @return True if the client is allowed to renew (the default),
     *  false otherwise.
     */
    public boolean renewOK (InetAddress cli) {
	DPRPRestriction pr = (DPRPRestriction)clientRestrictions.get (cli);
	if (pr != null) {
	    if (pr.renewAllowed())
		return (true);
	} else
	    return (true);

	return (false);
    } /* End renewOK(). */

    /**
     * Examines the number of times a lease has been renewed. If the
     * client is restricted to a certain number of renewals per lease,
     * than this method will enforce this security measure.
     * @param lea A lease, containing a client IP address and the
     *  number of times the lease has been renewed.
     * @return True if the number of renews is less than rule, false
     *  otherwise.
     */
    public boolean checkRetries (DPRPLease lea) {
	DPRPRestriction pr;
	pr = (DPRPRestriction)clientRestrictions.get (lea.getClientIP());
	if (pr != null) {
	    if (pr.checkNumRenews (lea.getNumRenews()))
		return (true);
	} else
	    return (true);

	return (false);
    } /* End checkRetries(). */

    /**
     * Check to see if a client is denied access to the DPRP service.
     * @param cli A client IP address to be looked up in the
     *  restriction database.
     * @return True if the client has been denied by rule, false otherwise.
     */
    public boolean checkDenied (InetAddress cli) {
	DPRPRestriction pr = (DPRPRestriction)clientRestrictions.get (cli);
	if (pr != null)
	    if (pr.isDenied())
		return (true);

	return (false);
    } /* End checkDenied(). */

    /**
     * Returns the maximum allowed duration for a given client. This
     * is needed in order to determine if a server-suggested duration
     * is in-bounds.
     * @param lea A lease, containing a client IP address.
     * @return The maximum allowed duration for that client, in seconds.
     */
    public int getMaxAllowedDuration (DPRPLease lea) {
	DPRPRestriction pr;
	pr = (DPRPRestriction)clientRestrictions.get (lea.getClientIP());
	if (pr != null)
	    return (pr.getMaxDuration());
	else
	    return (DPRPLease.DEFAULT_DURATION);
    } /* End getMaxAllowedDuration(). */

    /************************************************************************/
    /* External port allocation.                                            */
    /************************************************************************/
    // This is inclusive on both the min external port and max external
    // port.
    public boolean setExtPortRange (int sm, int lm) {
	if (sm < lm) {
	    minExtPort = sm;
	    maxExtPort = lm;
	    resvdPorts = new boolean[maxExtPort - minExtPort];
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPM::sEPR: Allocated " +
				    (maxExtPort - minExtPort) +
				    " slots for the reserved ports array.");

	    return (true);
	}

	return (false);
    } /* End setExtPortRange(). */

    /** Returns the minimum external port in range. */
    public int getMinExtPort() {
	return (minExtPort);
    }
    /** Returns the maximum external port in range. */
    public int getMaxExtPort() {
	return (maxExtPort);
    }
    /** Sets the external IP address. */
    public void setExternalIP (InetAddress ia) {
	externalIP = ia;
    }
    /** Sets the Linux flag. */
    public void markAsLinux() {
	hostIsLinux = true;
    }
    /** Returns true if Linux flag has been set, false otherwise. */
    public boolean isLinux() {
	return (hostIsLinux);
    }

    // this will be simplistic in a lot of ways. the first is that we'll
    // allocate ports in-sequence, not caring that tcp and udp could share
    // the same port #. also, we need to check that curPort is free,
    // according to the reservation table.
    //
    // NEED to throw exception here, in order to handle no_free_port errors,
    // and any other error that i can think of (transient linux probs,
    // for example).
    //
    // ALSO, need a datastructure to track the current free port, something
    // that we can increment, and walk through all of the ports in the
    // range, and find the "gaps" that are free. Circular buffer, or bitvector,
    // or something. Boolean array? true = allocated, false = unallocated.
    //
    // Concurrency in this method could be HELLA bad. Think about that.
    //
    // mention problems with the free-port-finding algorithm (like, how
    // it's all O(n) and shit). This is bad in the case that we have to
    // search the ENTIRE port space in order to find that there isn't a
    // free port. Couple that with multiple external IP addresses, and
    // things start to tend O(m^n), where m << n. Very much not making
    // Erickson proud. Maybe the unalloc version could reset the
    // 'curPort' to what it just rm'd?
    public synchronized int allocFreeExternalPort (InetAddress clientIP,
				 byte prot) throws DPRPAllocatePortException {
	// the port that it's okay to allocate.
	int allocPort = 0;

	// Check to see that curPort is free. This loop can't run infinitely
	// in the case that there are no free ports.
	int poc = 0;
	while (resvdPorts[curPort] && (poc < resvdPorts.length)) {
	    curPort = (curPort + 1) % resvdPorts.length;
	    poc++;
	}
	if (poc < resvdPorts.length) {
	    // We got one! (said as Janine, from "Ghostbusters")
	    resvdPorts[curPort] = true;  // for now.
	    allocPort = curPort + minExtPort;

	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPMdb-aR: marked locus '" + curPort +
				    "' as true in resvdPorts[].");
	} else {
	    // Throw an exception, no ports are free.
	    throw new DPRPAllocatePortException (DPRPErrors.NO_FREE_PORTS);
	}

	// Increment the current port, circularily.
	curPort = (curPort + 1) % resvdPorts.length;

	return (allocExternalPort (clientIP, allocPort, prot));
    } /* End allocFreeExternalPort(). */

    /**
     * Reserves a specific external port in netfilter. This should
     * also really be a private method.
     * @param clientIP The clients IP address for this reservation.
     * @param allocPort The external port to be reserved for the client.
     * @param prot The IP protocol for this reservation.
     * @throws DPRPAllocatePortException If an error occurs down in
     *  <b><tt>allocLinuxPort</tt></b>, or if <code>clientIP</code> is null.
     * @return The port number that was allocated.
     */
    public int allocExternalPort (InetAddress clientIP, int allocPort,
				  byte prot) throws DPRPAllocatePortException {
	if (clientIP == null) {
	    // error, throw exception or something?
	    throw new DPRPAllocatePortException (DPRPErrors.UNKNOWN_ERROR,
                             "Got null clientIP in SPM::allocExternalPort().");
	}

	if (hostIsLinux)
	    return (allocLinuxPort (clientIP, allocPort, prot));
	else
	    // do the slutty test-mode thing
	    return (allocPort);
    } /* End allocExternalPort(). */

    /**
     * Builds the <code>iptables</code> command that creates a DNAT hole
     * on the gateway. I really wish that this method was marked as
     * private. Also, I need to configure the internal and external
     * interface names into the SPM.
     * @param clientIP The clients IP address for this reservation.
     * @param allocPort The external port to be reserved for the client.
     * @param prot The IP protocol for this reservation.
     * @throws DPRPAllocatePortException If an error occurs down in
     *  <b><tt>execCmd</tt></b>.
     * @return The port number that was allocated.
     */
    public int allocLinuxPort (InetAddress clientIP, int allocPort,
				byte prot) throws DPRPAllocatePortException {
	// build the command.
	String[] cmd = new String[18];

	// do other distros have this in a different place?
	cmd[0] = "/sbin/iptables";
	cmd[1] = "-v";            // Enable verbose mode.
	cmd[2] = "-t";
	cmd[3] = "nat";
	cmd[4] = "-A";
	cmd[5] = "PREROUTING";
	cmd[6] = "-p";
	cmd[7] = DPRPMessage.protToString (prot);
	cmd[8] = "--dport";
	cmd[9] = "" + allocPort;
	cmd[10] = "-i";
	cmd[11] = "eth0";         // this can't be so static.
	cmd[12] = "-j";
	cmd[13] = "DNAT";
	cmd[14] = "--destination";
	cmd[15] = externalIP.getHostAddress();
	cmd[16] = "--to";
	cmd[17] = clientIP.getHostAddress() + ":" + allocPort;

	// Okay, fire in the hole!
	execCmd (cmd);

	return (allocPort);
    } /* End allocLinuxPort(). */

    /**
     * Removes a port allocation from the system. This entails marking
     * the appropriate location in the <code>resvdPorts</code> array
     * as false, and then exec'ing the proper <code>iptables</code>
     * command, via the <b><tt>unallocLinuxPort</tt></b> method.
     * @param clientIP The clients IP address for this reservation.
     * @param allocPort The external port reserved for the client.
     * @param prot The IP protocol for this reservation.
     * @return True if the <code>exec()</code> goes okay, false otherwise.
     * @see #unallocLinuxPort (InetAddress clientIP, int allocPort, byte prot)
     */
    public synchronized boolean unallocExternalPort (InetAddress clientIP,
						     int allocPort,
						     byte prot) {
	resvdPorts[allocPort - minExtPort] = false;
	if (SimpleDPRPServer.DEBUG)
	    System.out.println ("SPM::uEP: Released port: " + allocPort + ".");

	// Necessary in order to prevent "kaboom!" of the highest order.
	if (hostIsLinux)
	    return (unallocLinuxPort (clientIP, allocPort, prot));
	else
	    return (true);
    } /* End unallocExternalPort(). */

    /**
     * Prepares the iptables command for unallocating a reservation. I had
     * a hard time realizing this method, because it seemed like deleting
     * a rule had to be done by rule number. Tracking rule numbers would
     * essentially de-volve into running a list command, parsing, and then
     * finding the right number. But even that could fail, if a spry
     * administrator were to change the rule numbers in-between the list
     * and the delete. Luckily, I noticed that rule delete's can occur
     * by specifying the complete rule, which makes things <b>dramatically</b>
     * easier!
     * @param clientIP The clients IP address for this reservation.
     * @param allocPort The external port reserved for the client.
     * @param prot The IP protocol for this reservation.
     * @return True if the <code>exec()</code> goes okay, false otherwise.
     */
    private boolean unallocLinuxPort (InetAddress clientIP, int allocPort,
				      byte prot) {
	String[] cmd = new String[18];

	cmd[0] = "/sbin/iptables";
	cmd[1] = "-v";            // Enable verbose mode.
	cmd[2] = "-t";
	cmd[3] = "nat";
	cmd[4] = "-D";
	cmd[5] = "PREROUTING";
	cmd[6] = "-p";
	cmd[7] = DPRPMessage.protToString (prot);
	cmd[8] = "--dport";
	cmd[9] = "" + allocPort;
	cmd[10] = "-i";
	cmd[11] = "eth0";         // this can't be so static.
	cmd[12] = "-j";
	cmd[13] = "DNAT";
	cmd[14] = "--destination";
	cmd[15] = externalIP.getHostAddress();
	cmd[16] = "--to";
	cmd[17] = clientIP.getHostAddress() + ":" + allocPort;

	try {
	    execCmd (cmd);
	} catch (DPRPAllocatePortException e) {
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPM::uLP: Got error on delete:\n\t" + e);
	    return (false);
	}
	return (true);
    } /* End unallocLinuxPort(). */

    /**
     * Actually makes the <code>Runtime.exec()</code> call, for the pre-setup
     * command-array. The command array format is such that the command
     * itself (with full path name, just to be safe) is in slot 0. Each
     * parameter (that would normally be space-separated if invoking from
     * a shell) are in the ajoining array slots. If no exception is
     * thrown, then it is presumed that the command completed successfully.
     * @param cmdAry A properly-formatted command array, for the command to
     *  be executed.
     * @throws DPRPAllocatePortException If something goes arwy.
     */
    private void execCmd (String[] cmdAry) throws DPRPAllocatePortException {
	Runtime curRuntime = null;
	Process curProcess = null;
	int exitCode = -99;

	if (SimpleDPRPServer.DEBUG) {
	    System.out.print ("Executing command: ");
	    for (int i = 0; i < cmdAry.length; i++)
		System.out.print (cmdAry[i] + " ");
	    System.out.println ("");
	}

	try {
	    // First, we get the current runtime instance.
	    curRuntime = Runtime.getRuntime();
	    // Now, we can exec commands in our current environment.
	    curProcess = curRuntime.exec (cmdAry);
	    // This should execute fast, but just to be safe, we'll wait for
	    // iptables to be done.
	    exitCode = curProcess.waitFor();

	    if (SimpleDPRPServer.DEBUG) {
		System.out.println ("Command executed; return status = " +
				    exitCode);

		// Go through all of the girations necessary in order to get
		// the stdout emitted by the command.
		InputStreamReader outStream;
		outStream = new InputStreamReader(curProcess.getInputStream());
		LineNumberReader cmdOut = new LineNumberReader (outStream);
		String curLine = cmdOut.readLine();
		System.out.println ("<< Command output: >>");
		System.out.println ("=====================");
		while (curLine != null) {
		    System.out.println (cmdOut.getLineNumber() +": "+ curLine);
		    curLine = cmdOut.readLine();
		}
	    } /* debug */
	} catch (InterruptedException e) {
	    if (SimpleDPRPServer.DEBUG)
		System.err.println ("SPM::eC: " + e);
	    throw new DPRPAllocatePortException (DPRPErrors.UNKNOWN_ERROR,
						 "SPM::eC: Was interrupted " +
						 "while exec'ing command.\n" +
						 "Exception: " + e);
	} catch (IOException e) {
	    if (SimpleDPRPServer.DEBUG) {
		System.err.println ("SPM::eC: " + e);
		e.printStackTrace();
	    }
	    throw new DPRPAllocatePortException (DPRPErrors.UNKNOWN_ERROR,
						 "SPM::eC: Got IO error " +
						 "while in sub-process.\n" +
						 "Exception: " + e);
	} /* catch */

	// Make sure everything went OK.
	if (exitCode != 0)
	    throw new DPRPAllocatePortException (DPRPErrors.UNKNOWN_ERROR,
						 "SPM::eC: Got nonzero " +
						 "return code from " +
						 "iptables: " + exitCode);
    } /* End execCmd(). */

    /************************************************************************/
    /* Lease reservation database management.                               */
    /************************************************************************/
    /**
     * Adds a lease to the Map, and optionally synchronizes the Map with disk.
     * @param lea The lease object to be added.
     * @param wf If true, writes the leases out to disk.
     */
    public synchronized void addReservation (DPRPLease lea, boolean wf) {
	// Paranoid much?
	if (lea == null) {
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPMdb-aR: not adding bogus null lease!");
	    return;
	}

	// Update the lease management map.
	String key = lea.getPrimaryKey();
	if (SimpleDPRPServer.DEBUG)
	    System.out.println ("SPMdb-aR: add new lease @ key: " + key);
	portResv.put (key, lea);

	// For the sake of _some_ semblance of optimality, we don't always
	// want to write to disk.
	if (wf) {
	    int nl = writeLeaseFile();
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPMdb-aR: Wrote " + nl +
				    " leases out to file.");
	}
    } /* End addReservation(). */

    /**
     * Deletes a reservation (nee Lease) from the Map.
     * @param lea The lease object to be deleted.
     * @param wf Controls wether-or-not the leases should be flushed to disk.
     *  As an optimization in some cases, the writing of the leases to disk
     *  can be handled by the callee.
     */
    public synchronized void delReservation (DPRPLease lea, boolean wf) {
	String key = lea.getPrimaryKey();
	DPRPLease removed = (DPRPLease)portResv.remove (key);

	// Only write if the flag says, and if we actually deleted something!
	if (wf && (removed != null)) {
	    // write-file here.
	    int nl = writeLeaseFile();
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPM-dR: Wrote " + nl +
				    " leases out to file.");	    

	    // Finally, remove this reservation from the kernel.
	    if (unallocExternalPort (removed.getClientIP(),
				     removed.getExternalPort(),
				     removed.getProtocol())) {
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("SPM-dR: Removed reservation from " +
					"kernel without a hitch.");
	    } else {
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("SPM-dR: Error occured removing " +
					"reservation from kernel.");
	    }
	} else if ((SimpleDPRPServer.DEBUG) && (removed == null))
	    System.out.println ("SPM-dR: lease not found; not writing file " +
				"or purging netfilter rule.");

    } /* End delReservation(). */

    /**
     * Updates the object stored at the primary key with a new lease object.
     * @param lea The new lease object that replaces a prior object.
     */
    public synchronized void updateReservation (DPRPLease lea) {
	delReservation (lea, false);
	addReservation (lea, false);

	// write file here.
	int nl = writeLeaseFile();
	if (SimpleDPRPServer.DEBUG)
	    System.out.println ("uR: Wrote " + nl + " leases out to file.");
    } /* End updateReservation(). */

    /**
     * Retrieves a lease from the Map that corresponds to the given
     * message.
     * @param inMessage A valid DPRPMessage, which will be used to
     *  locate a lease.
     * @return A valid DPRPLease, if found.
     */
    public DPRPLease getReservation (DPRPMessage inMessage) {
	String key = inMessage.getPrimaryKey();
	DPRPLease ret = (DPRPLease)portResv.get (key);
	if (SimpleDPRPServer.DEBUG) {
	    System.out.println ("SPMdb: getting reservation for key: '" +
				key + "'.");
	    if (ret != null)
		System.out.println ("\tLease: " + ret);
	}
	return (ret);
    } /* End getReservation(). */

    /**
     * Writes all of the leases in the Map out to disk, and also purges
     * dead leases along the way.
     * @return The number of leases written.
     */
    private synchronized int writeLeaseFile() {
	Collection pr = portResv.values();
	HashSet initLeaseSet = new HashSet (pr);
	HashSet modLeaseSet = new HashSet (pr);

	if (SimpleDPRPServer.DEBUG) {
	    System.out.println ("SPM-wLF: # of leases in database: " +
				initLeaseSet.size());
	    System.out.println ("SPM-wLF: InitSet: " + initLeaseSet);
	}

	modLeaseSet = DPRPLease.writeLeaseFile (new File(leaseFileName),
						modLeaseSet);

	if (modLeaseSet == null) {
	    // some sort of error occured, or we just didn't have anything
	    // to write.
	    return (0);
	}

	if (SimpleDPRPServer.DEBUG) {
	    System.out.println ("SPM-wLF: # of leases in just-written " +
				"database: " + modLeaseSet.size());
	    System.out.println ("SPM-wLF: ModSet: " + modLeaseSet);
	}

	Iterator it = initLeaseSet.iterator();
	DPRPLease tmpL;
	DPRPLease rmd = null;
	while (it.hasNext()) {
	    // Nice, pre-skool bug.
	    tmpL = (DPRPLease)it.next();

	    System.out.println ("currently sniffin: " + tmpL);

	    if (!modLeaseSet.contains (tmpL)) {
		// this lease should be deleted.
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("SPM-wLF: modLease doesn't contain " +
					"initLease. Deleting lease @ key: " +
					tmpL.getPrimaryKey());

		rmd = (DPRPLease)portResv.remove (tmpL.getPrimaryKey());
		if ((rmd == null) && SimpleDPRPServer.DEBUG)
		    System.out.println ("wLF: Expired lease not found in " +
					"portResv database.");
		rmd = null;
	    } else if (SimpleDPRPServer.DEBUG)
		System.out.println ("SPM-wLF: modLease contains: " + tmpL);

	} /* while */
	return (modLeaseSet.size());
    } /* End writeLeaseFile(). */

    /**
     * Sucks in the on-disk lease database, and reconstructs the lease
     * map in memory. It is called once during server startup, and if
     * the specified file doesn't exist, then no harm is done, and the
     * server starts with a clean lease database. However, if the
     * lease file does exist, then this method reads in each lease
     * string, creates a <b><tt>DPRPLease</tt></b> object, and inserts
     * each object into the <b><tt>portResv</tt></b> map. Furthermore,
     * each lease is checked to make sure that it hasn't expired. If
     * it has, then it is not inserted into the database.
     * @param fname The name of the file to open.
     * @throws IOException If an error occurs dealing with the file
     *  (usually, indicates that the given file doesn't exist).
     * @return The number of leases created.
     */
    public synchronized int readLeaseFile (String fname) throws IOException {
	// Set the name of the lease file, permanently.
	leaseFileName = fname;
	HashSet leasesRead = DPRPLease.readLeaseFile (new File (leaseFileName));
	Iterator it = leasesRead.iterator();
	while (it.hasNext()) {
	    addReservation ((DPRPLease)it.next(), false);
	}

	return (leasesRead.size());
    }

    /************************************************************************/
    /* Build the restriction database.                                      */
    /************************************************************************/
    /**
     * Parses a "Host Security Configuration File" in order to build a map
     * of 'DPRPRestriction' objects.
     * @param filename The name of the file to be parsed.
     * @throws DPRPConfigException If some sort of parsing error is
     *  encountered within the configuration file.
     * @throws IOException If an IO error occurs.
     * @return The number of host blocks successfully parsed.
     */
    public int parseConfigFile (String filename) throws DPRPConfigException, IOException {
	// Local variables
	int numHosts = 0;

	// Open a file for reading.
	File configFile = new File (filename);
	FileReader inR = new FileReader (configFile);
	LineNumberReader inLNR = new LineNumberReader (inR);

	// prime the pump, so to speak.
	String curLine = inLNR.readLine();
	String fw;  // first word
	StringTokenizer tok;

	while (curLine != null) {
	    // there is more input to deal with.
	    tok = new StringTokenizer (curLine);
	    while (!tok.hasMoreTokens()) {
		curLine = inLNR.readLine();
		if (curLine == null)
		    return (numHosts);
		tok = new StringTokenizer (curLine);
	    }
	    fw = tok.nextToken();
	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("line read: " + curLine);
	    if (fw.startsWith ("#")) {
		// comment, ignore
		curLine = inLNR.readLine();
		continue;
	    } else if (fw.equalsIgnoreCase ("host") && (tok.countTokens() >= 1)) {
		// valid host state
		String hostIPStr = tok.nextToken();
		InetAddress hostIP;
		try {
		    hostIP = InetAddress.getByName (hostIPStr);
		} catch (UnknownHostException e) {
		    throw new DPRPConfigException ("Error on line " +
						   inLNR.getLineNumber() +
						   ": Invalid host name specified.");
		} /* catch */

		// now i need to look for host-y type options.
		curLine = inLNR.readLine();
		tok = new StringTokenizer (curLine);
		fw = tok.nextToken();

		// if this is host, then error. Otherwise, need to
		// check for all keywords, and do the 'appropriate' thing.
		if (fw.equalsIgnoreCase ("host"))
		    throw new DPRPConfigException ("Error on line " +
						   inLNR.getLineNumber() +
						   ": Empty host block encountered.");

		// Everything looks good, go ahead and parse all of
		// the options for this host.
		DPRPRestriction pr = parseAllHostOptions (hostIP, curLine, inLNR);
		if (SimpleDPRPServer.DEBUG) {
		    System.out.println ("pCF: Got a full-on DPRPRestriction:");
		    pr.prettyPrint();
		}

		// Drop this into the database.
		clientRestrictions.put (pr.getClientIP(), pr);

		// Cleanup.
		inLNR.reset();
		numHosts++;
	    } else {
		// error -- didn't find '#' or 'host'.
		throw new DPRPConfigException ("Error on line " +
					       inLNR.getLineNumber() +
					       ": Invalid line encountered");
	    }

	    // Update curLine before repeating.
	    curLine = inLNR.readLine();
	} /* while */

	inLNR.close();
	inR.close();
	return (numHosts);
    } /* End parseConfigFile(). */

    /**
     * Parses out all of the options inside of a given host
     * block. This means that we loop, grabbing successive lines of
     * text, until a new "host" block is encountered, or EOF. For each
     * line, we check to see what option it specifies, and then farm
     * off the parameters to individual parsing methods. Ain't I a
     * stinkah?
     * @param cliIP The client IP address that we're currently parsing
     *  options for.
     * @param curLine The current line in the configuration file.
     * @param lnr The buffer that we get new lines from.
     * @throws DPRPConfigException If one of the options was invalid.
     * @throws IOException If there was a problem reading from the file.
     * @return A fully-formed <b><tt>DPRPRestriction</tt></b> object.
     */
    private DPRPRestriction parseAllHostOptions (InetAddress cliIP,
						 String curLine,
						 LineNumberReader lnr)
	throws DPRPConfigException, IOException {

	DPRPRestriction ret  = new DPRPRestriction (cliIP);

	StringTokenizer tok = new StringTokenizer (curLine);
	String cw = tok.nextToken();  // current-word.
	int intVal = 0;
	boolean boolVal = true;					       

	while (!cw.equalsIgnoreCase ("host")) {
	    if (cw.equalsIgnoreCase ("renewAllowed") && (tok.countTokens() >= 2)) {
		boolVal = parseBoolean (tok, lnr.getLineNumber());
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("pAHO: found 'renewAllowed', val = " + boolVal);
		if (!boolVal)
		    ret.setRenewNotOK();
	    } else if (cw.equalsIgnoreCase ("numRetries") && (tok.countTokens() >= 2)) {
		intVal = parseInteger (tok, "=", lnr.getLineNumber());
		if (intVal < 0)
		    // Error, can't have negative numbers.
		    throw new DPRPConfigException ("Error on line " +
						   lnr.getLineNumber() +
						   ": Invalid numRetries '" +
						   intVal + "': negative number.");
		ret.setNumRenewals (intVal);
	    } else if (cw.equalsIgnoreCase ("portRange") && (tok.countTokens() >= 4)) {
		int loPort = parseInteger (tok, "=", lnr.getLineNumber());
		int hiPort = parseInteger (tok, "-", lnr.getLineNumber());

		if ((loPort < 0) || (hiPort < 0))
		    throw new DPRPConfigException ("Error on line " +
						   lnr.getLineNumber() +
						   ": Negative number in port range.");
		if (loPort > hiPort)
		    throw new DPRPConfigException ("Error on line " +
						   lnr.getLineNumber() +
						   ": Invalid port range " +
						   "(min greater than max).");

		// Commit these values.
		ret.setMinPort (loPort);
		ret.setMaxPort (hiPort);
	    } else if (cw.equalsIgnoreCase ("minDuration") && (tok.countTokens() >= 2)) {
		// There is probably some bare-minimum that I don't want
		// to go below (DoS, baby), but eh...
		intVal = parseInteger (tok, "=", lnr.getLineNumber());       
		if (ret.getMaxDuration() != DPRPRestriction.defaultMaxDuration)
		    // A max duration has been set.
		    if (intVal > ret.getMaxDuration())
			// Error, given minDuration is too big.
			throw new DPRPConfigException ("Error on line " +
						       lnr.getLineNumber() +
						       ": Specified duration '" +
						       intVal + "' invalid (min greater than max).");
		if (intVal < 0)
		    // Error, can't go back in time.
		    throw new DPRPConfigException ("Error on line " +
						   lnr.getLineNumber() +
						   ": Invalid duration '" +
						   intVal + "': negative number.");
		ret.setMinDuration (intVal);
	    } else if (cw.equalsIgnoreCase ("maxDuration") && (tok.countTokens() >= 2)) {
		intVal = parseInteger (tok, "=", lnr.getLineNumber());
		if (ret.getMinDuration() != DPRPRestriction.defaultMinDuration)
		    // A min duration has been set.
		    if (intVal < ret.getMinDuration())
			// Error, given minDuration is too big.
			throw new DPRPConfigException ("Error on line " +
						       lnr.getLineNumber() +
						       ": Specified duration '" +
						       intVal + "' invalid " +
						       "(min greater than max).");
		if (intVal < 0)
		    // Error, can't go back in time.
		    throw new DPRPConfigException ("Error on line " +
						   lnr.getLineNumber() +
						   ": Invalid duration '" +
						   intVal + "': negative number.");
		ret.setMaxDuration (intVal);
	    } else if (cw.equalsIgnoreCase ("deny") && (tok.countTokens() >= 2)) {
		boolVal = parseBoolean (tok, lnr.getLineNumber());
		if (boolVal)
		    ret.setIsDenied();
	    } else if (cw.startsWith ("#")) {
		// comment, go to next line.
		//curLine = lnr.readLine();
	    } else {
		// error - valid keyword not found
		throw new DPRPConfigException ("Error on line " +
					       lnr.getLineNumber() +
					       ": Invalid config " +
					       "option specified: '" +
					       cw + "'.");
	    }

	    // Advance to the next line. This is *so* gay!
	    lnr.mark (255);
	    curLine = lnr.readLine();
	    if (curLine == null) {
		if (SimpleDPRPServer.DEBUG)
		    System.out.println ("pAHO: found EOF while looking for next line.");
		break;
	    }

	    tok = new StringTokenizer (curLine);
	    while (!tok.hasMoreTokens()) {
		lnr.mark (255);
		curLine = lnr.readLine();
		if (curLine == null) {
		    // what i really need is a double-break.
		    if (SimpleDPRPServer.DEBUG)
			System.out.println ("pAHO: found EOF while jammin' through blank lines.");
		    return (ret);
		}

		tok = new StringTokenizer (curLine);
	    }
	    cw = tok.nextToken();

	    if (SimpleDPRPServer.DEBUG)
		System.out.println ("pAHO: line read: " + curLine);
	} /* while */

	// All's well that ends well.
	return (ret);
    } /* End parseAllHostOptions(). */

    /**
     * Parses out an integer value from the host security configuration
     * file. This is accomplished by looking for the separator character
     * from the string tokenizer, and then a valid integer number.
     * @param tok A string tokenizer, from which we can extract tokens.
     * @param sep A string representing the separator, what we should
     *  look for first. This gives this method a degree of flexability.
     * @param lineNum The current line number that we're on, used to
     *  make pretty error messages.
     * @throws DPRPConfigException If we don't find valid input.
     * @return The <tt>int</tt> representation of the number found.
     */
    private int parseInteger (StringTokenizer tok,
			      String sep,
			      int lineNum) throws DPRPConfigException {
	String cw;

	cw = tok.nextToken();
	if (!cw.equalsIgnoreCase (sep)) {
	    // error
	    throw new DPRPConfigException ("Error on line " +
					   lineNum +
					   ": Separator '" +
					   sep + "' not found.");
	}

	cw = tok.nextToken();
	try {
	    Integer tmp_i = new Integer (cw);
	    return (tmp_i.intValue());
	} catch (NumberFormatException e) {
	    throw new DPRPConfigException ("Error on line " +
					   lineNum +
					   ": Invalid integer '" +
					   cw + "' specified.");
	}
    } /* End parseInteger(). */

    /**
     * Parses out a boolean value from the host security configuration
     * file. This is accomplished by looking for an "=" character from
     * the string tokenizer, and then either the "true" or "false"
     * tokens.
     * @param tok A string tokenizer, from which we can extract tokens.
     * @param lineNum The current line number that we're on, used to
     *  make pretty error messages.
     * @throws DPRPConfigException If we don't find valid input.
     * @return The boolean representation of the string value found.
     */
    private boolean parseBoolean (StringTokenizer tok,
				  int lineNum) throws DPRPConfigException {
	String cw;

	cw = tok.nextToken();
	if (!cw.equalsIgnoreCase ("=")) {
	    // error
	    throw new DPRPConfigException ("Error on line " +
					   lineNum +
					   ": Invalid equals sign '" +
					   cw + "' specified.");
	}

	cw = tok.nextToken();
	if (cw.equalsIgnoreCase ("false"))
	    return (false);
	else if (cw.equalsIgnoreCase ("true"))
	    return (true);
	else
	    throw new DPRPConfigException ("Error on line " +
					   lineNum +
					   ": Invalid boolean parameter '" +
					   cw + "' specified.");
    } /* End parseBoolean(). */

    /**
     * Prints the format of the configuration file to stderr.
     */
    public void printConfigFormat() {
	System.err.println ("\nFormat for server host restriction configuration file:\n");
	System.err.println ("  host <hostname/ip>");
	System.err.println ("       [renewAllowed = true | false]    default = true");
	System.err.println ("       [numRetries = <#>]               default = 0 (unlimited)");
	System.err.println ("       [portRange = <#> - <#>]          default = 0 - 65536");
	System.err.println ("       [minDuration = <#>]              default = 500");
	System.err.println ("       [maxDuration = <#>]              default = 2147483647");
	System.err.println ("       [deny = true | false]            default = false");
	System.err.println ("\nAt least one option must be specified for each " +
			    "host block. Comments can be \ninserted anywhere, " +
			    "so long as they begin wtih a '#' character.");
    } /* End printConfigFormat(). */
} /* End class ServerPortMgr. */
