package edu.uiuc.cs.net.DPRPManager;
//package DPRPManager;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * The main server class, that starts the entire process. This class
 * started out as a really simple server, that just received a packet, and
 * sent it back out again. However through refactoring, this code 
 * became something much more complex. It probably should be renamed,
 * but it's kindof a nice reminder of where this code came
 * from. <tt>:)</tt>
 *
 * @version 1.3 - October 27th, 2001
 * @author Andy Reitz (<a href="mailto:areitz@cs.uiuc.edu">areitz@cs.uiuc.edu</a>)
 */
public class SimpleDPRPServer {
    /**
     * Controls the printing of debug output to stdout.
     */
    public static boolean DEBUG = false;
    /**
     * The version string, for the "-v" parameter.
     */
    public static final String VERSION = "1.3";

    // These two variables mean that you can't run two different
    // servers in the same VM. I would have to have server object
    // instances in order to get around this...
    /**
     * The <b><tt>ServerPortMgr</tt></b> instance for this server.
     */
    public static ServerPortMgr PortManager;
    /**
     * Maps "clientIP:port" strings to
     * <b><tt>DPRPServerSession</tt></b> instances.
     */
    public static HashMap sessionMap;

    /**
     * The start of control for this allegedly-simple DPRP server. This
     * method is not a mere skeleton -- it actually has a lot of functionality.
     * It first instantiates the <b><tt>ServerPortMgr</tt></b> singleton,
     * and then it parses its command-line parameters. It guarantees that
     * both the internal and external IP addresses are supplied, and if
     * a host configuration file is specified, it reads that in, and builds
     * the <b><tt>DPRPRestriction</tt></b> map. Addionally, if it finds a
     * stored lease file, this will also be sucked in, and on disk leases
     * restored.
     *
     * <p>
     *
     * Moving into the main server loop, we block on the internal socket,
     * waiting for a client request. When such a request is received, it
     * is matched against a map of open requests. If a previous request
     * is found, this new packet is routed to the appropriate handling
     * thread. Otherwise, after some error-checks succeed, a new thread
     * is spawned in order to handle this packet (as it represents a new
     * session).
     * @param args The command line parameters.
     */
    public static void main (String[] args) {
	InetAddress internalIP = null;
	InetAddress externalIP = null;
	int internalPort = DPRPMessage.DEFAULT_PORT;
	DPRPMessage recvdMessage = null;
	DPRPSocket serverSock = null;
	String configFileName = null;
	// Name of the lease file to open, with default.
	String leaseFileName = "dprp_lease_db.txt";

	String thrName;                         // Thread name.
	recvdMessage = new DPRPMessage();

	// Initialize necessary instances.
	sessionMap = new HashMap();
	PortManager = new ServerPortMgr();

	/*
	 * Do that thing (that you do) with the command-line parameters.
	 */
	try {
	    Integer tmp_i = null;
	    int i = 0;
	    for (i = 0; i < args.length; i++) {
		if (args[i].equals ("--intAddr") || args[i].equals ("-i"))
		    internalIP = InetAddress.getByName (args[++i]);
		else if (args[i].equals ("--extAddr") || args[i].equals ("-e"))
		    externalIP = InetAddress.getByName (args[++i]);
		else if (args[i].equals ("--intPort") ||
			 args[i].equals ("-s")) {
		    tmp_i = new Integer (args[++i]);
		    internalPort = tmp_i.intValue();
		} else if (args[i].equals("--debug") || args[i].equals("-d")) {
		    DEBUG = true;
		    // Goddamn jar files.
		    DPRPLease.DEBUG = true;
		} else if (args[i].equals ("--help") || args[i].equals ("-h"))
		    printCommandUsage();
		else if (args[i].equals ("--version") ||
			 args[i].equals ("-v")) {
		    System.err.println("SimpleDPRPServer version: " + VERSION);
		    System.exit(0);
		} else if (args[i].equals ("--cfgFile") ||
			   args[i].equals ("-f")) {
		    configFileName = args[++i];
		} else if (args[i].equals ("--leaseFile") ||
			   args[i].equals ("-l")) {
		    leaseFileName = args[++i];
		} else if (args[i].equals ("--extPortRange") ||
			   args[i].equals ("-r")) {
		    tmp_i = new Integer (args[++i]);
		    int sp = tmp_i.intValue();
		    if (!args[++i].equals ("-"))
			throw new Exception ("Invalid external port range " +
					     "specified.");
		    tmp_i= new Integer (args[++i]);
		    int lp = tmp_i.intValue();
		    if (!PortManager.setExtPortRange (sp, lp))
			throw new Exception ("Invalid external port range " +
					     "specified.");
		} else
		    throw new Exception ("Invalid parameter '" + args[i] +
					 "' specified.");
	    } /* for */
	} catch (UnknownHostException e) {
	    System.err.println ("Invalid IP address specified: " + e);
	    printCommandUsage();
	} catch (NumberFormatException e) {
	    System.err.println ("Invalid port specified.");
	    printCommandUsage();
	} catch (Exception e) {
	    System.err.println (e);
	    printCommandUsage();
	}

	// Enforce required parameters.
	if ((internalIP == null) || (externalIP == null)) {
	    System.err.println ("Error: must specify both internal " +
				"and external IP addresses for this gateway.");
	    printCommandUsage();
	}

	// This really deserves commenting.
	PortManager.setExternalIP (externalIP);

	// OS detection.
	if (System.getProperty ("os.name").equals ("Linux") &&
	    System.getProperty ("user.name").equals ("root") &&
	    (new File ("/sbin/iptables")).exists())
	    PortManager.markAsLinux();
	else {
	    System.out.println ("***************************************" +
				"***************************************");
	    System.out.println ("Warning: Linux OS with iptables was " +
				"not detected. This could mean that\nthis " +
				"VM is not running on the Linux platform, " +
				"this VM is not runnning as\nuser 'root', " +
				"or that this VM doesn't have access to a " +
				"valid netfilter\ninstallation. This " +
				"program will continue to run, but in " +
				"test-mode only.");
	    System.out.println ("***************************************" +
				"***************************************");
	}

	// Suck in the configuration file, if we have one.
	int numHostsFound = 0;
	try {
	    if (configFileName != null) {
		numHostsFound = PortManager.parseConfigFile (configFileName);
		if (DEBUG)
		    System.out.println ("Read configuration data for " +
					numHostsFound + " hosts.");
	    }
	} catch (DPRPConfigException e) {
	    System.err.print ("\nERROR: Problem with host security " +
			      "configuration file:\n\t");
	    System.err.println (e.getErrorMsg());
	    PortManager.printConfigFormat();
	    System.exit (2);
	} catch (IOException e) {
	    System.out.println (e);
	    System.exit (3);
	}

	// Suck in any leases that are lying around on disk.
	try {
	    int leaseCount = PortManager.readLeaseFile (leaseFileName);
	    if (DEBUG)
		System.out.println ("Sucked in " + leaseCount +
				    " leases from file: '" + leaseFileName
				    + "'.");
	} catch (IOException e) {
	    if (DEBUG) {
		System.out.println ("Got exception whist trying to " +
				    "read-in past leases.");
		System.out.println (e);
	    }
	}

	/*
	 * Simply listening on our default port, on all interfaces,
	 * presents an absolutely disasterous security hole. So, we're
	 * not going to do that, ever. (we don't want the outside
	 * sending us port-opening control packets, eh?)
	 */
	try {
	    serverSock = new DPRPSocket(internalPort, internalIP);
	    serverSock.setSoTimeout (0);
	} catch (SocketException e) {
	    System.err.println ("Socket could not be created with specified " +
				"IP address/port combination. Either the " +
				"port is not available, or the IP address ");
	    System.err.println ("doesn't exist on this system. Please " +
				"change these parameters, and try again.");
	    System.err.println ("\tIP Address specified: " +
				internalIP.getHostAddress());
	    System.err.println ("\tUDP Port specified: " + internalPort);
	    System.err.println ("Exception: " + e);
	    System.exit(1);
	} catch (SecurityException e2) {}   // This should never happen.

	if (DEBUG) {
	    System.out.println ("*******************************************");
	    System.out.println ("Server "+VERSION+" initialized. Parameters:");
	    System.out.println ("\tInternal IP Address: " +
				internalIP.getHostAddress());
	    System.out.println ("\tInternal Port: " + internalPort);
	    System.out.println ("\tExternal IP Address: " +
				externalIP.getHostAddress());
	    System.out.println ("\tExternal port range: " +
				PortManager.getMinExtPort() + " - " +
				PortManager.getMaxExtPort());
	    System.out.println ("\tisLinux: " + PortManager.isLinux());
	    System.out.println ("\tConfig file name: " + configFileName);
	    System.out.println ("\tLease output file name: " + leaseFileName);
	    System.out.println ("Now, Waiting for client(s):");
	    System.out.println ("*******************************************");
	}

	/*
	 * Go into the standard server infinite receive loop.
	 */
	while (true) {
	    // I think that I want to receive forever here.
	    if (!serverSock.receive(recvdMessage)) {
		System.out.println ("*** Error: got false from receive! ***");
		System.exit(69);
	    }

	    // If this client is set as 'denied' in the ServerPortManager,
	    // then knock them down right away.
	    if (PortManager.checkDenied (recvdMessage.getSourceAddress())) {
		if (DEBUG)
		    System.out.println ("SPS: got message from denied host: " +
			  recvdMessage.getSourceAddress().getHostAddress() +
			  ". Nooge.");

		// Send an error packet to this foolio.
		recvdMessage.flipSendRecv();
		serverSock.sendError (recvdMessage,
				      DPRPErrors.CLIENT_DENIED);
		continue;
	    }

	    // If this packet corresponds to a currently-running session,
	    // then hand it off to said session.
	    if (sessionMap.containsKey (recvdMessage.getSourceKey())) {
		DPRPServerSession s;
		s = (DPRPServerSession) sessionMap.get (recvdMessage.getSourceKey());
		try {
		    // test this by forging a source address on incoming
		    // packets. Brilliant!
		    //
		    // bigger problem, right now, is how to actually
		    // *pass* this blasted message into a currently
		    // running thread. i could maybe set a variable in it,
		    // and then interrupt it? what happens if i invoke a
		    // foolish method?
		    s.passedMessage = recvdMessage;
		    s.interrupt();
		} catch (SecurityException e) {
		    System.err.println (e);
		}
		continue;
	    } /* if containsKey. */

	    // If we're here, then we're making a new session for the
	    // 'CLI_REQUEST'. Start by setting the initial thread name.
	    thrName = "thr_" + recvdMessage.getSourceAddress().getHostAddress()
		+ ":" + recvdMessage.getSourcePort() + "->";

	    if (recvdMessage.getMessageType() == DPRPMessage.CLI_REQUEST) {
		if (DEBUG)
		    System.out.println ("SPS: Got a 'CLI_REQUEST' from: " +
			     recvdMessage.getSourceAddress().getHostAddress() +
			     ":" + recvdMessage.getSourcePort());

		// Instantiate a new server.
		DPRPServerSession pdss = new
		    DPRPServerSession (recvdMessage, thrName, externalIP);

		// We haven't seen this request, so add it to the map.
		sessionMap.put (recvdMessage.getSourceKey(), pdss);
		if (DEBUG)
		    System.out.println ("SPS: New 'CLI_REQUEST' added to map" +
					" @ key: " +
					recvdMessage.getSourceKey());

		// Start the thread.
		pdss.start();
	    } else if (recvdMessage.getMessageType() == DPRPMessage.CLI_RENEW_REQUEST) {
		// Got a renew request. First check to see if client is
		// allowed to renew.
		if (!PortManager.renewOK (recvdMessage.getSourceAddress())) {
		    // Doh! Not allowed to renew. Notify 'em.
		    if (DEBUG)
			System.out.println ("SPS: Got RENEW request from '" +
			     recvdMessage.getSourceAddress().getHostAddress() +
			     "' that is barred by rule.");
		    recvdMessage.flipSendRecv();
		    recvdMessage.setMessageType (DPRPMessage.SRV_RENEW_NACK);
		    recvdMessage.setFlags(DPRPErrors.CLIENT_NOT_ALLOWED_RENEW);
		    try {
			serverSock.send (recvdMessage);
		    } catch (IOException e) {
			System.out.println ("** SPS-Error: problem sending " +
					    "'SRV_RENEW_NACK' by rule:");
			System.out.println (e);
			continue;
		    } /* catch */

		    // Snuff this lease from the db.
		    DPRPLease tmpLea = new DPRPLease (recvdMessage, true);
		    PortManager.delReservation (tmpLea, true);
		    continue;
		} /* if not renewOK */

		// Renew permitted -- start a new session, then add it.
		DPRPServerSession ps = new
		    DPRPServerSession (recvdMessage, thrName, externalIP);
		sessionMap.put (recvdMessage.getSourceKey(), ps);
		ps.start();
	    } else if (recvdMessage.getMessageType() == DPRPMessage.CLI_TERMINATE_LEASE) {
		// Snuff the lease that corresponds to the given parameters.
		// This is a quick hack. Instead of spawning a new session,
		// that will guarantee that the ACK is sent out, we'll
		// just send the ACK, and rely on the error message to
		// clear things up for the client in case it isn't
		// received.
		if (DEBUG)
		    System.out.println ("<< Got TERM_LEASE from " +
			     recvdMessage.getSourceAddress().getHostAddress() +
			     ":" + recvdMessage.getSourcePort() +
                             " for lease: " + recvdMessage.getPrimaryKey() +
			     " >>");

		DPRPLease rmLea = PortManager.getReservation (recvdMessage);
		if (rmLea == null) {
		    // Lease not found, so send 'em an error, jimmy!
		    if (DEBUG)
			System.out.println ("<< TERM_LEASE of key: " +
					    recvdMessage.getPrimaryKey() +
					    " not found! >>");
		    recvdMessage.flipSendRecv();
		    serverSock.sendError (recvdMessage,
					  DPRPErrors.CLIENT_LEASE_NOT_FOUND);
		} else {
		    // Ack to the wack!
		    PortManager.delReservation (rmLea, true);
		    recvdMessage.flipSendRecv();
		    recvdMessage.setMessageType(DPRPMessage.SRV_TERMINATE_ACK);
		    if (DEBUG) {
			System.out.println ("<< TERM_LEASE of key: " +
					    recvdMessage.getPrimaryKey() +
					    " found, ACK'ing: >>");
			recvdMessage.printDump();
		    }
		    try {
			serverSock.send (recvdMessage);
		    } catch (IOException e) {
			if (DEBUG)
			    System.out.println ("SPS: error sending " +
						"terminate ack: " + e);
		    } /* catch */
		}
	    } else {
		if (DEBUG)
		    System.out.println ("SPS: Got a request of type " +
				    recvdMessage.getMessageType() +
				    " from: " + recvdMessage.getSourceKey() +
				    ". Discarding...");

		// Send an error packet to this foolio.
		recvdMessage.flipSendRecv();
		serverSock.sendError (recvdMessage,
				      DPRPErrors.INVALID_PACKET_TYPE);
	    } // else
	} // while.
    } /* End main(). */

    /**
     * Removes a "session" from the <b><tt>sessionMap</tt></b>. When a
     * session finishes, the last thing it does is to remove it's
     * identifier from the map, affecting a full cleanup.
     * @param sourceKey The source identifier, serves as the key in
     *  the map.
     */
    public synchronized static void removeSession (String sourceKey) {
	DPRPServerSession ss = (DPRPServerSession)sessionMap.remove(sourceKey);
	if (ss == null) {
	    // Uh-oh, that should have worked.
	    System.out.println ("** SPS-Error: couldn't remove session for: " +
				sourceKey + ".");
	}
	if (DEBUG) {
	    System.out.println ("SPS: Session of cliNfo:" + sourceKey +
				" removed.");
	    System.out.println ("SPS: # of outstanding connections " +
				"remaining in map: " +
				sessionMap.size());
	    if (sessionMap.size() > 0) {
		Set ks = sessionMap.keySet();
		Iterator ksi = ks.iterator();
		System.out.println ("Dumping list of all keys:");
		int ci = 0;
		while (ksi.hasNext()) {
		    String st = (String)ksi.next();
		    System.out.println ("\t" + ci + ": cliNfo = " + st);
		    ci++;
		} /* while */
	    } /* if size */
	} /* debug */
    } /* end removeSession(). */

    /**
     * Simply dumps a message to stderr about how to start this utility.
     */
    private static void printCommandUsage() {
	System.err.println ("Usage: SimpleDPRPServer -i <Internal " +
			    "Address> -e <External Address> [OPTIONS]\n\t " +
			    "i.e. SimpleDPRPServer -i 10.0.0.1 -e 10.10.10.2");
	System.err.println ("\nOptions:");
	System.err.println ("  -h,  --help\t\tDisplay this help message.");
	System.err.println ("  -v,  --version\tDisplay server version.");
	System.err.println ("  -d,  --debug\t\tDisplay debugging messages.");
	System.err.println ("  -i,  --intAddr\tSet the IP Address of the " +
			    "internal interface.");
	System.err.println ("  -s,  --intPort\tSet the listening port on " +
			    "the internal interface.");
	System.err.println ("  -e,  --extAddr\tSet the IP address of the " +
			    "external interface.");
	System.err.println ("  -r,  --extPortRange\tSet the range of " +
			    "external ports to allocate from.");
	System.err.println ("  -f,  --cfgFile\tSpecify the host security " +
			    "configuration file.");
	System.err.println ("  -l,  --leaseFile\tSpecify the name of the " +
			    "on-disk lease database.");
	System.exit(1);
    } /* End printCommandUsage(). */
} /* End class SimpleDPRPServer. */
